/******************************************************************************
* Cagd_Arc.c - Curve representation of arcs and circles			      *
*******************************************************************************
* (C) Gershon Elber, Technion, Israel Institute of Technology                 *
*******************************************************************************
* Written by Gershon Elber, Jun. 90.					      *
******************************************************************************/

#include <ctype.h>
#include <stdio.h>
#include <string.h>
#include "irit_sm.h"
#include "extra_fn.h"
#include "cagd_loc.h"

#define CONIC_MAX_VAL		-10
#define CONIC_MIN_VAL		10
#define CONIC_DELTA_VAL		(CONIC_MAX_VAL - CONIC_MIN_VAL)

#define CAGD_QUADRIC_EXTENT	(CONIC_MAX_VAL)

#define QUADRIC_INVERT(x)	((x) == 0 ? IRIT_INFNTY : 1.0 / sqrt(x))

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Construct rational quadratic curve out of the 6 coefficients of the      M
* conic section: A x^2 + B xy + C y^2 + D x + E y + F = 0.		     M
*   Based on:								     M
* Bezier Curves and Surface Patches on Quadrics, by Josef Hoschek,           M
* Mathematical methods in Computer aided Geometric Design II, Tom Lyche and  M
* Larry L. Schumaker (eds.), pp 331-342, 1992.				     M
*                                                                            *
* PARAMETERS:                                                                M
*   A, B, C, D, E, F:   The six coefficients of the conic curve.             M
*   ZLevel:		Sets the Z level of this XY parallel conic curve.    M
*                                                                            *
* RETURN VALUE:                                                              M
*   CagdCrvStruct *:  A quadratic curve representing the conic.		     M
*                                                                            *
* SEE ALSO:                                                                  M
*   BspCrvCreateCircle, BspCrvCreateUnitCircle, CagdCrvCreateArc,	     M
*   BzrCrvCreateArc, CagdCreateConicCurve2, CagdCreateQuadricSrf	     M
*                                                                            *
* KEYWORDS:                                                                  M
*   CagdCreateConicCurve                                                     M
*****************************************************************************/
CagdCrvStruct *CagdCreateConicCurve(CagdRType A,
				    CagdRType B,
				    CagdRType C,
				    CagdRType D,
				    CagdRType E,
				    CagdRType F,
				    CagdRType ZLevel)
{
    VectorType Trans;
    CagdRType **Points, 
	RotAngle = APX_EQ(B, 0.0) ? 0.0 : atan2(B, A - C) / 2.0,
	A1 = ((A + C) + B * sin(2 * RotAngle) +
	      (A - C) * cos(2 * RotAngle)) / 2.0,
	B1 = B * cos(2 * RotAngle) - (A - C) * sin(2 *RotAngle),
	C1 = ((A + C) - B * sin(2 * RotAngle) -
	      (A - C) * cos(2 * RotAngle)) / 2.0,
	D1 = D * cos(RotAngle) + E * sin(RotAngle),
	E1 = -D * sin(RotAngle) + E * cos(RotAngle),
	F1 = F;
    MatrixType Mat;
    CagdCrvStruct *PwrCrv,
	*BzrCrv = NULL;

    if (!APX_EQ(B1, 0.0) ||
	(APX_EQ(A1, 0.0) && APX_EQ(C1, 0.0))) {
	CAGD_FATAL_ERROR(CAGD_ERR_INVALID_CONIC_COEF);
	return NULL;
    }
    if (APX_EQ(A1, 0.0) || APX_EQ(C1, 0.0)) {
	/* It is a parabola. */
	return CagdCreateConicCurve2(A, B, C, D, E, F, ZLevel, NULL, NULL);
    }

    /* We now have conic: A1 x^2 + C1 y^2 + D1 x + E1 y + F1 = 0.            */
    /* Compute the translation factors by converting to,		     */
    /* A1(x + D1/(2A1))^2 + C1(y + E1/(2C1))^2 +			     */
    /*                                        F1 - D1^2/4A1 - E1^2/4C1 = 0.  */
    Trans[0] = -D1 / (2.0 * A1);
    Trans[1] = -E1 / (2.0 * C1);
    Trans[2] = ZLevel;

    F1 -= SQR(Trans[0]) + SQR(Trans[1]);

    if (A1 < 0.0) {			        /* Make sure A1 is positive. */
	A1 = -A1;
	C1 = -C1;
	F1 = -F1;
    }
    if (A1 * C1 > 0.0) {
        /* Conic is now A1 x^2 + C1 y^2 = -F1,    A1 > 0,  C1 > 0,  -F1 > 0. */
	/* The above ellipse can be written as a parametrization of,	     */
	/*              x = (1/w) * (a^2 - b^2) / sqrt(A1),		     */
	/*              y = (1/w) * 2ab / sqrt(C1),			     */
	/* and w equals w = (a^2 + b^2) / sqrt(-F1).			     */
	/* The only (large) remaining question is how to parametrize a, b.   */
	/* For no better selection, we select a = t, b = 1 - t, or	     */
	/*              x = (1/w) * (-1 + 2 * t) / sqrt(A1),		     */
	/*              y = (1/w) * (2 * t - 2 * t^2) / sqrt(C1),	     */
	/* and w equals w = (1 - 2 * t + 2 * t^2) / sqrt(-F1).		     */
	if (F1 >= 0.0) {
	    CAGD_FATAL_ERROR(CAGD_ERR_INVALID_CONIC_COEF);
	    return NULL;
	}
	A1 = 1.0 / sqrt(A1);
	C1 = 1.0 / sqrt(C1);
	F1 = 1.0 / sqrt(-F1);  

	PwrCrv = CagdCrvNew(CAGD_CPOWER_TYPE, CAGD_PT_P2_TYPE, 3);
	Points = PwrCrv -> Points;
    
	Points[0][0] = F1;				      /* The w term. */
	Points[0][1] = -2.0 * F1;
	Points[0][2] = 2.0 * F1;
	Points[1][0] = -A1;				      /* The x term. */
	Points[1][1] = 2.0 * A1;
	Points[1][2] = 0.0;
	Points[2][0] = 0.0;				      /* The y term. */
	Points[2][1] = 2.0 * C1;
	Points[2][2] = -2.0 * C1;
    
	BzrCrv = CnvrtPower2BezierCrv(PwrCrv);
	CagdCrvFree(PwrCrv);
    }
    else {
        CagdBType SwapXY;

	if (F1 > 0.0) {
	    SWAP(CagdRType, A1, C1);
	    A1 = -A1;
	    C1 = -C1;
	    F1 = -F1;
	    SwapXY = TRUE;
	}
	else
	    SwapXY = FALSE;

	C1 = -C1;
        /* Conic is now A1 x^2 - C1 y^2 = -F1,    A1 > 0,  C1 > 0,  -F1 > 0. */
	/* The above hyperbola can be written as a parametrization of,	     */
	/*              x = (1/w) * (a^2 + b^2) / sqrt(A1),		     */
	/*              y = (1/w) * 2ab / sqrt(C1),			     */
	/* and w equals w = (a^2 - b^2) / sqrt(-F1).			     */
	/* The only (large) remaining question is how to parametrize a, b.   */
	/* For no better selection, we select a = t, b = 1 - t, or	     */
	/*              x = (1/w) * (1 - 2 * t + 2 * t^2) / sqrt(A1),	     */
	/*              y = (1/w) * (2 * t - 2 * t^2) / sqrt(C1),	     */
	/* and w equals w = (-1 + 2 * t) / sqrt(-F1).			     */
	A1 = 1.0 / sqrt(A1);
	C1 = 1.0 / sqrt(C1);
	F1 = F1 == 0.0 ? IRIT_INFNTY : 1.0 / sqrt(-F1);  

	PwrCrv = CagdCrvNew(CAGD_CPOWER_TYPE, CAGD_PT_P2_TYPE, 3);
	Points = PwrCrv -> Points;
    
	Points[0][0] = -F1;				      /* The w term. */
	Points[0][1] = 2.0 * F1;
	Points[0][2] = 0.0;
	Points[1][0] = A1;				      /* The x term. */
	Points[1][1] = -2.0 * A1;
	Points[1][2] = 2.0 * A1;
	Points[2][0] = 0.0;				      /* The y term. */
	Points[2][1] = 2.0 * C1;
	Points[2][2] = -2.0 * C1;

	if (SwapXY) {
	    SWAP(CagdRType, Points[1][0], Points[2][0]);
	    SWAP(CagdRType, Points[1][1], Points[2][1]);
	    SWAP(CagdRType, Points[1][2], Points[2][2]);
	}
	
	BzrCrv = CnvrtPower2BezierCrv(PwrCrv);
	CagdCrvFree(PwrCrv);
    }

    CagdCrvTransform(BzrCrv, Trans, 1.0);
    MatGenMatRotZ1(RotAngle, Mat);
    CagdCrvMatTransform(BzrCrv, Mat);

    CAGD_SET_GEOM_TYPE(BzrCrv, CAGD_GEOM_CONIC_SEC);

    return BzrCrv;
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Construct rational quadratic curve out of the 6 coefficients of the      M
* conic section: A x^2 + B xy + C y^2 + D x + E y + F = 0.		     M
*                                                                            *
* PARAMETERS:                                                                M
*   A, B, C, D, E, F:   The six coefficients of the conic curve.             M
*   ZLevel:		Sets the Z level of this XY parallel conic curve.    M
*   PStartXY, PEndXY:   Domain of conic section - starting/end points, in    M
*			the XY plane. If NULL, the most complete conic       M
*			possible is created.				     M
*                                                                            *
* RETURN VALUE:                                                              M
*   CagdCrvStruct *:  A quadratic curve representing the conic.		     M
*                                                                            *
* SEE ALSO:                                                                  M
*   BspCrvCreateCircle, BspCrvCreateUnitCircle, CagdCrvCreateArc,	     M
*   BzrCrvCreateArc, CagdCreateConicCurve, CagdCreateQuadricSrf		     M
*                                                                            *
* KEYWORDS:                                                                  M
*   CagdCreateConicCurve2                                                    M
*****************************************************************************/
CagdCrvStruct *CagdCreateConicCurve2(CagdRType A,
				     CagdRType B,
				     CagdRType C,
				     CagdRType D,
				     CagdRType E,
				     CagdRType F,
				     CagdRType ZLevel,
				     CagdRType *PStartXY,
				     CagdRType *PEndXY)
{
    VectorType Trans;
    CagdRType Desc, x, x1, x2, dx, y1, y2, dy,
	RotAngle = APX_EQ(B, 0.0) ? 0.0 : atan2(B, A - C) / 2.0,
	A1 = ((A + C) + B * sin(2 * RotAngle) +
	      (A - C) * cos(2 * RotAngle)) / 2.0,
	B1 = B * cos(2 * RotAngle) - (A - C) * sin(2 *RotAngle),
	C1 = ((A + C) - B * sin(2 * RotAngle) -
	      (A - C) * cos(2 * RotAngle)) / 2.0,
	D1 = D * cos(RotAngle) + E * sin(RotAngle),
	E1 = -D * sin(RotAngle) + E * cos(RotAngle),
	F1 = F;
    MatrixType Mat;
    CagdCrvStruct
	*Crv = NULL;

    if (!APX_EQ(B1, 0.0) ||
	(APX_EQ(A1, 0.0) && APX_EQ(C1, 0.0))) {
	CAGD_FATAL_ERROR(CAGD_ERR_INVALID_CONIC_COEF);
	return NULL;
    }

    /* Computes the translation factors. */
    Trans[0] = Trans[1] = 0.0;
    Trans[2] = ZLevel;

    /* We now have conic: A1 x^2 + C1 y^2 + D1 x + E1 y + F1 = 0. */

    Desc = A1 * C1;

    if (APX_EQ_EPS(Desc, 0.0, IRIT_UEPS)) {
	CagdCrvStruct
	    *PwrCrv = CagdCrvNew(CAGD_CPOWER_TYPE, CAGD_PT_E3_TYPE, 3);
	 CagdRType
	     **Points= PwrCrv -> Points;

	/* Its a parabola. */

	PwrCrv -> Order = PwrCrv -> Length = 3;
	ZAP_MEM(PwrCrv -> Points[1], 3 * sizeof(CagdRType));
	ZAP_MEM(PwrCrv -> Points[2], 3 * sizeof(CagdRType));
	ZAP_MEM(PwrCrv -> Points[3], 3 * sizeof(CagdRType));

	if (!APX_EQ(A1, 0.0) && !APX_EQ(E1, 0.0)) {
	    /* Update the translation factors. */
	    Trans[0] = -D1 / (2.0 * A1);

	    /* Update the control points. */
	    if (PStartXY != NULL && PEndXY != NULL) {
		x1 = PStartXY[0] - Trans[0];
		x2 = PEndXY[0]   - Trans[0];
		y1 = PStartXY[1] - Trans[1];
		y2 = PEndXY[1]   - Trans[1];
		dx = x2 - x1;
		dy = y2 - y1;
	    }
	    else {
		x1 = CONIC_MIN_VAL - Trans[0];
		x2 = CONIC_MAX_VAL - Trans[0];
		y1 = CONIC_MIN_VAL - Trans[1];
		y2 = CONIC_MAX_VAL - Trans[1];
		dx = CONIC_DELTA_VAL;
		dy = CONIC_DELTA_VAL;
	    }

	    x  = x1 * cos(RotAngle) - y1 * sin(RotAngle);
	    y1 = x1 * sin(RotAngle) + y1 * cos(RotAngle);
	    x1 = x;

	    Points[1][0] = x1;
	    Points[1][1] = dx;
	    Points[2][0] = SQR(x1) * (- A1 / E1);
	    Points[2][1] = 2 * dx * x1 * (- A1 / E1);
	    Points[2][2] = SQR(dx) * (- A1 / E1);
	}
	else if (!APX_EQ(C1, 0.0) && !APX_EQ(D1, 0.0)) {
	    /* Update the translation factors. */
	    Trans[1] = -E1 / (2.0 * C1);

	    /* Update the control points. */
	    if (PStartXY != NULL && PEndXY != NULL) {
		x1 = PStartXY[0] - Trans[0];
		x2 = PEndXY[0]   - Trans[0];
		y1 = PStartXY[1] - Trans[1];
		y2 = PEndXY[1]   - Trans[1];
		dx = x2 - x1;
		dy = y2 - y1;
	    }
	    else {
		x1 = CONIC_MIN_VAL - Trans[0];
		x2 = CONIC_MAX_VAL - Trans[0];
		y1 = CONIC_MIN_VAL - Trans[1];
		y2 = CONIC_MAX_VAL - Trans[1];
		dx = CONIC_DELTA_VAL;
		dy = CONIC_DELTA_VAL;
	    }

	    x  = x1 * cos(RotAngle) - y1 * sin(RotAngle);
	    y1 = x1 * sin(RotAngle) + y1 * cos(RotAngle);
	    x1 = x;

	    Points[1][0] = SQR(y1) * (- C1 / D1);
	    Points[1][1] = 2 * dy * y1 * (- C1 / D1);
	    Points[1][2] = SQR(dy) * (- C1 / D1);
	    Points[2][0] = y1;
	    Points[2][1] = dy;
	}
	else {
	    CAGD_FATAL_ERROR(CAGD_ERR_INVALID_CONIC_COEF);
	    return NULL;
	}
	Crv = CnvrtPower2BezierCrv(PwrCrv);
	CagdCrvFree(PwrCrv);
    }
    else {
	/* Update the translation factors. */
	Trans[0] = -D1 / (2.0 * A1);
	Trans[1] = -E1 / (2.0 * C1);

	if (PStartXY != NULL && PEndXY != NULL) {
	    x1 = PStartXY[0] - Trans[0];
	    x2 = PEndXY[0]   - Trans[0];
	    y1 = PStartXY[1] - Trans[1];
	    y2 = PEndXY[1]   - Trans[1];
	    x  = x1 * cos(RotAngle) - y1 * sin(RotAngle);
	    y1 = x1 * sin(RotAngle) + y1 * cos(RotAngle);
	    x1 = x;
	    dx = x2 - x1;
	    dy = y2 - y1;
	}

	F1 = (4 * A1 * C1 * F1 - C1 * SQR(D1) - A * SQR(E1)) / (4 * A1 * C1);

	/* We now have conic in canonic form of: A1 x^2 + C1 y^2 + F1 = 0. */
	if (Desc > 0) {
	    CagdRType R[2];

	    /* Its an ellipse. */
	    if (F1 / A1 < 0 && F1 / C1 < 0) {
		STATIC_DATA CagdPtStruct
		    POrigin = { NULL, NULL, { 0.0, 0.0, 0.0 } };

		R[0] = sqrt(-F1 / A1);
		R[1] = sqrt(-F1 / C1);

		if (PStartXY != NULL && PEndXY != NULL) {
		    CagdRType StartAngle, EndAngle;

		    /* Create an ellipse from PStartXY to PEndXY! */
		    StartAngle = atan2(y1 / R[1], x1 / R[0]);
		    if (StartAngle < 0.0)
			StartAngle = RAD2DEG(StartAngle) + 360.0;
		    else
			StartAngle = RAD2DEG(StartAngle);
		    EndAngle = atan2(y2 / R[1], x2 / R[0]);
		    if (EndAngle < 0.0)
			EndAngle = RAD2DEG(EndAngle) + 360.0;
		    else
			EndAngle = RAD2DEG(EndAngle);

		    Crv = CagdCrvCreateArc(&POrigin, 1.0,
					   StartAngle, EndAngle);
		}
		else {
		    /* Create a full ellipse! */
		    Crv = BspCrvCreateUnitCircle();
		}
	    }
	    else {
		CAGD_FATAL_ERROR(CAGD_ERR_INVALID_CONIC_COEF);
		return NULL;
	    }

	    /* Scale the circular shape into an ellipse. */
	    MatGenMatScale(R[0], R[1], 1.0, Mat);
	    CagdCrvMatTransform(Crv, Mat);
	}
	else { /* t < 0 */
	    /* Its an hyperbola. */
	    CAGD_FATAL_ERROR(CAGD_ERR_HYPERBOLA_NO_SUPPORT);
	    return NULL;
	}
    }

    CagdCrvTransform(Crv, Trans, 1.0);
    MatGenMatRotZ1(RotAngle, Mat);
    CagdCrvMatTransform(Crv, Mat);

    CAGD_SET_GEOM_TYPE(Crv, CAGD_GEOM_CONIC_SEC);

    return Crv;
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Construct rational quadric surface out of the 9 coefficients of:         M
* A x^2 + B y^2 + C z^2 + D xy + E xz + F yz + G x + H y + I z + J = 0.      M
*   Based on:								     M
* Bezier Curves and Surface Patches on Quadrics, by Josef Hoschek,           M
* Mathematical methods in Computer aided Geometric Design II, Tom Lyche and  M
* Larry L. Schumaker (eds.), pp 331-342, 1992.				     M
*                                                                            *
* PARAMETERS:                                                                M
*   A, B, C, D, E, F, G, H, I, J:   The ten coefficients of the quadric.     M
*                                                                            *
* RETURN VALUE:                                                              M
*   CagdSrfStruct *:  A quadric surface representing the given form.	     M
*                                                                            *
* SEE ALSO:                                                                  M
*   BspCrvCreateCircle, BspCrvCreateUnitCircle, CagdCrvCreateArc,	     M
*   BzrCrvCreateArc, CagdCreateConicCurve, CagdCreateConicCurve2	     M
*                                                                            *
* KEYWORDS:                                                                  M
*   CagdCreateQuadricSrf                                                     M
*****************************************************************************/
CagdSrfStruct *CagdCreateQuadricSrf(CagdRType A,
				    CagdRType B,
				    CagdRType C,
				    CagdRType D,
				    CagdRType E,
				    CagdRType F,
				    CagdRType G,
				    CagdRType H,
				    CagdRType I,
				    CagdRType J)
{
    CagdBType
	SwitchXY = FALSE,
	SwitchXZ = FALSE,
	SwitchYZ = FALSE;
    int i;
    CagdRType **Points, N[9], b[3];
    VectorType Trans;
    MatrixType Mat, Mat2, M, U, d, V;
    CagdSrfStruct *PwrSrf,
	*BzrSrf = NULL;

    /* Computes the translation factors.  Substituting (x-X0) for x (and     */
    /* same for y and z), one gets the following constraints for the linear  */
    /* terms to vanish (G, H, I in the original general quadric):	     */
    /* [2A   D   E] [X0] = [G]						     */
    /* [ D  2B   F] [Y0] = [H]						     */
    /* [ E   F  2C] [Z0] = [I]						     */
    /* The solution for X0, Y0, Z0 zeros these linear terms and yields new   */
    /* constant J term: J + A X0^2 + B Y0^2 + C Z0^2 +			     */
    /*			D X0 Y0 + E X0 Z0 + F Y0 Z0 - G X0 - H Y0 - I Z0.    */
    N[0] = 2 * A;     N[3] = D;      	N[6] = E;
    N[1] = D;	      N[4] = 2 * B;	N[7] = F; 
    N[2] = E;	      N[5] = F;		N[8] = 2 * C;
    b[0] = G;
    b[1] = H;
    b[2] = I;
    IritQRUnderdetermined(N, NULL, NULL, 3, 3);
    IritQRUnderdetermined(NULL, Trans, b, 3, 3);

    /* Find the translational factors. */
    J += A * SQR(Trans[0]) + B * SQR(Trans[1]) + C * SQR(Trans[2])
	     + D * Trans[0] * Trans[1]
	     + E * Trans[0] * Trans[2]
	     + F * Trans[1] * Trans[2]
	     - G * Trans[0] - H * Trans[1] - I * Trans[2];
    G = H = I = 0.0;

    /* Compute the rotation factors.  Now we have a quadrics of the form:    */
    /* A x^2 + B y^2 + C z^2 + D xy + E xz + F yz + J = 0.		     */
    /* Writing the quadric form as:					     */
    /* [x y z 1] [A    D/2  E/2  G/2] [x] = [x y z 1] M [x] = 0		     */
    /*		 [D/2  B    F/2  H/2] [y]		[y]		     */
    /*		 [E/2  F/2  C    I/2] [z]		[z]		     */
    /*		 [G/2  H/2  I/2  J  ] [1]		[1]		     */
    /* we seek a rotation of M that make it diagonal.  Toward this end we    */
    /* employ SVD to decompose M into U N V, N diagonal and U and V are two  */
    /* rotational matrices that are also transpose (==inverse) of each other.*/
    M[0][0] = A;
    M[1][1] = B;
    M[2][2] = C;
    M[3][3] = J;

    M[0][1] = M[1][0] = D / 2;
    M[0][2] = M[2][0] = E / 2;
    M[1][2] = M[2][1] = F / 2;

    M[0][3] = M[3][0] = G / 2;
    M[1][3] = M[3][1] = H / 2;
    M[2][3] = M[3][2] = I / 2;

    JacobiMatrixDiag4x4(M, U, d, V);

    A = d[0][0];
    B = d[1][1];
    C = d[2][2];
    J = d[3][3];

    /* We now have quadric: A x^2 + B y^2 + C z^2 + J = 0. */
    if (J == 0.0) {
        if (A < 0.0) {
	    A = -A;
	    B = -B;
	    C = -C;
	}
    }
    else {
        if (J > 0.0) {			    /* Make sure J is not positive. */
	    A = -A;
	    B = -B;
	    C = -C;
	    J = -J;
	}

        if (A * J >= 0.0) {
	    if (B * J >= 0.0) {
	        if (C * J >= 0.0) {
		    CAGD_FATAL_ERROR(CAGD_ERR_INVALID_CONIC_COEF);
		    return NULL;
		}
		else {
		    SwitchXZ = TRUE;
		    SWAP(CagdRType, A, C);
		}
	    }
	    else {
	        SwitchXY = TRUE;
		SWAP(CagdRType, A, B);
	    }
	}
    }

    /* Now J is not positive and A is not negative. */

    if (B >= 0.0 && C >= 0.0) {
        /* Quadrics is now A x^2 + B y^2 + C z^2 + J w^2 = 0,	             */
	/*				    A >= 0, B >= 0, C >= 0, J <= 0.  */
	/* The above ellipsoid can be written as a parametrization of,	     */
	/*              x = (1/w) * (a^2 - b^2 - c^2) / sqrt(A),      	     */
	/*              y = (1/w) * 2ab / sqrt(B),			     */
	/*              z = (1/w) * 2ac / sqrt(C),			     */
	/* and w equals w = (a^2 + b^2 + c^2) / sqrt(-J).     		     */
	/* The only (large) remaining question is how to parametrize a, b, c.*/
	/* For no better selection, we select a = u, b = 1 - v, c = v, or    */
	/*              x = (1/w) * (u^2 - 1 + 2 * v - 2 * v^2) / sqrt(A),   */
	/*              y = (1/w) * (2 * u - 2 * uv) / sqrt(B),		     */
	/*              z = (1/w) * (2 * u * v) / sqrt(C),		     */
	/* and w equals w = (u^2 + 1 - 2 * v + 2 * v^2) / sqrt(-J).	     */

	A = QUADRIC_INVERT(A);
	B = QUADRIC_INVERT(B);
	C = QUADRIC_INVERT(C);
	J = -QUADRIC_INVERT(-J);

	PwrSrf = CagdSrfNew(CAGD_SPOWER_TYPE, CAGD_PT_P3_TYPE, 3, 3);
	Points = PwrSrf -> Points;

	for (i = 0; i <= 3; i++)
	    ZAP_MEM(Points[i], 9 * sizeof(CagdRType));

	/* Coefficients of the biquadratic surface are orders as follows     */
	/*    0) 1          1)  u         2) u^2			     */
	/*    3) v          4)  u * v     5) u^2 * v			     */
	/*    6) v^2        7)  u * v^2   8) u^2 * v^2			     */
	Points[0][0] = J;				      /* The w term. */
	Points[0][2] = J;
	Points[0][3] = -2.0 * J;
	Points[0][6] = 2.0 * J;
	Points[1][0] = -A;				      /* The x term. */
	Points[1][2] = A;
	Points[1][3] = 2.0 * A;
	Points[1][6] = -2.0 * A;
	Points[2][1] = 2.0 * B;				      /* The y term. */
	Points[2][4] = -2.0 * B;
	Points[3][4] = 2.0 * C;				      /* The z term. */

	BzrSrf = CnvrtPower2BezierSrf(PwrSrf);
	CagdSrfFree(PwrSrf);
    }
    else if (B <= 0.0 && C <= 0.0) {
        /* Quadrics is now A x^2 + B y^2 + C z^2 + J w^2 = 0,	             */
	/*				   A >= 0, B <= 0, C <= 0, J <= 0.   */
	/* The above hyperboloid of two sheets can be written as,	     */
	/*              x = (1/w) * (a^2 + b^2 + c^2) / sqrt(A),      	     */
	/*              y = (1/w) * 2ab / sqrt(B),			     */
	/*              z = (1/w) * 2ac / sqrt(C),			     */
	/* and w equals w = (a^2 - b^2 - c^2) / -sqrt(-J).     		     */
	/* The only (large) remaining question is how to parametrize a, b, c.*/
	/* For no better selection, we select a = u, b = 1 - v, c = v, or    */
	/*              x = (1/w) * (u^2 + 1 - 2 * v + 2 * v^2) / sqrt(A),   */
	/*              y = (1/w) * (2 * u - 2 * uv) / sqrt(B),		     */
	/*              z = (1/w) * (2 * u * v) / sqrt(C),		     */
	/* and w equals w = (u^2 - 1 + 2 * v - 2 * v^2) / sqrt(-J)           */

	A = QUADRIC_INVERT(A);
	B = QUADRIC_INVERT(-B);
	C = QUADRIC_INVERT(-C);
	J = -QUADRIC_INVERT(-J);

	PwrSrf = CagdSrfNew(CAGD_SPOWER_TYPE, CAGD_PT_P3_TYPE, 3, 3);
	Points = PwrSrf -> Points;

	for (i = 0; i <= 3; i++)
	    ZAP_MEM(Points[i], 9 * sizeof(CagdRType));

	/* Coefficients of the biquadratic surface are orders as follows     */
	/*    0) 1          1)  u         2) u^2			     */
	/*    3) v          4)  u * v     5) u^2 * v			     */
	/*    6) v^2        7)  u * v^2   8) u^2 * v^2			     */
	Points[0][0] = -J;				      /* The w term. */
	Points[0][2] = J;
	Points[0][3] = 2.0 * J;
	Points[0][6] = -2.0 * J;
	Points[1][0] = A;				      /* The x term. */
	Points[1][2] = A;
	Points[1][3] = -2.0 * A;
	Points[1][6] = 2.0 * A;
	Points[2][1] = 2.0 * B;				      /* The y term. */
	Points[2][4] = -2.0 * B;
	Points[3][4] = 2.0 * C;				      /* The z term. */

	BzrSrf = CnvrtPower2BezierSrf(PwrSrf);
	CagdSrfFree(PwrSrf);
    }
    else if ((B >= 0.0 && C <= 0.0) || (B <= 0.0 && C >= 0.0)) {
        if (B <= 0.0 && C >= 0.0) {
	    /* Exchange the role of the B and C coefficients. */
	    SWAP(CagdRType, B, C);
	    SwitchYZ = TRUE;
	}

	/* Quadrics is now A x^2 + B y^2 + C z^2 + J w^2 = 0,	             */
	/*				    A >= 0, B >= 0, C <= 0, J <= 0.  */
	/* The above hyperboloid of one sheet can be written as,	     */
	/*              x = (1/w) * (a^2 - b^2 + c^2) / sqrt(A),      	     */
	/*              y = (1/w) * 2ab / sqrt(B),			     */
	/*              z = (1/w) * 2ac / -sqrt(-C),			     */
	/* and w equals w = (a^2 + b^2 - c^2) / -sqrt(-J).     		     */
	/* The only (large) remaining question is how to parametrize a, b, c.*/
	/* For no better selection, we select a = u, b = 1 - v, c = v, or    */
	/*              x = (1/w) * (u^2 - 1 + 2 * v) / sqrt(A),	     */
	/*              y = (1/w) * (2 * u - 2 * uv) / sqrt(B),  	     */
	/*              z = (1/w) * (2 * u * v) / -sqrt(C),		     */
	/* and w equals w = (1 + u^2 - 2 * v) / -sqrt(-J).	             */

	A = QUADRIC_INVERT(A);
	B = QUADRIC_INVERT(B);
	C = -QUADRIC_INVERT(-C);
	J = -QUADRIC_INVERT(-J);

	PwrSrf = CagdSrfNew(CAGD_SPOWER_TYPE, CAGD_PT_P3_TYPE, 3, 3);
	Points = PwrSrf -> Points;

	for (i = 0; i <= 3; i++)
	    ZAP_MEM(Points[i], 9 * sizeof(CagdRType));

	/* Coefficients of the biquadratic surface are orders as follows     */
	/*    0) 1          1)  u         2) u^2			     */
	/*    3) v          4)  u * v     5) u^2 * v			     */
	/*    6) v^2        7)  u * v^2   8) u^2 * v^2			     */
	Points[0][0] = J;				      /* The w term. */
	Points[0][2] = J;
	Points[0][3] = -2.0 * J;
	Points[1][0] = -A;				      /* The x term. */
	Points[1][2] = A;
	Points[1][3] = 2.0 * A;
	Points[2][1] = 2.0 * B;				      /* The y term. */
	Points[2][4] = -2.0 * B;
	Points[3][4] = 2.0 * C;				      /* The z term. */
    
	BzrSrf = CnvrtPower2BezierSrf(PwrSrf);
	CagdSrfFree(PwrSrf);
    }
    else {
        CAGD_FATAL_ERROR(CAGD_ERR_INVALID_CONIC_COEF);
	return NULL;
    }

    if (SwitchYZ) {
	MatGenMatRotX1(M_PI / 2.0, Mat);
	MatGenMatScale(1, 1, -1, Mat2);
	MatMultTwo4by4(Mat, Mat2, Mat);
	CagdSrfMatTransform(BzrSrf, Mat);
    }
    if (SwitchXZ) {
	MatGenMatRotY1(-M_PI / 2.0, Mat);
	MatGenMatScale(1, 1, -1, Mat2);
	MatMultTwo4by4(Mat, Mat2, Mat);
	CagdSrfMatTransform(BzrSrf, Mat);
    }
    if (SwitchXY) {
	MatGenMatRotZ1(M_PI / 2.0, Mat);
	MatGenMatScale(1, -1, 1, Mat2);
	MatMultTwo4by4(Mat, Mat2, Mat);
	CagdSrfMatTransform(BzrSrf, Mat);
    }

    if (BzrSrf != NULL) {
        PT_SCALE(Trans, -1.0);
	CagdSrfTransform(BzrSrf, Trans, 1.0);
	CagdSrfMatTransform(BzrSrf, U);

	CAGD_SET_GEOM_TYPE(BzrSrf, CAGD_GEOM_CONIC_SEC);

	return BzrSrf;
    }

    return BzrSrf;
}
