/*****************************************************************************
* Filter to convert IRIT data files to STL layered manufacturing format.     *
******************************************************************************
* (C) Gershon Elber, Technion, Israel Institute of Technology                *
******************************************************************************
* Written by:  Gershon Elber			       Ver 1.0, April 2000   *
*****************************************************************************/

#include <stdio.h>
#include <math.h>
#include <string.h>
#include "irit_sm.h"
#include "iritprsr.h"
#include "allocate.h"
#include "attribut.h"
#include "iritgrap.h"
#include "ip_cnvrt.h"
#include "misc_lib.h"

#define SAME_VRTX_DEF_EPS	1e-4

#ifdef NO_CONCAT_STR
STATIC_DATA char *VersionStr =
	"Irit2Stl		Version 9.0,	Gershon Elber,\n\
	 (C) Copyright 1989-2004 Gershon Elber, Non commercial use only.";
#else
STATIC_DATA char *VersionStr = "Irit2Stl	" IRIT_VERSION ",	Gershon Elber,	"
	__DATE__ ",   " __TIME__ "\n" IRIT_COPYRIGHT ", Non commercial use only.";
#endif /* NO_CONCAT_STR */

STATIC_DATA char
#ifdef IRIT_DOUBLE
    *CtrlStr = "Irit2Stl l%- 4%- r%- F%-PolyOpti|FineNess!d!F E%-VrtxEps!F s%- S%- o%-OutName!s m%- u%- z%- DFiles!*s";
#else
    *CtrlStr = "Irit2Stl l%- 4%- r%- F%-PolyOpti|FineNess!d!f E%-VrtxEps!f s%- S%- o%-OutName!s m%- u%- z%- DFiles!*s";
#endif /* IRIT_DOUBLE */

#define IHT_VERTEX_KEY(Pt)	(Pt[0] * 0.301060 + \
				 Pt[1] * 0.050964 + \
				 Pt[2] * 0.161188)

STATIC_DATA int
    GlblRegularTriang = TRUE,
    GlblMultiObjSplit = 0;
STATIC_DATA RealType
    GlblSameVrtxEps = SAME_VRTX_DEF_EPS;

static int CmpTwoVertices(VoidPtr V1, VoidPtr V2);
static void MakeVerticesIdentical(IPObjectStruct *PObj);
static void DumpDataForStl(IPObjectStruct *PObjects, char *OutFileName);
static void Irit2StlExit(int ExitCode);

/*****************************************************************************
* DESCRIPTION:                                                               M
* Main module of irit2scn - Read command line and do what is needed...	     M
*                                                                            *
* PARAMETERS:                                                                M
*   argc, argv:  Command line.                                               M
*                                                                            *
* RETURN VALUE:                                                              M
*   void                                                                     M
*                                                                            *
* KEYWORDS:                                                                  M
*   main                                                                     M
*****************************************************************************/
void main(int argc, char **argv)
{
    int Error,
	SrfFineNessFlag = FALSE,
	EpsFlag = FALSE,
	VerFlag = FALSE,
	SplitObjFlag = FALSE,
	SplitObjFileFlag = FALSE,
	OutFileFlag = FALSE,
	ForceUnitMat = FALSE,
	NumFiles = 0;
    char
	**FileNames = NULL,
        *OutFileName = NULL;
    IPObjectStruct *PObjects;
    MatrixType CrntViewMat;

#ifdef DEBUG_IRIT_MALLOC
    IritInitTestDynMemory();
#endif /* DEBUG_IRIT_MALLOC */

    if ((Error = GAGetArgs(argc, argv, CtrlStr,
			   &IPFFCState.LinearOnePolyFlag,
			   &IPFFCState.FourPerFlat, &GlblRegularTriang,
			   &SrfFineNessFlag, &IPFFCState.OptimalPolygons,
			   &IPFFCState.FineNess, &EpsFlag, &GlblSameVrtxEps,
			   &SplitObjFlag, &SplitObjFileFlag,
			   &OutFileFlag, &OutFileName, &IPFFCState.Talkative,
			   &ForceUnitMat, &VerFlag,
			   &NumFiles, &FileNames)) != 0) {

	GAPrintErrMsg(Error);
	GAPrintHowTo(CtrlStr);
	Irit2StlExit(1);
    }

    if (VerFlag) {
	fprintf(stderr, "\n%s\n\n", VersionStr);
	GAPrintHowTo(CtrlStr);
	Irit2StlExit(0);
    }

    if (SplitObjFlag && SplitObjFileFlag) {
	fprintf(stderr, "Only one of '-s' and '-S' at a time please, exit.\n");
	GAPrintHowTo(CtrlStr);
	Irit2StlExit(1);
    }
    if (SplitObjFlag)
        GlblMultiObjSplit = 1;
    else if (SplitObjFileFlag)
        GlblMultiObjSplit = 2;
	
    if (!NumFiles) {
	fprintf(stderr, "No data file names were given, exit.\n");
	GAPrintHowTo(CtrlStr);
	Irit2StlExit(1);
    }

    if (IPFFCState.Talkative)
	fprintf(stderr, "%s triangles per flat will be created.\n",
		IPFFCState.FourPerFlat ? "Four" : "Two");

    /* Get the data files: */
    IPSetFlattenObjects(TRUE);
    IPSetPolyListCirc(TRUE);
    if ((PObjects = IPGetDataFiles(FileNames, NumFiles, TRUE, FALSE)) == NULL)
	Irit2StlExit(1);
    PObjects = IPResolveInstances(PObjects);

    if ((!IPWasPrspMat && !IPWasViewMat) || ForceUnitMat) {
	MatGenUnitMat(CrntViewMat);
    }
    else {
        if (IPWasPrspMat)
	    MatMultTwo4by4(CrntViewMat, IPViewMat, IPPrspMat);
	else
	    GEN_COPY(CrntViewMat, IPViewMat, sizeof(MatrixType));

	IPTraverseObjListHierarchy(PObjects, CrntViewMat, IPMapObjectInPlace);
    }

    DumpDataForStl(PObjects, OutFileName);

    Irit2StlExit(0);
}

/*****************************************************************************
* DESCRIPTION:                                                               *
*   Compare two vertices if same or not.                                     *
*                                                                            *
* PARAMETERS:                                                                *
*   V1, V2:   Two vertices to compare.                                       *
*                                                                            *
* RETURN VALUE:                                                              *
*   int:      -1, 0, +1 if V1 less, equal, greater than V2.                  *
*****************************************************************************/
static int CmpTwoVertices(VoidPtr V1, VoidPtr V2)
{
    RealType
	*Coord1 = (RealType *) V1,
	*Coord2 = (RealType *) V2;

    if (PT_APX_EQ(Coord1, Coord2))
	return 0;
    else {
        int i;

        for (i = 0; i < 3; i++) {
	    if (Coord1[i] < Coord2[i])
	        return -1;
	    else if (Coord1[i] > Coord2[i])
	        return 1;
	}

	return 0; /* Should never get here. */
    }
}

/*****************************************************************************
* DESCRIPTION:                                                               *
*   Make sure all epsilon like vertices are made identically the same.       *
*                                                                            *
* PARAMETERS:                                                                *
*   PObj:   Polygonal object to process.                                     *
*                                                                            *
* RETURN VALUE:                                                              *
*   void                                                                     *
*****************************************************************************/
static void MakeVerticesIdentical(IPObjectStruct *PObj)
{
    RealType Min, Max;
    GMBBBboxStruct 
	*BBox = GMBBComputeBboxObject(PObj);
    IritHashTableStruct *IHT;
    IPVertexStruct *PVertex;
    IPPolygonStruct *PPoly;

    Min = MIN(MIN(BBox -> Min[0], BBox -> Min[1]), BBox -> Min[2]);
    Max = MAX(MAX(BBox -> Max[0], BBox -> Max[1]), BBox -> Max[2]);

    /* Create the hash table. */
    IHT = IritHashTableCreate(Min, Max, IRIT_EPS,
			      IPPolyListLen(PObj -> U.Pl));

    /* Insert all vertices to the hash table.  Similar vertices upto      */
    /* the epsilon prescribed in CmpTwoVertices are inserted once only.   */
    for (PPoly = PObj -> U.Pl;
	 PPoly != NULL;
	 PPoly = PPoly -> Pnext) {
        PVertex = PPoly -> PVertex;
	do {
	    IritHashTableInsert(IHT, PVertex -> Coord, CmpTwoVertices,
				IHT_VERTEX_KEY(PVertex -> Coord), FALSE);

	    PVertex = PVertex -> Pnext;
	}
	while (PVertex != NULL && PVertex != PPoly -> PVertex);
    }

    /* Find duplicated entries amd kae them identical. */
    for (PPoly = PObj -> U.Pl;
	 PPoly != NULL;
	 PPoly = PPoly -> Pnext) {
        PVertex = PPoly -> PVertex;
	do {
	    VoidPtr Data;

	    if ((Data = IritHashTableFind(IHT, PVertex -> Coord,
					  CmpTwoVertices,
					  IHT_VERTEX_KEY(PVertex -> Coord)))
								  != NULL &&
		Data != (VoidPtr) PVertex -> Coord) {
	        /* It is an entry that is similar to this one.  Coerce   */
	        /* them both to be identically the same.		 */
		PT_COPY(PVertex -> Coord, Data);
	    }

	    PVertex = PVertex -> Pnext;
	}
	while (PVertex != NULL && PVertex != PPoly -> PVertex);
    }
    
    IritHashTableFree(IHT);
}

/*****************************************************************************
* DESCRIPTION:                                                               *
* Dumps the data for Stl to OutFileName.                                     *
*                                                                            *
* PARAMETERS:                                                                *
*   PObjects:   To dump into file.                                           *
*                                                                            *
* RETURN VALUE:                                                              *
*   void                                                                     *
*****************************************************************************/
static void DumpDataForStl(IPObjectStruct *PObjects, char *OutFileName)
{
    char
	*Name = "irit2stl";
    int i,
        VIndex = 0,
	TotalVertices = 0;
    IPPolygonStruct *PPoly;
    IPObjectStruct *PObj;
    FILE
	*f = NULL;

    if (GlblMultiObjSplit < 2) {
	if (OutFileName != NULL) {
	    if ((f = fopen(OutFileName, "w")) == NULL) {
	        fprintf(stderr, "Failed to open \"%s\".\n", OutFileName);
		Irit2StlExit(2);
	    }
	}
	else
	    f = stdout;
    }

    if (GlblMultiObjSplit == 0)
        fprintf(f, "solid %s\n", Name);

    /* Count how many polygons/vertices we have in this data set and print. */
    for (PObj = PObjects; PObj != NULL; PObj = PObj -> Pnext) {
	IPObjectStruct *PObjReg, *PObjRegTris, *PPObj;

	/* Process freeform geometry into polys. */
	if (IP_IS_FFGEOM_OBJ(PObj))
	    PPObj = IPConvertFreeForm(PObj, &IPFFCState);
	else if (IP_IS_GEOM_OBJ(PObj))
	    PPObj = PObj;
	else
	    continue;

	Name = PObj -> Name;

	if (GlblMultiObjSplit >= 2) {
	    STATIC_DATA int
	        FileCount = 1;
	    char OutFileNameIndex[LINE_LEN_LONG];
	    
	    if (OutFileName != NULL) {
	        char
		    *p = strchr(OutFileName, '.');

		if (p != NULL)
		    *p = 0;
	        sprintf(OutFileNameIndex, "%s%d.stl",
			OutFileName, FileCount++);

	        if ((f = fopen(OutFileNameIndex, "w")) == NULL) {
		    fprintf(stderr, "Failed to open \"%s\".\n",
			    OutFileNameIndex);
		    Irit2StlExit(2);
		}
	    }
	    else
	        f = stdout;

	    if (IPFFCState.Talkative)	
	        fprintf(stderr,
			"Processing object \"%s\" into file \"%s\".\n",
			Name, f == stdout ? "stdout" : OutFileNameIndex);
	}
	else {
	    if (IPFFCState.Talkative)	
	        fprintf(stderr,
			"Processing object \"%s\" into file \"%s\".\n",
			Name, f == stdout ? "stdout" : OutFileName);
	}
	if (GlblMultiObjSplit > 0)
	    fprintf(f, "solid %s\n", Name);

	if (!IP_IS_POLY_OBJ(PPObj)) {
	    fprintf(stderr, "Non polygonal object \"%s\" ignored.\n", Name);
	    continue;
	}

	if (GlblRegularTriang) {
	    PObjReg = GMRegularizePolyModel(PPObj);
	    PObjRegTris = GMConvertPolysToTriangles(PObjReg);
	    IPFreeObject(PObjReg);
	}
	else
	    PObjRegTris = GMConvertPolysToTriangles(PPObj);

	if (PPObj != PObj)
	    IPFreeObject(PPObj);

	MakeVerticesIdentical(PObjRegTris);

	/* Dump the geometry out. */
	for (PPoly = PObjRegTris -> U.Pl;
	     PPoly != NULL;
	     PPoly = PPoly -> Pnext) {
	    IPVertexStruct *V;

	    if (PT_APX_EQ_ZERO_EPS(PPoly -> Plane, IRIT_EPS))
	        fprintf(stderr, "Warning: zero normals detected!\n");

	    fprintf(f, "  facet normal %9.6f %9.6f %9.6f\n    outer loop\n",
		    -PPoly -> Plane[0],
		    -PPoly -> Plane[1],
		    -PPoly -> Plane[2]);
	    for (i = 0, V = PPoly -> PVertex;
		 i < 3 && V != NULL;
		 i++, V = V -> Pnext) {
	        fprintf(f, "      vertex %9.6f %9.6f %9.6f\n",
			V -> Coord[0], V -> Coord[1], V -> Coord[2]);
	    }
	    fprintf(f, "    endloop\n  endfacet\n");
	}

	IPFreeObject(PObjRegTris);

	if (GlblMultiObjSplit > 0)
	    fprintf(f, "endsolid\n\n");
	if (GlblMultiObjSplit >= 2 && f != stdout)
	    fclose(f);
    }

    if (GlblMultiObjSplit == 0)
	fprintf(f, "endsolid\n\n");
    if (GlblMultiObjSplit < 2 && f != stdout)
	fclose(f);
}

/*****************************************************************************
* DESCRIPTION:                                                               *
* Irit2Stl exit routine.						     *
*                                                                            *
* PARAMETERS:                                                                *
*   ExitCode:                                                                *
*                                                                            *
* RETURN VALUE:                                                              *
*   void                                                                     *
*****************************************************************************/
static void Irit2StlExit(int ExitCode)
{
    exit(ExitCode);
}

#ifdef DEBUG

/*****************************************************************************
* DESCRIPTION:                                                               *
*    Dummy function to link at debugging time.                               *
*                                                                            *
* PARAMETERS:                                                                *
*                                                                            *
* RETURN VALUE:                                                              *
*   void                                                                     *
*                                                                            *
* KEYWORDS:                                                                  *
*****************************************************************************/
void DummyLinkCagdDebug(void)
{
    IPDbg();
}

#endif /* DEBUG */
