/*****************************************************************************
* Filter to convert STL data files to IRIT .irt files.			     *
******************************************************************************
* (C) Gershon Elber, Technion, Israel Institute of Technology                *
******************************************************************************
* Written by:  Gershon Elber				Ver 1.0, June 2002   *
*****************************************************************************/

#include <stdio.h>
#include <ctype.h>
#include <math.h>
#include <string.h>
#include "irit_sm.h"
#include "geom_lib.h"
#include "misc_lib.h"

#ifdef __WINNT__
#include <io.h>
#include <fcntl.h>
#endif /* __WINNT__ */

#ifdef NO_CONCAT_STR
STATIC_DATA char *VersionStr =
	"stl2irit		Version 9.0,		Gershon Elber,\n\
	 (C) Copyright 1989-2004 Gershon Elber, Non commercial use only.";
#else
STATIC_DATA char *VersionStr = "stl2irit	" IRIT_VERSION ",	Gershon Elber,	"
	__DATE__ ",   " __TIME__ "\n" IRIT_COPYRIGHT ", Non commercial use only.";
#endif /* NO_CONCAT_STR */

STATIC_DATA char
    *CtrlStr = "stl2irit b%- w%- n%- o%-OutName!s z%- STLFile!*s";

static void EndianSwapFloats(float *RP, int n);
static void EndianSwapLongInts(long int *IP, int n);
static void Stl2IritExit(int ExitCode);

/*****************************************************************************
* DESCRIPTION:                                                               M
* Main module of stl2irit - Read command line and do what is needed...	     M
*                                                                            *
* PARAMETERS:                                                                M
*   argc, argv:  Command line.                                               M
*                                                                            *
* RETURN VALUE:                                                              M
*   void                                                                     M
*                                                                            *
* KEYWORDS:                                                                  M
*   main                                                                     M
*****************************************************************************/
void main(int argc, char **argv)
{
    int i, Error, NumFiles,
	VerFlag = FALSE,
	OutFileFlag = FALSE,
        BinarySTL = FALSE,
        EndianSwap = FALSE,
        NormalFlip = FALSE,
        NumInputPolys = 0,
        NumOutputPolys = 0;
    char *ReadAttr, Line[LINE_LEN_VLONG],
	*OutFileName = NULL,
	**FileNames = NULL;
    FILE *StlFile, *DATFile;

#ifdef DEBUG_IRIT_MALLOC
    IritInitTestDynMemory();
#endif /* DEBUG_IRIT_MALLOC */

    if ((Error = GAGetArgs(argc, argv, CtrlStr,
			   &BinarySTL, &EndianSwap, &NormalFlip,
			   &OutFileFlag, &OutFileName, &VerFlag,
			   &NumFiles, &FileNames)) != 0) {
	GAPrintErrMsg(Error);
	GAPrintHowTo(CtrlStr);
	Stl2IritExit(1);
    }

    if (VerFlag) {
	fprintf(stderr, "\n%s\n\n", VersionStr);
	GAPrintHowTo(CtrlStr);
	Stl2IritExit(0);
    }

    if (!NumFiles) {
	fprintf(stderr, "No data file names were given, exit.\n");
	GAPrintHowTo(CtrlStr);
	Stl2IritExit(1);
    }
    else if (NumFiles > 1) {
	fprintf(stderr, "Cannot handle more than one STL file at a time, exit.\n");
	GAPrintHowTo(CtrlStr);
	Stl2IritExit(1);
    }

#   ifdef __WINNT__
        ReadAttr = BinarySTL ? "rb" : "r";
#   else
	ReadAttr = "r";
#   endif /* __WINNT__ */

    if (strcmp(FileNames[0], "-") == 0) {
	StlFile = stdin;
#	if defined(__WINNT__)
	    if (BinarySTL)
	        _setmode(_fileno(stdin), _O_BINARY);        /* Binary mode. */
#	endif /* __WINNT__ */
    }
    else if ((StlFile = fopen(FileNames[0], ReadAttr)) == NULL) {
	fprintf(stderr, "Cannot open STL file \"%s\", exit.\n", FileNames[0]);
	Stl2IritExit(1);
    }

    if (OutFileName != NULL) {
	if ((DATFile = fopen(OutFileName, "w")) < 0) {
	    fprintf(stderr, "Failed to open \"%s\".\n", OutFileName);
	    Stl2IritExit(2);
	}
    }
    else
        DATFile = stdout;

    /* Skip the header. */
    if (BinarySTL) {
        long int Len;

        for (i = 0; i < 80; i++)   /* 80 bytes in header of binary STL file. */
	    getc(StlFile);

	fread(&Len, sizeof(long int), 1, StlFile);
	if (EndianSwap)
	    EndianSwapLongInts(&Len, 1);
	fprintf(stderr, "%d triangles in STL binary file\n", Len);
    }
    else {
        /* Skip until a first facet is found. */
        while (!strstr(fgets(Line, LINE_LEN_VLONG - 1, StlFile), "facet") &&
	       !feof(StlFile));
    }

    fprintf(DATFile, "[OBJECT STL\n");

    /* Read triangles and dump them immediately. */
    do {
	char Pad[2];
        float Normal[3], Points[3][3];

	NumInputPolys++;

        if (BinarySTL) {
	    fread(Normal, sizeof(float), 3, StlFile);
	    if (EndianSwap)
	        EndianSwapFloats(&Normal[0], 3);
	    fread(Points, sizeof(float), 9, StlFile);
	    if (EndianSwap)
	        EndianSwapFloats(&Points[0][0], 9);
	    fread(Pad, sizeof(char), 2, StlFile);
	}
	else {
	    /* Parse the normal. */
	    sscanf(Line, " facet normal  %f %f %f\n",
		   &Normal[0], &Normal[1], &Normal[2]);

	    /* Sync on the first vertex. */
	    do
	        fgets(Line, LINE_LEN_VLONG - 1, StlFile);
	    while (!feof(StlFile) && !strstr(Line, "vertex"));
	    if (feof(StlFile))
		break;
	    sscanf(Line, " vertex %f %f %f\n",
		   &Points[0][0], &Points[0][1], &Points[0][2]);

	    /* Sync on the second vertex. */
	    fgets(Line, LINE_LEN_VLONG - 1, StlFile);
	    sscanf(Line, " vertex %f %f %f\n",
		   &Points[1][0], &Points[1][1], &Points[1][2]);

	    /* Sync on the third vertex. */
	    fgets(Line, LINE_LEN_VLONG - 1, StlFile);
	    sscanf(Line, " vertex %f %f %f\n",
		   &Points[2][0], &Points[2][1], &Points[2][2]);

	    /* Skip until next facet is found. */
	    while (!strstr(fgets(Line, LINE_LEN_VLONG - 1, StlFile),
			   "facet") &&
		   !strstr(Line, "normal") &&
		   !feof(StlFile));
	}

	if (!feof(StlFile)) {
	    VectorType V1, V2, V;

	    /* Estimate normal for this triangle. */
	    for (i = 0; i < 3; i++) {
	        V1[i] = Points[1][i] - Points[0][i];
		V2[i] = Points[2][i] - Points[1][i];
	    }
	    GMVecCrossProd(V, V1, V2);

	    if (DOT_PROD(V, V) > IRIT_UEPS) {
	        /* Dump the triangle. */
	        fprintf(DATFile, "    [POLYGON 3\n");

		NumOutputPolys++;

	        if (!NormalFlip ||
		    (V[0] * Normal[0] +
		     V[1] * Normal[1] +
		     V[2] * Normal[2] > 0)) {
		    /* Flip the vertices order to flip the normal... */
		    fprintf(DATFile, "\t[%f %f %f]\n",
			    Points[0][0], Points[0][1], Points[0][2]);
		    fprintf(DATFile, "\t[%f %f %f]\n",
			    Points[2][0], Points[2][1], Points[2][2]);
		    fprintf(DATFile, "\t[%f %f %f]\n",
			    Points[1][0], Points[1][1], Points[1][2]);
		}
		else {
		    fprintf(DATFile, "\t[%f %f %f]\n",
			    Points[0][0], Points[0][1], Points[0][2]);
		    fprintf(DATFile, "\t[%f %f %f]\n",
			    Points[1][0], Points[1][1], Points[1][2]);
		    fprintf(DATFile, "\t[%f %f %f]\n",
			    Points[2][0], Points[2][1], Points[2][2]);
		}
		fprintf(DATFile, "    ]\n");
	    }
	}
    }
    while (!feof(StlFile));

    fprintf(DATFile, "]\n");

    fprintf(stderr,
	    "%d triangles processed in STL file, %d dumped out\n",
	    NumInputPolys, NumOutputPolys);

    Stl2IritExit(0);
}

/*****************************************************************************
* DESCRIPTION:                                                               *
*   Swaps a vector of n float type numbers, in place.                        *
*                                                                            *
* PARAMETERS:                                                                *
*   RP:   A pointer to the vector of floats.                                 *
*   n:    Number of entities in RP.                                          *
*                                                                            *
* RETURN VALUE:                                                              *
*   void                                                                     *
*****************************************************************************/
static void EndianSwapFloats(float *RP, int n)
{
    int i, j, k;
    char
	*c = (char *) RP;

    for (i = 0; i < n; i++) {
	for (j = 0, k = sizeof(float) - 1; j < k; j++, k--)
	    SWAP(char, c[j], c[k]);

	c += sizeof(float);
    }
}

/*****************************************************************************
* DESCRIPTION:                                                               *
*   Swaps a vector of n integer type numbers, in place.                      *
*                                                                            *
* PARAMETERS:                                                                *
*   IP:   A pointer to the vector of integers.                               *
*   n:    Number of entities in IP.                                          *
*                                                                            *
* RETURN VALUE:                                                              *
*   void                                                                     *
*****************************************************************************/
static void EndianSwapLongInts(long int *IP, int n)
{
    int i, j, k;
    char
	*c = (char *) IP;

    for (i = 0; i < n; i++) {
	for (j = 0, k = sizeof(long int) - 1; j < k; j++, k--)
	    SWAP(char, c[j], c[k]);

	c += sizeof(long int);
    }
}

/*****************************************************************************
* DESCRIPTION:                                                               *
* Stl2Irit exit routine.						     *
*                                                                            *
* PARAMETERS:                                                                *
*   ExitCode:                                                                *
*                                                                            *
* RETURN VALUE:                                                              *
*   void                                                                     *
*****************************************************************************/
static void Stl2IritExit(int ExitCode)
{
    exit(ExitCode);
}
