/******************************************************************************
* PlyCrvtr.c - principal curvatures for polygonal data sets.		      *
*******************************************************************************
* (C) Gershon Elber, Technion, Israel Institute of Technology                 *
*******************************************************************************
* Written by Tatiana Surazhsky, October 2002                                  *
******************************************************************************/

#include <math.h>
#include "geom_loc.h"

#define BOUND_CRVTR(k)	\
	(fabs( k ) < IRIT_UEPS ? 0.0 \
			       : BOUND(k, -IRIT_INFNTY, IRIT_INFNTY) )

typedef enum GMPlCrvtrlabel {
    GMPlCrvtrUnused = -1,
    GMPlCrvtrUsed = 0,
    GMPlCrvtrWhite = 1,
    GMPlCrvtrBlack = 2
} GMPlCrvtrlabel;

typedef struct GMPlCrvtrMatrix33Struct {
    RealType Data[3][3];
} GMPlCrvtrMatrix33Struct;

typedef struct GMPlCrvtrMatrix3NStruct {
    RealType *Data[3];
    int N;
} GMPlCrvtrMatrix3NStruct;


typedef struct GMPlCrvtrEdgeStruct {
    GMPlCrvtrVertexStruct *Source, *Target;

    struct GMPlCrvtrEdgeStruct *Twin;
    struct GMPlCrvtrFaceStruct *Face;

    /* InFace is an iterator - a pointer to a list entry in facestructure. */
    GMPlCrvtrDListStruct *InFace;
} GMPlCrvtrEdgeStruct;

typedef struct GMPlCrvtrFaceStruct {
    GMPlCrvtrDListStruct *Edges;
} GMPlCrvtrFaceStruct;

typedef struct GMPlCrvtrPVertexStruct {
    GMPlCrvtrDListStruct *MeshVertex; /* Pointer to <VertexStruct*> in Mesh. */
    /* Array of <VertexStruct*> in Mesh to put color for GMPlCrvtrUsed and   */
    GMPlCrvtrDListStruct *VNeighbours;			 /* GMPlCrvtrUnused. */
    VectorType *Neighbours, *PNeighbours;
    RealType K, H;
    int Boundary;
    int *RingSize, NSize;
} GMPlCrvtrPVertexStruct;

#define DOT_PROD_COLUMN(Pt, Mat, Row)	((Pt)[0] * (Mat)[0][Row] + \
				         (Pt)[1] * (Mat)[1][Row] + \
				          (Pt)[2] * (Mat)[2][Row])

STATIC_DATA GMPlCrvtrMeshStruct 
    *GMPlCrvtrMeshInWork = NULL;
STATIC_DATA GMPlCrvtrVertexStruct 
    *GMPlCrvtrLastInitialized = NULL;
STATIC_DATA IPVertexStruct 
    **GMPlCrvtrVertexArray = NULL;

static void GMPlCrvtrSetNormalAux(GMPlCrvtrPVertexStruct *v);
static void GMPlCrvtrInitBasisAux(GMPlCrvtrPVertexStruct *v, 
				  VectorType *V1,
				  VectorType *V2,
				  VectorType *Normal);
static void GMPlCrvtrProjectAux(GMPlCrvtrPVertexStruct *v);
static void GMPlCrvtrInitAttributesAux(GMPlCrvtrVertexStruct *v);
static GMPlCrvtrVertexStruct *GMPlCrvtrNewVertexAux(void);
static void  GMPlCrvtrVertexAttach(GMPlCrvtrVertexStruct *Vertex,
                                   GMPlCrvtrEdgeStruct *e);
static void GMPlCrvtrInitVertexAux(GMPlCrvtrVertexStruct *Vertex, 
				   IPVertexStruct *v);
static GMPlCrvtrEdgeStruct *GMPlCrvtrVertexGetEdgeBetween(
					   const GMPlCrvtrVertexStruct *From,
			                   const GMPlCrvtrVertexStruct *To);
static void GMPlCrvtrVertexAttachAsBoundary(GMPlCrvtrVertexStruct *Vertex, 
                                            GMPlCrvtrEdgeStruct *Edge);
static int GMPlCrvtrVertexIsBoundaryEdge(GMPlCrvtrVertexStruct *Vertex, 
                                          GMPlCrvtrEdgeStruct *Edge);
static void GMPlCrvtrVertexValidateBoundaries(GMPlCrvtrVertexStruct *Vertex);
static int GMPlCrvtrVertexIsManifold(const GMPlCrvtrVertexStruct *Vertex);
static int GMPlCrvtrVertexIsOnBoundary(const GMPlCrvtrVertexStruct *Vertex);
static RealType GMPlCrvtrMatrix33Determinant(const GMPlCrvtrMatrix33Struct *m);
static void GMPlCrvtrMatrix33Inverse(const GMPlCrvtrMatrix33Struct *a, 
                                     GMPlCrvtrMatrix33Struct *To);
static void GMPlCrvtrMatrix33Mult(const GMPlCrvtrMatrix33Struct *a, 
                                  const GMPlCrvtrMatrix33Struct *b, 
                                  GMPlCrvtrMatrix33Struct *Res);
static void GMPlCrvtrMatrix3NInit(GMPlCrvtrMatrix3NStruct *Mat, int N);
static void GMPlCrvtrMatrix3NDelete(GMPlCrvtrMatrix3NStruct *m);
static GMPlCrvtrPVertexStruct *GMPlCrvtrNewPVertexAux(void);
static void GMPlCrvtrAddNeighboursAux(GMPlCrvtrPVertexStruct *v, 
				      GMPlCrvtrVertexStruct *Center);
static void GMPlCrvtrAddOneNeighborAux(GMPlCrvtrPVertexStruct *v,
				       GMPlCrvtrVertexStruct *Curr);
static GMPlCrvtrEdgeStruct *GMPlCrvtrNewEdge(void);
static GMPlCrvtrEdgeStruct *GMPlCrvtrEdgePrevInFace(const GMPlCrvtrEdgeStruct *e);
static GMPlCrvtrEdgeStruct *GMPlCrvtrEdgeNextInFace(const GMPlCrvtrEdgeStruct *e);
static GMPlCrvtrEdgeStruct *GMPlCrvtrEdgePrevOfS(const GMPlCrvtrEdgeStruct *e);
static GMPlCrvtrEdgeStruct *GMPlCrvtrEdgeNextOfS(const GMPlCrvtrEdgeStruct *e);
static GMPlCrvtrFaceStruct *GMPlCrvtrNewFace(void);
static void GMPlCrvtrInsertVertexAux(IPVertexStruct *v);
static void GMPlCrvtrPutVertexIntoMirrorAux(IPVertexStruct *First, IPVertexStruct *Vert);
static void GMPlCrvtrCleanMesh(GMPlCrvtrMeshStruct *Mesh);
static void GMPlCrvtrAttachTwins(GMPlCrvtrEdgeStruct *E1,
				 GMPlCrvtrEdgeStruct *E2);
static 
GMPlCrvtrFaceStruct *GMPlCrvtrMeshAddFace(GMPlCrvtrMeshStruct *Mesh, 
					  const GMPlCrvtrDListStruct *Vertices);
static GMPlCrvtrEdgeStruct *GMPlCrvtrMeshInsertEdge(GMPlCrvtrMeshStruct *Mesh, 
                                                    GMPlCrvtrVertexStruct *s, 
                                                    GMPlCrvtrVertexStruct *t);

static GMPlCrvtrDListStruct *GMPlCrvtrNewList(size_t Size);
static GMPlCrvtrDListStruct *GMPlCrvtrNewElementAux(GMPlCrvtrDListStruct *DList);
static VoidPtr GMPlCrvtrPopFront(GMPlCrvtrDListStruct *DList);
static void GMPlCrvtrCleanList(GMPlCrvtrDListStruct *DList);
static void GMPlCrvtrDeleteList(GMPlCrvtrDListStruct *DList);
static void GMPlCrvtrPushFrontEmpty(GMPlCrvtrDListStruct *DList);
static void GMPlCrvtrPushBackEmpty(GMPlCrvtrDListStruct *DList);
static GMPlCrvtrDListStruct *GMPlCrvtrDeleteAt(GMPlCrvtrDListStruct *Place);
static int GMPlCrvtrIsEmpty(const GMPlCrvtrDListStruct *DList);
static void GMPlCrvtrInitFirstAux(GMPlCrvtrPVertexStruct *v);
static void GMPlCrvtrInitNextAux(GMPlCrvtrPVertexStruct *v, int n);
static void GMPlCrvtrSetNeighboursAux(GMPlCrvtrPVertexStruct *v, 
                                      GMPlCrvtrAlgorithmType AType, 
                                      int NumberOfRings);
static void GMPlCrvtrSetCurvatureParAux(GMPlCrvtrPVertexStruct *v,
					int NumberOfRings);
static RealType GMPlCrvtrGetWeightAux(GMPlCrvtrPVertexStruct *v, 
				      int i, 
				      RealType Sigma);
static RealType GMPlCrvtrGetSigma(GMPlCrvtrPVertexStruct *v);
static RealType GMPlCrvtrFindAlphaAux(GMPlCrvtrPVertexStruct *v, int Id);
static RealType GMPlCrvtrFindAreaAux(GMPlCrvtrPVertexStruct *v, int Id);
static void GMPlCrvtrSetCurvatureGaussAux(GMPlCrvtrPVertexStruct *v);
static void GMPlCrvtrInitPVertexAux(GMPlCrvtrPVertexStruct *v,
                                    GMPlCrvtrDListStruct *InMesh, 
                                    GMPlCrvtrAlgorithmType AType,
                                    int NumberOfRings);
static void GMPlCrvtrCleanMeshNeighbours(GMPlCrvtrPVertexStruct *v);
static void GMPlCrvtrCleanPVertex(GMPlCrvtrPVertexStruct *v);
static void GMPlCrvtrDeletePVertex(GMPlCrvtrPVertexStruct *v);
static void GMPlCrvtrCleanAlg(GMPlCrvtrAlgorithmStruct *Alg);
static void GMPlCrvtrChooseNextVertexStruct(GMPlCrvtrAlgorithmStruct *Alg);
static int GMPlCrvtrCompareVerticesByXYZAux(const void *elem1,
					    const void *elem2);
void GMPlCrvtrForEachVertex(IPPolygonStruct *PolyList, 
                            void (*ActionNew)(IPVertexStruct *Vertex), 
                            void (*ActionOld)(IPVertexStruct *Prev,
					      IPVertexStruct *Current));

/*****************************************************************************
* DESCRIPTION:                                                               *
*  Approximates normal of the Vertex using neighbours coordinates.           *
*                                                                            *
* PARAMETERS:                                                                *
*   v: The Vertex for which the normal vector must be approximated.          *
*                                                                            *
* RETURN VALUE:                                                              *
*   void                                                                     *
*****************************************************************************/
static void GMPlCrvtrSetNormalAux(GMPlCrvtrPVertexStruct *v)
{
    int i;
    VectorType N, Acc, Tmp1, Tmp2,
        *Vert = &(((GMPlCrvtrVertexStruct *) (v -> MeshVertex -> Data))
							    -> Vert -> Coord);

    if (v -> RingSize[0] < 3)
        return;
    PT_SUB(Tmp1, (v -> Neighbours[v -> RingSize[0] - 1]), *Vert);
    PT_SUB(Tmp2, (v -> Neighbours[0]), *Vert);
    CROSS_PROD(N, Tmp1, Tmp2);
    PT_SAFE_NORMALIZE(N);
    PT_COPY(Acc, N);

    for (i = 1; i < v -> RingSize[0]; ++i) {
        VectorType Tmp3;

        PT_SUB(Tmp1, (v -> Neighbours[i-1]), *Vert);
        PT_SUB(Tmp2, (v -> Neighbours[i]), *Vert);
        CROSS_PROD(Tmp3, Tmp1, Tmp2);
        
        if (DOT_PROD(N, Tmp3) < 0)
            PT_SCALE(Tmp3, -1);
        PT_SAFE_NORMALIZE(Tmp3);
        PT_ADD(Acc, Tmp3, Acc);
    }
    PT_SAFE_NORMALIZE(Acc);
    PT_COPY((((GMPlCrvtrVertexStruct *) (v -> MeshVertex -> Data))
							   -> Vert -> Normal), 
	    Acc);
}

/*****************************************************************************
* DESCRIPTION:                                                               *
*  Find local basis for the neighborhood of the given vertex                 *
*  (x,y in tangent plane).                                                   *
*                                                                            *
* PARAMETERS:                                                                *
*   v:  The central Vertex.                                                  *
*   V1: The first basis vector.                                              *
*   V2: The second basis vector.                                             *
*   Normal: The normal vector.                                               *
*                                                                            *
* RETURN VALUE:                                                              *
*   void                                                                     *
*****************************************************************************/
static void GMPlCrvtrInitBasisAux(GMPlCrvtrPVertexStruct *v, 
				  VectorType *V1,
				  VectorType *V2,
				  VectorType *Normal)
{
    VectorType Tmp;

    PT_COPY(*Normal, 
	    (((GMPlCrvtrVertexStruct *)
	      (v -> MeshVertex -> Data)) -> Vert -> Normal));
    PT_SUB(Tmp, v -> Neighbours[0],
	   (((GMPlCrvtrVertexStruct *) (v -> MeshVertex -> Data))
							  -> Vert -> Coord));
    CROSS_PROD(*V1, Tmp, *Normal);
    PT_SAFE_NORMALIZE(*V1);
    
    CROSS_PROD(*V2, *Normal, *V1);
    PT_SAFE_NORMALIZE(*V2);
}

/*****************************************************************************
* DESCRIPTION:                                                               *
*  Project the neighborhood to the local coordinates (x,y in tangent plane). *
*                                                                            *
* PARAMETERS:                                                                *
*   v: The central Vertex.                                                   *
*                                                                            *
* RETURN VALUE:                                                              *
*   void                                                                     *
*****************************************************************************/
static void GMPlCrvtrProjectAux(GMPlCrvtrPVertexStruct *v)
{
    int i = 0;
    VectorType V1, V2, Tmp, N;
    GMPlCrvtrMatrix33Struct T, TL, TMP;

    GMPlCrvtrSetNormalAux(v);    
    if (v -> NSize < 3)
        return;

    GMPlCrvtrInitBasisAux(v, &V1, &V2, &N);

    PT_COPY(T.Data[0], V1);
    PT_COPY(T.Data[1], V2);
    PT_COPY(T.Data[2], N);
    GMPlCrvtrMatrix33Inverse(&T, &TL);
    GMPlCrvtrMatrix33Mult(&TL, &T, &TMP);
    
    for (i = 0; i < v -> NSize; ++i) {
        PT_SUB(Tmp, (v -> Neighbours[i]), 
	       (((GMPlCrvtrVertexStruct *)
		               (v -> MeshVertex -> Data)) -> Vert -> Coord));
        v -> PNeighbours[i][0] = DOT_PROD_COLUMN(Tmp, TL.Data, 0);
        v -> PNeighbours[i][1] = DOT_PROD_COLUMN(Tmp, TL.Data, 1);
        v -> PNeighbours[i][2] = DOT_PROD_COLUMN(Tmp, TL.Data, 2);
    }
}

/*****************************************************************************
* DESCRIPTION:                                                               *
*   Allocate a new Vertex for the curvature estimation.                      *
*                                                                            *
* PARAMETERS:                                                                *
*   None                                                                     *
*                                                                            *
* RETURN VALUE:                                                              *
*   GMPlCrvtrPVertexStruct *:   Newly allocated Vertex                       *
*****************************************************************************/
static GMPlCrvtrPVertexStruct *GMPlCrvtrNewPVertexAux(void)
{
    GMPlCrvtrPVertexStruct 
	*v = (GMPlCrvtrPVertexStruct *)
				   IritMalloc(sizeof(GMPlCrvtrPVertexStruct));

    ZAP_MEM(v, sizeof(GMPlCrvtrPVertexStruct));
    return v;
}

/*****************************************************************************
* DESCRIPTION:                                                               *
*  Adds one more neighbor to the list of neighbours.                         *
*                                                                            *
* PARAMETERS:                                                                *
*   v:      The current Vertex for curvature estimation.                     *
*   Curr:   The Vertex that must be inserted into the list of neighbors.     *
*                                                                            *
* RETURN VALUE:                                                              *
*   void                                                                     *
*****************************************************************************/
static void GMPlCrvtrAddOneNeighborAux(GMPlCrvtrPVertexStruct *v, 
				       GMPlCrvtrVertexStruct *Curr)
{
    if (Curr -> Vert -> Tags == GMPlCrvtrWhite)
	return;
    v -> NSize += 1;
        
    GMPlCrvtrPushBackEmpty(v -> VNeighbours);
    v -> VNeighbours -> Tail -> Pprev -> Data = Curr;
        
    Curr -> Vert -> Tags = GMPlCrvtrWhite;
}

/*****************************************************************************
* DESCRIPTION:                                                               *
*  Initialize the immediate neighbours of the Vertex center.                 *
*                                                                            *
* PARAMETERS:                                                                *
*   v:      the current Vertex for curvature estimation.                     *
*   Center: the Vertex for which the first level neighborhood must be        *
*                                                             initialized.   *
* RETURN VALUE:                                                              *
*   void                                                                     *
*****************************************************************************/
static void GMPlCrvtrAddNeighboursAux(GMPlCrvtrPVertexStruct *v, 
                                      GMPlCrvtrVertexStruct *Center)
{
    GMPlCrvtrEdgeStruct 
	*E, *EN, *Tmp;

    if (Center -> Edges == NULL)
	return;

    E = EN = Tmp = (GMPlCrvtrEdgeStruct *)
	(Center -> Edges -> Tail -> Pnext -> Data);

    do {
	GMPlCrvtrVertexStruct *Curr;
        Tmp = GMPlCrvtrEdgeNextOfS(EN);
        if (Tmp == NULL) {
            Tmp = GMPlCrvtrEdgePrevInFace(EN);
            GMPlCrvtrAddOneNeighborAux(v, Tmp -> Source);

            EN = NULL;
            Curr = E -> Target;
        }
        else {
            EN = Tmp;
            Curr = EN -> Target;
        }
        
        GMPlCrvtrAddOneNeighborAux(v, Curr);
    } while (EN != E && EN);
}

/*****************************************************************************
* DESCRIPTION:                                                               *
*   Initialize first level neighborhood of the vervex v.                     *
*                                                                            *
* PARAMETERS:                                                                *
*   v: The central Vertex.                                                   *
*                                                                            *
* RETURN VALUE:                                                              *
*   void                                                                     *
*****************************************************************************/
static void GMPlCrvtrInitFirstAux(GMPlCrvtrPVertexStruct *v)
{
    GMPlCrvtrVertexStruct 
	*Mv = (GMPlCrvtrVertexStruct *) (v -> MeshVertex -> Data);

    GMPlCrvtrAddNeighboursAux(v, Mv);
    v -> RingSize[0] = v -> NSize;
}

/*****************************************************************************
* DESCRIPTION:                                                               *
*  Initialize the next, number n, level neighborhood of the Vertex v.        *
*                                                                            *
* PARAMETERS:                                                                *
*   v: The central Vertex.                                                   *
*   n: The neighborhood level number.                                        *
*                                                                            *
* RETURN VALUE:                                                              *
*   void                                                                     *
*****************************************************************************/
static void GMPlCrvtrInitNextAux(GMPlCrvtrPVertexStruct *v, int n)
{
    int Count = 0;
    GMPlCrvtrDListStruct 
	*vi = v -> VNeighbours -> Tail -> Pnext;
    
    if (n == 0)
        GMPlCrvtrInitFirstAux(v);
    
    for ( ; 
	 Count < v -> RingSize[n-1] && vi != vi -> Tail; 
	 vi = vi -> Pnext, Count += 1) {
        GMPlCrvtrVertexStruct 
	    *Curr = (GMPlCrvtrVertexStruct *)(vi -> Data);

        GMPlCrvtrAddNeighboursAux(v, Curr);
    }
    v -> RingSize[n] = v -> NSize;
}

/*****************************************************************************
* DESCRIPTION:                                                               *
*  Initialize all neighbors for the curvature estimation.                    *
*                                                                            *
* PARAMETERS:                                                                *
*   v:                  the Vertex where the curvature must be estimated.    *
*   AType:              algorithm type for curvature estimation.             *
*   NumberOfRings:      how many levels of neighborhoods must be Used.       *
*                                                                            *
* RETURN VALUE:                                                              *
*   void                                                                     *
*****************************************************************************/
static void GMPlCrvtrSetNeighboursAux(GMPlCrvtrPVertexStruct *v, 
                                      GMPlCrvtrAlgorithmType AType, 
                                      int NumberOfRings)
{
    int Count = 0, 
	i = 0;
    GMPlCrvtrVertexStruct 
	*MV = (GMPlCrvtrVertexStruct *)(v -> MeshVertex -> Data);
    GMPlCrvtrDListStruct 
	*VI = NULL;

    v -> VNeighbours = GMPlCrvtrNewList(sizeof(GMPlCrvtrVertexStruct));
    v -> RingSize = (int *) IritMalloc(sizeof(int)*NumberOfRings);
    for (i = 0; i < NumberOfRings; ++i)
        v -> RingSize[i] = 0;
    GMPlCrvtrInitFirstAux(v);
    
    for (i = 1; i < NumberOfRings; ++i) 
        GMPlCrvtrInitNextAux(v, i);
    
    v -> Neighbours = (VectorType *)
				   IritMalloc(v -> NSize * sizeof(VectorType));
    for (VI = v -> VNeighbours -> Tail -> Pnext; 
	 Count < v -> NSize; 
	 VI = VI -> Pnext, Count += 1) {
        GMPlCrvtrVertexStruct 
	    *CV = (GMPlCrvtrVertexStruct *) (VI -> Data);

        PT_COPY(v -> Neighbours[Count], CV -> Vert -> Coord);
    }
    if (AType == GMPL_CRVTR_PARABOLOID_FIT) {
        v -> PNeighbours = (VectorType *)
				   IritMalloc(v -> NSize * sizeof(VectorType));
        GMPlCrvtrProjectAux(v);
    }
}

/*****************************************************************************
* DESCRIPTION:                                                               *
*   Principal curvatures estimation using osculating paraboloid fit method.  *
*                                                                            *
* PARAMETERS:                                                                *
*   v:             The Vertex where principal curvatures must be estimated.  *
*   NumberOfRings: How many levels of neighborhood must be Used.             *
*                                                                            *
* RETURN VALUE:                                                              *
*   void                                                                     *
*****************************************************************************/
static void GMPlCrvtrSetCurvatureParAux(GMPlCrvtrPVertexStruct *v,
					int NumberOfRings)
{
    char SDirection[LINE_LEN];
    int i = 0,
	j = 0,
	k = 0;
    RealType D, Sin2Fi, Fi, Sigma, Abc[3], 
	*B = (RealType *) IritMalloc(v -> NSize * sizeof(RealType));
    VectorType V1, V2, N, D1, D2, TMP;
    MatrixType Mat1, Mat1Inv, MatFi;
    GMPlCrvtrMatrix3NStruct A, APInv;
    GMPlCrvtrMatrix33Struct M, InvM;
    GMPlCrvtrVertexStruct 
	*Tmp = NULL;

    if (v -> NSize < 3)
        return;

    GMPlCrvtrMatrix3NInit(&A, v -> NSize);
    GMPlCrvtrMatrix3NInit(&APInv, v -> NSize);

    Sigma = GMPlCrvtrGetSigma(v);
    for (i = 0; i < v -> NSize; ++i) {
        RealType 
	    Wgt = GMPlCrvtrGetWeightAux(v,i, Sigma);

	A.Data[0][i] = v -> PNeighbours[i][0] * v -> PNeighbours[i][0] * Wgt;
	A.Data[1][i] = v -> PNeighbours[i][0] * v -> PNeighbours[i][1] * Wgt;
	A.Data[2][i] = v -> PNeighbours[i][1] * v -> PNeighbours[i][1] * Wgt;
	B[i] = v -> PNeighbours[i][2] *  Wgt;
    }
    
    for (i = 0; i < 3; ++i) {
        for (j = 0; j < 3; ++j) {
	    M.Data[i][j] = 0;
	    for (k = 0; k < A.N; ++k)
	        M.Data[i][j] += A.Data[i][k] * A.Data[j][k];
	}
    }
    GMPlCrvtrMatrix33Inverse(&M, &InvM);

    for (i = 0; i < 3; ++i) {
        for (j = 0; j < APInv.N; ++j) {
	    APInv.Data[i][j] = 0;
	    for (k = 0; k < 3; ++k)
	        APInv.Data[i][j] += InvM.Data[i][k] * A.Data[k][j];
	}
    }

    for (i = 0; i < 3; ++i) {
        Abc[i] = 0;
	for ( k = 0; k < APInv.N; ++k)
	    Abc[i] += APInv.Data[i][k] * B[k];
    }
    v -> K = 4 * Abc[0] * Abc[2] - Abc[1] * Abc[1];
    v -> H = Abc[0] + Abc[2];

    D = (v -> H) * (v -> H) - v -> K;
    if (D < 0)
        D = 0;

    if (D > 0)
        Sin2Fi = Abc[1] / sqrt(D);
    else 
        Sin2Fi = 0;

    if (Sin2Fi > 1.0)
        Sin2Fi = 1.0;
    if (Sin2Fi < -1.0)
        Sin2Fi = -1.0;
    Fi = 0.5 * asin(Sin2Fi);

    GMPlCrvtrInitBasisAux(v, &V1, &V2, &N);

    GMGenRotateMatrix(Mat1, N);
    MatInverseMatrix(Mat1, Mat1Inv);
    MatGenMatRotZ1(-1*Fi, MatFi);

    MatMultPtby4by4(D1, V1, Mat1);
    MatMultPtby4by4(TMP, D1, MatFi);
    MatMultPtby4by4(D1, TMP, Mat1Inv);

    MatMultPtby4by4(D2, V2, Mat1);
    MatMultPtby4by4(TMP, D2, MatFi);
    MatMultPtby4by4(D2, TMP, Mat1Inv);
    PT_SAFE_NORMALIZE(D1);
    PT_SAFE_NORMALIZE(D2);

    PT_CLEAN_NORMALIZE(D1);
    PT_CLEAN_NORMALIZE(D2);

    Tmp = (GMPlCrvtrVertexStruct *) (v -> MeshVertex -> Data);
    AttrSetRealAttrib(&(Tmp -> Vert -> Attr), "K1Curv",
		      BOUND_CRVTR(v -> H + sqrt(D)));
    AttrSetRealAttrib(&(Tmp -> Vert -> Attr), "K2Curv",
		      BOUND_CRVTR(v -> H - sqrt(D)));
    AttrSetRealAttrib(&(Tmp -> Vert -> Attr), "KCurv",
		      BOUND_CRVTR(v -> K));
    AttrSetRealAttrib(&(Tmp -> Vert -> Attr), "HCurv",
		      BOUND_CRVTR(v -> H));

    sprintf(SDirection, "%g, %g, %g", D1[0], D1[1], D1[2]);
    AttrSetStrAttrib(&(Tmp -> Vert -> Attr), "D1", SDirection);
    sprintf(SDirection, "%g, %g, %g", D2[0], D2[1], D2[2]);
    AttrSetStrAttrib(&(Tmp -> Vert -> Attr), "D2", SDirection);

    GMPlCrvtrMatrix3NDelete(&A);
    GMPlCrvtrMatrix3NDelete(&APInv);
    IritFree(B);
}

/*****************************************************************************
* DESCRIPTION:                                                               *
*    Find the angle between the two Edges at Vertex v.                       *
*                                                                            *
* PARAMETERS:                                                                *
*   v:  The central Vertex.                                                  *
*   id: The id number of the first Edge.                                     *
*                                                                            *
* RETURN VALUE:                                                              *
*   RealType: The angle between the Edges.                                   *
*****************************************************************************/
static RealType GMPlCrvtrFindAlphaAux(GMPlCrvtrPVertexStruct *v, int Id)
{
    int Nid = Id + 1;
    RealType Len1, Len2, CosFi;
    VectorType P1, P2, 
	*Vert = &(((GMPlCrvtrVertexStruct *) (v -> MeshVertex -> Data))
							    -> Vert -> Coord);

    if (Nid == v -> RingSize[0])
	Nid = 0;
    PT_SUB(P1, (v -> Neighbours[Id]), *Vert);
    PT_SUB(P2, (v -> Neighbours[Nid]), *Vert);
    Len1 = PT_LENGTH(P1);
    Len2 = PT_LENGTH(P2);

    CosFi = DOT_PROD(P1,P2) / (Len1 * Len2);
    if (CosFi > 1.0)
        CosFi = 1.0;
    if (CosFi < -1.0)
        CosFi = -1.0;
    return acos(CosFi);
}

/*****************************************************************************
* DESCRIPTION:                                                               *
*   Find area of the triangle between the two Edges with common Vertex v.    *
*                                                                            *
* PARAMETERS:                                                                *
*   v:  The central Vertex.                                                  *
*   id: The id number of the first Edge of the triangle.                     *
*                                                                            *
* RETURN VALUE:                                                              *
*   RealType: The area of the triangle,                                      *
*****************************************************************************/
static RealType GMPlCrvtrFindAreaAux(GMPlCrvtrPVertexStruct *v, int Id)
{
    int Nid = Id+1;
    VectorType P1, P2, Area,
        *Vert = &(((GMPlCrvtrVertexStruct *)
		   (v -> MeshVertex -> Data)) -> Vert -> Coord);

    if (Nid == v -> RingSize[0])
        Nid = 0;
    PT_SUB(P1, (v -> Neighbours[Id]), *Vert);
    PT_SUB(P2, (v -> Neighbours[Nid]), *Vert);

    CROSS_PROD(Area, P1, P2);
    return PT_LENGTH(Area) * 0.5;
}

/*****************************************************************************
* DESCRIPTION:                                                               *
*   Principal curvatures estimation using gauss-bonnet scheme.               *
*                                                                            *
* PARAMETERS:                                                                *
*   v:       The Vertex where principal curvatures must be estimated.        *
*                                                                            *
* RETURN VALUE:                                                              *
*   void                                                                     *
*****************************************************************************/
static void GMPlCrvtrSetCurvatureGaussAux(GMPlCrvtrPVertexStruct *v)
{
    int i = 0;
    RealType 
	AccAlpha = 0, 
	Alpha = 0, 
	AccArea = 0, 
	Area = 0;
    GMPlCrvtrVertexStruct 
	*Tmp = NULL;
    
    if (v -> Boundary)
        return;
    
    for (i = 0; i < v -> RingSize[0]; ++i) {
        Alpha = GMPlCrvtrFindAlphaAux(v, i);
        Area = GMPlCrvtrFindAreaAux(v, i);
        
        AccAlpha += Alpha;
        AccArea += Area;
    }
    
    v -> K = 2.0 * M_PI - AccAlpha;
    v -> K *= 3;
    v -> K /= AccArea;
    v -> H = 0;

    Tmp = (GMPlCrvtrVertexStruct *)(v -> MeshVertex -> Data);
    AttrSetRealAttrib(&(Tmp -> Vert -> Attr), "KCurv", BOUND_CRVTR(v -> K));
    AttrSetRealAttrib(&(Tmp -> Vert -> Attr), "HCurv", 0.0);
}

/*****************************************************************************
* DESCRIPTION:                                                               *
*   Initialize Vertex structure for principal curvatures estimation.         * 
*                                                                            *
* PARAMETERS:                                                                *
*   v:      The Vertex structure to initialize.                              *
*   InMesh: The iterator entry in Mesh data structure.                       *
*   AType:  Algorithm type to use for curvature estimation.                  *
*   NumberOfRings: How many levels of neighborhood are to be used.           *
*                                                                            *
* RETURN VALUE:                                                              *
*   void                                                                     *
*****************************************************************************/
static void GMPlCrvtrInitPVertexAux(GMPlCrvtrPVertexStruct *v,
                                    GMPlCrvtrDListStruct *InMesh, 
                                    GMPlCrvtrAlgorithmType AType,
                                    int NumberOfRings)
{
    v -> MeshVertex = InMesh;

    v -> Boundary = GMPlCrvtrVertexIsOnBoundary(
        (const GMPlCrvtrVertexStruct *)(InMesh -> Data));
    if (v -> Boundary) {
        AttrSetIntAttrib(
		&(((GMPlCrvtrVertexStruct *)(InMesh -> Data)) -> Vert -> Attr),
                "boundary", 1);
    }
    else {
	AttrSetIntAttrib(
		&(((GMPlCrvtrVertexStruct *)(InMesh -> Data)) -> Vert -> Attr),
		"boundary", 0);
    }

    GMPlCrvtrSetNeighboursAux(v, AType, NumberOfRings);
    if (AType == GMPL_CRVTR_PARABOLOID_FIT)
        GMPlCrvtrSetCurvatureParAux(v, NumberOfRings);
    if (AType == GMPL_CRVTR_GAUSS_BONNET)
        GMPlCrvtrSetCurvatureGaussAux(v);
    GMPlCrvtrCleanMeshNeighbours(v);
}

/*****************************************************************************
* DESCRIPTION:                                                               *
*    Clean pointers to the neighbors in the Mesh data structure.             *
*                                                                            *
* PARAMETERS:                                                                *
*   v:   The central Vertex.                                                 *
*                                                                            *
* RETURN VALUE:                                                              *
*   void                                                                     *
*****************************************************************************/
static void GMPlCrvtrCleanMeshNeighbours(GMPlCrvtrPVertexStruct *v)
{
    GMPlCrvtrDListStruct 
	*VI = NULL;

    for (VI = v -> VNeighbours -> Tail -> Pnext; 
	 VI != VI -> Tail; 
	 VI = VI -> Pnext) {
        GMPlCrvtrVertexStruct 
            *Curr = (GMPlCrvtrVertexStruct *)(VI -> Data);
        Curr -> Vert -> Tags = GMPlCrvtrUnused;
        VI -> Data = NULL;
    }
    GMPlCrvtrDeleteList(v -> VNeighbours);
    v -> VNeighbours = NULL;
}

/*****************************************************************************
* DESCRIPTION:                                                               *
*   Clean GMPlCrvtrPVertexStruct data structure, initialized by              *
*                                                  GMPlCrvtrInitPVertexAux.  *
*                                                                            *
* PARAMETERS:                                                                *
*   v:   The structure to be cleaned.                                        *
*                                                                            *
* RETURN VALUE:                                                              *
*   void                                                                     *
*****************************************************************************/
static void GMPlCrvtrCleanPVertex(GMPlCrvtrPVertexStruct *v)
{
    if (v -> VNeighbours)
        GMPlCrvtrCleanMeshNeighbours(v);
    IritFree(v -> Neighbours);
    v -> Neighbours = NULL;
    if (v -> PNeighbours) {
        IritFree(v -> PNeighbours);
        v -> PNeighbours = NULL;
    }
    IritFree(v -> RingSize);
    v -> NSize = 0;
    v -> K = v -> H = 0;
    v -> MeshVertex = NULL;
    v -> RingSize = NULL;
}

/*****************************************************************************
* DESCRIPTION:                                                               *
*   Delete the structure allocated by GMPlCrvtrInitPVertexAux.               *
*                                                                            *
* PARAMETERS:                                                                *
*   v:   The structure to deallocate.                                        *
*                                                                            *
* RETURN VALUE:                                                              *
*   void                                                                     *
*****************************************************************************/
static void GMPlCrvtrDeletePVertex(GMPlCrvtrPVertexStruct *v)
{
    GMPlCrvtrCleanPVertex(v);
    IritFree(v);
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Allocate new GMPlCrvtrAlgorithmStruct structure.                         M
*                                                                            *
* PARAMETERS:                                                                M
*   None                                                                     M
*                                                                            *
* RETURN VALUE:                                                              M
*   GMPlCrvtrAlgorithmStruct *:  Newly allocated structure.                  M
*                                                                            *
* SEE ALSO:                                                                  M
*   GMPlCrvtrInitAlg, GMPlCrvtrDeleteAlg                                     M
*                                                                            *
* KEYWORDS:                                                                  M
*   GMPlCrvtrNewAlg                                                          M
*****************************************************************************/
GMPlCrvtrAlgorithmStruct *GMPlCrvtrNewAlg(void)
{
    GMPlCrvtrAlgorithmStruct 
        *Alg = (GMPlCrvtrAlgorithmStruct *)
				 IritMalloc(sizeof(GMPlCrvtrAlgorithmStruct));

    Alg -> AlgVertices = NULL;
    Alg -> CurrVertex = NULL;
    return Alg;
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Initialize the structure for principal curvarures estimation on the      M
*   triangular Mesh data structure.                                          M
*                                                                            *
* PARAMETERS:                                                                M
*   Alg:   The algorithm structure to initialize.                            M
*   Mesh:  The triangular Mesh data.                                         M
*   AType: The method for principal curvatures estimation.                   M
*   NumberOfRings: How many levels of neighborhood must be used.             M
*                                                                            *
* RETURN VALUE:                                                              M
*   void                                                                     M
*                                                                            *
* SEE ALSO:                                                                  M
*    GMPlCrvtrNewAlg                                                         M
*                                                                            *
* KEYWORDS:                                                                  M
*   GMPlCrvtrInitAlg                                                         M
*****************************************************************************/
void GMPlCrvtrInitAlg(GMPlCrvtrAlgorithmStruct *Alg, 
                      GMPlCrvtrMeshStruct *Mesh, 
                      GMPlCrvtrAlgorithmType AType, 
                      int NumberOfRings)
{
    GMPlCrvtrPVertexStruct 
        *NV = GMPlCrvtrNewPVertexAux();

    Alg -> WhichAlg = AType;
    Alg -> NumberOfRings = (NumberOfRings >= 1) ? NumberOfRings : 1;
    Alg -> AlgVertices = GMPlCrvtrNewList(sizeof(GMPlCrvtrPVertexStruct));
    GMPlCrvtrPushBackEmpty(Alg -> AlgVertices);

    Alg -> AlgVertices -> Tail -> Pnext -> Data = NV;
    GMPlCrvtrInitPVertexAux(NV, 
			    Mesh -> Vertices -> Tail -> Pnext, 
			    Alg -> WhichAlg, 
			    Alg -> NumberOfRings);
    Alg -> CurrVertex = Alg -> AlgVertices -> Tail -> Pnext;
}

/*****************************************************************************
* DESCRIPTION:                                                               *
*   Clean GMPlCrvtrAlgorithmStruct structure, initialized by                 *
*                                                          GMPlCrvtrInitAlg. *
*                                                                            *
* PARAMETERS:                                                                *
*   Alg:   The data structure that must be cleaned.                          *
*                                                                            *
* RETURN VALUE:                                                              *
*   void                                                                     *
*****************************************************************************/
static void GMPlCrvtrCleanAlg(GMPlCrvtrAlgorithmStruct *Alg)
{
    GMPlCrvtrDListStruct 
        *VI = NULL;

    Alg -> CurrVertex = NULL;

    for (VI = Alg -> AlgVertices -> Tail -> Pnext; 
	 VI != VI -> Tail; 
	 VI = VI -> Pnext) {
        GMPlCrvtrDeletePVertex((GMPlCrvtrPVertexStruct*)(VI -> Data));
        VI -> Data = NULL;
    }
    GMPlCrvtrDeleteList(Alg -> AlgVertices);
    Alg -> AlgVertices = NULL;
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*  Deallocates the GMPlCrvtrAlgorithmStruct data structure, allocated by     M
*                     GMPlCrvtrNewAlg, and initialized by GMPlCrvtrInitAlg.  M
*                                                                            *
* PARAMETERS:                                                                M
*   Alg:   The data structure that must be deallocated.                      M
*                                                                            *
* RETURN VALUE:                                                              M
*   void                                                                     M
*                                                                            *
* SEE ALSO:                                                                  M
*   GMPlCrvtrNewAlg, GMPlCrvtrInitAlg                                        M
*                                                                            *
* KEYWORDS:                                                                  M
*   GMPlCrvtrDeleteAlg                                                       M
*****************************************************************************/
void GMPlCrvtrDeleteAlg(GMPlCrvtrAlgorithmStruct *Alg)
{
    GMPlCrvtrCleanAlg(Alg);
    IritFree(Alg);
}

/*****************************************************************************
* DESCRIPTION:                                                               *
*  Estimate principal curvatures on the next Vertex of the Mesh.             *
*                                                                            *
* PARAMETERS:                                                                *
*   Alg:   The algorithm data structure.                                     *
*                                                                            *
* RETURN VALUE:                                                              *
*   void                                                                     *
*****************************************************************************/
static void GMPlCrvtrChooseNextVertexStruct(GMPlCrvtrAlgorithmStruct *Alg)
{
    GMPlCrvtrPVertexStruct 
        *Curr = (GMPlCrvtrPVertexStruct *)(Alg -> CurrVertex -> Data);
    GMPlCrvtrPVertexStruct 
        *NV = GMPlCrvtrNewPVertexAux();

    if (Curr -> MeshVertex -> Pnext == Curr -> MeshVertex -> Tail)
        return;

    GMPlCrvtrPushBackEmpty(Alg -> AlgVertices);
    Alg -> AlgVertices -> Tail -> Pprev -> Data = NV;
    GMPlCrvtrInitPVertexAux(NV, 
			    Curr -> MeshVertex -> Pnext, 
			    Alg -> WhichAlg,
			    Alg -> NumberOfRings);
    Alg -> CurrVertex = Alg -> AlgVertices -> Tail -> Pprev;
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Perform principal curvature estimation.                                  M
*                                                                            *
* PARAMETERS:                                                                M
*   Alg:   The algorithm structure.                                          M
*                                                                            *
* RETURN VALUE:                                                              M
*   void                                                                     M
*                                                                            *
* SEE ALSO:                                                                  M
*   GMPlCrvtrGo2NextVertex                                                   M
*                                                                            *
* KEYWORDS:                                                                  M
*   GMPlCrvtrRunAlg                                                          M
*****************************************************************************/
void GMPlCrvtrRunAlg(GMPlCrvtrAlgorithmStruct *Alg)
{
    GMPlCrvtrPVertexStruct 
        *Curr = (GMPlCrvtrPVertexStruct *) (Alg -> CurrVertex -> Data);

    while (Curr -> MeshVertex -> Pnext != Curr -> MeshVertex -> Tail) {
        GMPlCrvtrChooseNextVertexStruct(Alg);
        Curr = (GMPlCrvtrPVertexStruct *)(Alg -> CurrVertex -> Data);
    }
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Choose the next Vertex as current, note that first the curvatures        M
*                                                        must be evaluated.  M
*                                                                            *
* PARAMETERS:                                                                M
*   Alg:   The algorithm data structure.                                     M
*                                                                            *
* RETURN VALUE:                                                              M
*   void                                                                     M
*                                                                            *
* SEE ALSO:                                                                  M
*   GMPlCrvtrRunAlg                                                          M
*                                                                            *
* KEYWORDS:                                                                  M
*   GMPlCrvtrGo2NextVertex                                                   M
*****************************************************************************/
void GMPlCrvtrGo2NextVertex(GMPlCrvtrAlgorithmStruct *Alg)
{
    if (Alg -> CurrVertex -> Pnext == Alg -> AlgVertices -> Tail) 
        Alg -> CurrVertex = Alg -> AlgVertices -> Tail -> Pnext;
    else 
        Alg -> CurrVertex = Alg -> CurrVertex -> Pnext;
}

/*****************************************************************************
* DESCRIPTION:                                                               *
*   Calculate the normalization factor for gaussian filter.                  *
*                                                                            *
* PARAMETERS:                                                                *
*   v:          The Vertex structure.                                        *
*                                                                            *
* RETURN VALUE:                                                              *
*   RealType:   The normalization factor.                                    *
*****************************************************************************/
static RealType GMPlCrvtrGetSigma(GMPlCrvtrPVertexStruct *v)
{
    int i = 0;
    RealType
	XMax = 0;
    const RealType
	FivePerCentLog = log(0.05);

    while (i < v -> NSize) {
        RealType
        x = PT_LENGTH(v -> PNeighbours[i]);
        XMax = MAX(XMax, x);
        ++i;
    }
    return -1 * XMax * XMax / FivePerCentLog;
}


/*****************************************************************************
* DESCRIPTION:                                                               *
*   Estimate the weight of the current Vertex.                               *
*                                                                            *
* PARAMETERS:                                                                *
*   v:       The Vertex structure.	                                     *
*   i:       The id of the Vertex.	                                     *
*   Sigma:   The normalization factor.					     *
*                                                                            *
* RETURN VALUE:                                                              *
*   RealType:   The weight value.                                            *
*****************************************************************************/
static RealType GMPlCrvtrGetWeightAux(GMPlCrvtrPVertexStruct *v, 
				      int i,
				      RealType Sigma)
{
    RealType x, Result;

    if (i < v -> RingSize[0] && v -> NSize == v -> RingSize[0])
        return 1;

    x = PT_LENGTH(v -> PNeighbours[i]);
    Result = exp(-x*x / (Sigma));

    return Result;
}

/*****************************************************************************
* DESCRIPTION:                                                               *
*   Initialize 3xN matrix structure.                                         *
*                                                                            *
* PARAMETERS:                                                                *
*   Mat:  The structure to be initialized.                                   *
*   N:    The second dimension of the matrix.                                *
*                                                                            *
* RETURN VALUE:                                                              *
*   void                                                                     *
*****************************************************************************/
static void GMPlCrvtrMatrix3NInit(GMPlCrvtrMatrix3NStruct *Mat, int N)
{
    int i = 0,
	j = 0;

    Mat -> N = N;
    for (i = 0; i < 3; ++i)
        Mat -> Data[i] = (RealType*)IritMalloc(sizeof(RealType) * N);
    for (i = 0; i < 3; ++i)
        ZAP_MEM(Mat -> Data[i], sizeof(RealType) * N);
}

/*****************************************************************************
* DESCRIPTION:                                                               *
*   Multiplication of the two 3x3 matrices.                                  *
*                                                                            *
* PARAMETERS:                                                                *
*   a:     The first matrix.                                                 *
*   b:     The second matrix.                                                *
*   Res:   The result of the multiplication.                                 *
*                                                                            *
* RETURN VALUE:                                                              *
*   void                                                                     *
*****************************************************************************/
static void GMPlCrvtrMatrix33Mult(const GMPlCrvtrMatrix33Struct *a, 
                                  const GMPlCrvtrMatrix33Struct *b, 
                                  GMPlCrvtrMatrix33Struct *Res)
{
    int i = 0, 
	j = 0;

    for (i = 0; i < 3; ++i)
        for (j = 0; j < 3; ++j)
            Res -> Data[i][j] = a -> Data[i][0] * b -> Data[0][j] + 
			        a -> Data[i][1] * b -> Data[1][j] + 
			        a -> Data[i][2] * b -> Data[2][j];
}

/*****************************************************************************
* DESCRIPTION:                                                               *
*   Deallocate 3xN matrix, allocated by GMPlCrvtrMatrix3NInit.               *
*                                                                            *
* PARAMETERS:                                                                *
*   m:   The matrix structure.                                               *
*                                                                            *
* RETURN VALUE:                                                              *
*   void                                                                     *
*****************************************************************************/
static void GMPlCrvtrMatrix3NDelete(GMPlCrvtrMatrix3NStruct *m)
{
    int i = 0;

    for (i = 0; i < 3; ++i) {
        IritFree(m -> Data[i]);
        m -> Data[i] = NULL;
    }
}

/*****************************************************************************
* DESCRIPTION:                                                               *
*   Is the given Vertex located on boundary of the Mesh.                     *
*                                                                            *
* PARAMETERS:                                                                *
*   Vertex:   The Vertex to be checked.                                      *
*                                                                            *
* RETURN VALUE:                                                              *
*   int:   TRUE if the Vertex is boundary, FALSE otherwise.                  *
*****************************************************************************/
static int GMPlCrvtrVertexIsOnBoundary(const GMPlCrvtrVertexStruct *Vertex)
{
    return !(GMPlCrvtrVertexIsManifold(Vertex));
}

/*****************************************************************************
* DESCRIPTION:                                                               *
*   Is the given Vertex two-manifold.                                        *
*                                                                            *
* PARAMETERS:                                                                *
*   Vertex:  The Vertex to be checked.                                       *
*                                                                            *
* RETURN VALUE:                                                              *
*   int:   TRUE if two-manifold, FALSE if the Vertex is boundary.            *
*****************************************************************************/
static int GMPlCrvtrVertexIsManifold(const GMPlCrvtrVertexStruct *Vertex)
{
    GMPlCrvtrEdgeStruct 
        *Front = NULL;
    
    if (Vertex -> Edges == NULL || GMPlCrvtrIsEmpty(Vertex -> Edges))
        return FALSE;
    Front = (GMPlCrvtrEdgeStruct *)(Vertex -> Edges -> Tail -> Pnext -> Data);
    if (Vertex -> Edges -> Tail -> Pnext != Vertex -> Edges -> Tail -> Pprev ||
        Front -> Twin == NULL) 
        return FALSE;
    return TRUE;
}

/*****************************************************************************
* DESCRIPTION:                                                               *
*    Allocate new GMPlCrvtrDListStruct structure.                            *
*                                                                            *
* PARAMETERS:                                                                *
*   Size:   Size of the data slot.                                           *
*                                                                            *
* RETURN VALUE:                                                              *
*   GMPlCrvtrDListStruct *:   The newly allocated structure.                 *
*****************************************************************************/
static GMPlCrvtrDListStruct *GMPlCrvtrNewList(size_t Size)
{
    GMPlCrvtrDListStruct 
        *NewListV = (GMPlCrvtrDListStruct *)
				      IritMalloc(sizeof(GMPlCrvtrDListStruct));

    NewListV -> ElSize = Size;
    NewListV -> Data = NULL;
    NewListV -> Tail = NewListV;
    NewListV -> Pnext = NewListV;
    NewListV -> Pprev = NewListV;
    return NewListV;
}

/*****************************************************************************
* DESCRIPTION:                                                               *
*   Deallocate the GMPlCrvtrDListStruct structure, allocated by              *
*                                                         GMPlCrvtrNewList.  *
*                                                                            *
* PARAMETERS:                                                                *
*   DList:   The structure to deallocate.                                    *
*                                                                            *
* RETURN VALUE:                                                              *
*   void                                                                     *
*****************************************************************************/
static void GMPlCrvtrDeleteList(GMPlCrvtrDListStruct *DList)
{
    GMPlCrvtrCleanList(DList);
    if (DList -> Data)
        IritFree(DList -> Data);
    IritFree(DList);
}

/*****************************************************************************
* DESCRIPTION:                                                               *
*   Is the given list empty?                                                 *
*                                                                            *
* PARAMETERS:                                                                *
*   DList:   The list to check.                                              *
*                                                                            *
* RETURN VALUE:                                                              *
*   int:   TRUE if the list is empty, FALSE otherwise.                       *
*****************************************************************************/
static int GMPlCrvtrIsEmpty(const GMPlCrvtrDListStruct *DList)
{
    return (DList -> Tail -> Pnext == DList -> Tail);
}

/*****************************************************************************
* DESCRIPTION:                                                               *
*   Clean the GMPlCrvtrDListStruct structure.                                *
*                                                                            *
* PARAMETERS:                                                                *
*   DList:   The list that must be cleaned.                                  *
*                                                                            *
* RETURN VALUE:                                                              *
*   void                                                                     *
*****************************************************************************/
static void GMPlCrvtrCleanList(GMPlCrvtrDListStruct *DList)
{
    while (DList -> Tail != DList -> Pnext) {
        VoidPtr 
            El = GMPlCrvtrPopFront(DList);

        if (El)
            IritFree(El);
    }
}

/*****************************************************************************
* DESCRIPTION:                                                               *
*    Pop front element of the list.                                          *
*                                                                            *
* PARAMETERS:                                                                *
*   DList:   The list for the operation.                                     *
*                                                                            *
* RETURN VALUE:                                                              *
*   VoidPtr:   The data of the "poped" element.                              *
*****************************************************************************/
static VoidPtr GMPlCrvtrPopFront(GMPlCrvtrDListStruct *DList)
{
    GMPlCrvtrDListStruct 
        *Front = DList -> Tail -> Pnext;
    VoidPtr 
        El = Front -> Data;

    DList -> Tail -> Pnext = Front -> Pnext;
    DList -> Tail -> Pnext -> Pprev = DList -> Tail;
    Front -> Pnext = NULL;
    Front -> Pprev = NULL;
    Front -> Tail = NULL;

    Front -> Data = NULL;

    IritFree(Front);

    return El;
}

/*****************************************************************************
* DESCRIPTION:                                                               *
*   Next Edge in the star (incident to the same source).                     *
*                                                                            *
* PARAMETERS:                                                                *
*   e:   The current Edge.                                                   *
*                                                                            *
* RETURN VALUE:                                                              *
*   GMPlCrvtrEdgeStruct *:   The next Edge.                                  *
*****************************************************************************/
static GMPlCrvtrEdgeStruct *GMPlCrvtrEdgeNextOfS(const GMPlCrvtrEdgeStruct *e)
{
    GMPlCrvtrEdgeStruct 
        *p = GMPlCrvtrEdgePrevInFace(e);
    
    if (p)
        return p -> Twin;
    else
        return NULL;
}

/*****************************************************************************
* DESCRIPTION:                                                               *
*   Get the previous Edge in the same face.                                  *
*                                                                            *
* PARAMETERS:                                                                *
*   e:   The current Edge.                                                   *
*                                                                            *
* RETURN VALUE:                                                              *
*   GMPlCrvtrEdgeStruct *:  The next Edge.                                   *
*****************************************************************************/
static GMPlCrvtrEdgeStruct *GMPlCrvtrEdgePrevInFace(const GMPlCrvtrEdgeStruct *e)
{
    if (!e)
        return NULL;
    if (e -> InFace -> Pprev == e -> InFace -> Tail)
        return (GMPlCrvtrEdgeStruct *) (e -> InFace -> Tail -> Pprev -> Data);
    return (GMPlCrvtrEdgeStruct *) (e -> InFace -> Pprev -> Data);
}

/*****************************************************************************
* DESCRIPTION:                                                               *
*   Find an inverse for 3x3 matrix.                                          *
*                                                                            *
* PARAMETERS:                                                                *
*   a:   The matrix.                                                         *
*   To:  The result (inverse of "a").                                        *
*                                                                            *
* RETURN VALUE:                                                              *
*   void                                                                     *
*****************************************************************************/
static void GMPlCrvtrMatrix33Inverse(const GMPlCrvtrMatrix33Struct *a, 
                                     GMPlCrvtrMatrix33Struct *To)
{
    RealType 
        Det = GMPlCrvtrMatrix33Determinant(a);

    if (fabs(Det) <= 0)
        return;
    To -> Data[0][0] = (a -> Data[1][1] * a -> Data[2][2] -
			a -> Data[1][2] * a -> Data[2][1]) / Det;
    To -> Data[0][1] = (a -> Data[0][2] * a -> Data[2][1] -
			a -> Data[0][1] * a -> Data[2][2]) / Det;
    To -> Data[0][2] = (a -> Data[0][1] * a -> Data[1][2] -
			a -> Data[0][2] * a -> Data[1][1]) / Det;
    To -> Data[1][0] = (a -> Data[1][2] * a -> Data[2][0] -
			a -> Data[1][0] * a -> Data[2][2]) / Det;
    To -> Data[1][1] = (a -> Data[0][0] * a -> Data[2][2] -
			a -> Data[0][2] * a -> Data[2][0]) / Det;
    To -> Data[1][2] = (a -> Data[0][2] * a -> Data[1][0] -
			a -> Data[0][0] * a -> Data[1][2]) / Det;
    To -> Data[2][0] = (a -> Data[1][0] * a -> Data[2][1] -
			a -> Data[1][1] * a -> Data[2][0]) / Det;
    To -> Data[2][1] = (a -> Data[0][1] * a -> Data[2][0] -
			a -> Data[0][0] * a -> Data[2][1]) / Det;
    To -> Data[2][2] = (a -> Data[0][0] * a -> Data[1][1] -
			a -> Data[0][1] * a -> Data[1][0]) / Det;
}

/*****************************************************************************
* DESCRIPTION:                                                               *
*   Find a determinant of the 3x3 matrix.                                    *
*                                                                            *
* PARAMETERS:                                                                *
*   m:   The 3x3 matrix.                                                     *
*                                                                            *
* RETURN VALUE:                                                              *
*   RealType:   The matrix's determinant.                                    *
*****************************************************************************/
static RealType GMPlCrvtrMatrix33Determinant(const GMPlCrvtrMatrix33Struct *m)
{
    return (m -> Data[0][0] * m -> Data[1][1] * m -> Data[2][2] + 
	    m -> Data[0][1] * m -> Data[1][2] * m -> Data[2][0] +
	    m -> Data[0][2] * m -> Data[1][0] * m -> Data[2][1] -
	    m -> Data[0][2] * m -> Data[1][1] * m -> Data[2][0] -
	    m -> Data[0][1] * m -> Data[1][0] * m -> Data[2][2] -
	    m -> Data[0][0] * m -> Data[1][2] * m -> Data[2][1]);
}

/*****************************************************************************
* DESCRIPTION:                                                               *
*   Push the new empty entry into the end of the list.                       *
*                                                                            *
* PARAMETERS:                                                                *
*   DList:   The list structure.                                             *
*                                                                            *
* RETURN VALUE:                                                              *
*   void                                                                     *
*****************************************************************************/
static void GMPlCrvtrPushBackEmpty(GMPlCrvtrDListStruct *DList)
{
    GMPlCrvtrDListStruct 
        *p = DList -> Tail -> Pprev;

    DList -> Tail -> Pprev = GMPlCrvtrNewElementAux(DList);
    DList -> Tail -> Pprev -> Pnext = DList -> Tail;
    DList -> Tail -> Pprev -> Pprev = p;
    p -> Pnext = DList -> Tail -> Pprev;
}

/*****************************************************************************
* DESCRIPTION:                                                               *
*   Allocate a new entry for the list.                                       *
*                                                                            *
* PARAMETERS:                                                                *
*   DList:   The list data structure.                                        *
*                                                                            *
* RETURN VALUE:                                                              *
*   GMPlCrvtrDListStruct *:   The newly allocated entry.                     *
*****************************************************************************/
static GMPlCrvtrDListStruct *GMPlCrvtrNewElementAux(GMPlCrvtrDListStruct *DList)
{
    GMPlCrvtrDListStruct 
        *El = GMPlCrvtrNewList(DList -> ElSize);

    El -> Tail = DList -> Tail;
    El -> Pnext = NULL;
    El -> Pprev = NULL;
    return El;
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Deallocate the Mesh data structure.                                      M
*                                                                            *
* PARAMETERS:                                                                M
*   Mesh:   The Mesh data structure that must be deallocated.                M
*                                                                            *
* RETURN VALUE:                                                              M
*   void                                                                     M
*                                                                            *
* SEE ALSO:                                                                  M
*   GMPlCrvtrNemMesh                                                         M
*                                                                            *
* KEYWORDS:                                                                  M
*   GMPlCrvtrDeleteMesh                                                      M
*****************************************************************************/
void GMPlCrvtrDeleteMesh(GMPlCrvtrMeshStruct *Mesh)
{
    GMPlCrvtrCleanMesh(Mesh);
    IritFree(Mesh);
}

/*****************************************************************************
* DESCRIPTION:                                                               *
*   Clean the Mesh data structure.                                           *
*                                                                            *
* PARAMETERS:                                                                *
*   Mesh:   To be cleaned.                                                   *
*                                                                            *
* RETURN VALUE:                                                              *
*   void                                                                     *
*****************************************************************************/
static void GMPlCrvtrCleanMesh(GMPlCrvtrMeshStruct *Mesh)
{
    GMPlCrvtrDListStruct *It;

    Mesh -> NumOfFaces = 0;
    Mesh -> NumOfEdges = 0;
    Mesh -> NumOfVertices = 0;

    GMPlCrvtrDeleteList(Mesh -> Edges);

    for (It = Mesh -> Faces -> Tail -> Pnext; 
	 It != Mesh -> Faces -> Tail; 
	 It = It -> Pnext) {
	GMPlCrvtrFaceStruct
	    *Face = (GMPlCrvtrFaceStruct *) (It -> Data);
	GMPlCrvtrDListStruct *Ed,
	    *EdLast = Face -> Edges -> Tail;

	for (Ed = Face -> Edges -> Tail -> Pnext;
	     Ed != EdLast;
	     ) {
	    GMPlCrvtrDListStruct
		*EdNext = Ed -> Pnext;

	    IritFree(Ed);
	    Ed = EdNext;
	}
	IritFree(Ed);
    }
    GMPlCrvtrDeleteList(Mesh -> Faces);

    for (It = Mesh -> Vertices -> Tail -> Pnext; 
	 It != Mesh -> Vertices -> Tail; 
	 It = It -> Pnext) {
	GMPlCrvtrVertexStruct
	    *Vert = (GMPlCrvtrVertexStruct *) (It -> Data);
	GMPlCrvtrDListStruct *Ed,
	    *EdLast = Vert -> Edges -> Tail;

	for (Ed = Vert -> Edges -> Tail -> Pnext;
	     Ed != EdLast;
	     ) {
	    GMPlCrvtrDListStruct
		*EdNext = Ed -> Pnext;

	    IritFree(Ed);
	    Ed = EdNext;
	}
	IritFree(Ed);

	IPFreeVertex(Vert -> Vert);
	Vert -> Vert = NULL;
    }
    GMPlCrvtrDeleteList(Mesh -> Vertices);

    Mesh -> Faces = NULL;
    Mesh -> Edges = NULL;
    Mesh -> Vertices = NULL;
}


/*****************************************************************************
* DESCRIPTION:                                                               *
*   Add a new face, consisting of the given list of Vertices to the Mesh.    *
*                                                                            *
* PARAMETERS:                                                                *
*   Mesh:      The Mesh structure.                                           *
*   Vertices:  The list of face Vertices.                                    *
*                                                                            *
* RETURN VALUE:                                                              *
*   GMPlCrvtrFaceStruct *:  The pointer to the new face structure.           *
*****************************************************************************/
static GMPlCrvtrFaceStruct *GMPlCrvtrMeshAddFace(GMPlCrvtrMeshStruct *Mesh, 
                                                 const GMPlCrvtrDListStruct *Vertices)
{
    GMPlCrvtrFaceStruct 
        *Face = GMPlCrvtrNewFace();
    GMPlCrvtrVertexStruct 
        *Curr = NULL,
	*Next = NULL;
    GMPlCrvtrEdgeStruct 
        *E = NULL,
	*Twin = NULL,
	*Aux = NULL;
    GMPlCrvtrDListStruct 
        *Vi = NULL,
	*EI = NULL;

    GMPlCrvtrPushBackEmpty(Mesh -> Faces);
    Mesh -> Faces -> Tail -> Pprev -> Data = Face;

    Face -> Edges = GMPlCrvtrNewList(sizeof(GMPlCrvtrEdgeStruct));

    for (Vi = Vertices -> Tail -> Pnext; 
	 Vi != Vertices -> Tail; 
	 Vi = Vi -> Pnext) {
        Curr = (GMPlCrvtrVertexStruct *)(Vi -> Data);

        if (Vi -> Pnext == Vi -> Tail) 
            Next = (GMPlCrvtrVertexStruct *)(Vi -> Tail -> Pnext -> Data);
        else 
            Next = (GMPlCrvtrVertexStruct *)(Vi -> Pnext -> Data);

        if (Next == NULL || Curr == NULL)
            return NULL;

        E = GMPlCrvtrVertexGetEdgeBetween(Curr, Next);

        if (E != NULL)
            return NULL;
        E = GMPlCrvtrMeshInsertEdge(Mesh, Curr, Next);
        if (E == NULL)
            return NULL;
        GMPlCrvtrPushBackEmpty(Face -> Edges);
        Face -> Edges -> Tail -> Pprev -> Data = E;
        E -> InFace = Face -> Edges -> Tail -> Pprev;
        E -> Face = Face;
    }

    EI = Face -> Edges -> Tail -> Pnext;
    do {
        E = (GMPlCrvtrEdgeStruct *) (EI -> Data);
        GMPlCrvtrVertexAttach(E -> Source, E);

        Twin = GMPlCrvtrVertexGetEdgeBetween(E -> Target, E -> Source);
        if (Twin != NULL)
            GMPlCrvtrAttachTwins(E, Twin);
        EI = EI -> Pnext;
    } while (EI != Face -> Edges -> Tail);

    return Face;
}

/*****************************************************************************
* DESCRIPTION:                                                               *
*   Initialize the attributes of the Vertex.                                 *
*                                                                            *
* PARAMETERS:                                                                *
*   v:   The Vertex with the attributes.                                     *
*                                                                            *
* RETURN VALUE:                                                              *
*   void                                                                     *
*****************************************************************************/
static void GMPlCrvtrInitAttributesAux(GMPlCrvtrVertexStruct *v)
{
    v -> Vert -> Tags = GMPlCrvtrUnused;
    AttrSetRealAttrib(&(v -> Vert -> Attr), "K1Curv", 0.0);
    AttrSetRealAttrib(&(v -> Vert -> Attr), "K2Curv", 0.0);
    AttrSetRealAttrib(&(v -> Vert -> Attr), "KCurv", 0.0);
    AttrSetRealAttrib(&(v -> Vert -> Attr), "HCurv", 0.0);
    AttrSetStrAttrib(&(v -> Vert -> Attr), "D1Curv", "0, 0, 0");
    AttrSetStrAttrib(&(v -> Vert -> Attr), "D2Curv", "0, 0, 0");

    AttrSetIntAttrib(&(v -> Vert -> Attr), "boundary", 0);
}

/*****************************************************************************
* DESCRIPTION:                                                               *
*   Find an Edge between the two given Vertices.                             *
*                                                                            *
* PARAMETERS:                                                                *
*   From: the source Vertex                                                  *
*   To:   the target Vertex                                                  *
*                                                                            *
* RETURN VALUE:                                                              *
*   GMPlCrvtrEdgeStruct *:  Pointer to the Edge if exists, NULL otherwise.   *
*****************************************************************************/
static GMPlCrvtrEdgeStruct *GMPlCrvtrVertexGetEdgeBetween(
                                            const GMPlCrvtrVertexStruct  *From,
					    const GMPlCrvtrVertexStruct *To)
{
    GMPlCrvtrDListStruct 
        *It = NULL;

    if (From == To || From -> Edges == NULL)
        return NULL;
    for (It = From -> Edges -> Tail -> Pnext; 
	 It  != From -> Edges -> Tail; 
	 It = It -> Pnext) {
        GMPlCrvtrEdgeStruct
            *E = (GMPlCrvtrEdgeStruct *) (It -> Data),
            *EI = NULL;

        /* Check for an invalid Edge. */
        if (E == NULL)
            continue;
        if (E -> Target == To)
            return E;
        for (EI = GMPlCrvtrEdgeNextOfS(E); 
	     EI && EI != E; 
	     EI = GMPlCrvtrEdgeNextOfS(EI)) {
            if (E -> Target == To)
                return E;
        }
    }
    return NULL;
}

/*****************************************************************************
* DESCRIPTION:                                                               *
*   Insert a new Edge into the Mesh.                                         *
*                                                                            *
* PARAMETERS:                                                                *
*   Mesh:  The Mesh structure.                                               *
*   s:     The source Vertex.                                                *
*   t:     The target Vertex.                                                *
*                                                                            *
* RETURN VALUE:                                                              *
*   GMPlCrvtrEdgeStruct *:  The pointer to the new Edge.                     *
*****************************************************************************/
static GMPlCrvtrEdgeStruct *GMPlCrvtrMeshInsertEdge(GMPlCrvtrMeshStruct *Mesh, 
                                                    GMPlCrvtrVertexStruct *s, 
                                                    GMPlCrvtrVertexStruct *t)
{
    GMPlCrvtrEdgeStruct 
        *e = NULL;

    if (s == NULL || t == NULL)
        return NULL;
    
    e = GMPlCrvtrNewEdge();
    ++Mesh -> NumOfEdges;
    GMPlCrvtrPushBackEmpty(Mesh -> Edges);
    Mesh -> Edges -> Tail -> Pprev -> Data = e;
    e -> Source = s;
    e -> Target = t;
    return e;
}


/*****************************************************************************
* DESCRIPTION:                                                               *
*   Allocates a new vertex structure.                                        *
*                                                                            *
* PARAMETERS:                                                                *
*   None                                                                     *
*                                                                            *
* RETURN VALUE:                                                              *
*   GMPlCrvtrVertexStruct *:  The pointer to the newly allocated vertex.     *
*****************************************************************************/
static GMPlCrvtrVertexStruct *GMPlCrvtrNewVertexAux(void)
{
    GMPlCrvtrVertexStruct 
        *v = (GMPlCrvtrVertexStruct *)
				    IritMalloc(sizeof(GMPlCrvtrVertexStruct));

    v -> Vert = IPAllocVertex2(NULL);
    ZAP_MEM(v -> Vert, sizeof(IPVertexStruct));

    v -> Edges = NULL;
    return v;
}

/*****************************************************************************
* DESCRIPTION:                                                               *
*   Attach the given Edges to each other as twin Edges.                      *
*                                                                            *
* PARAMETERS:                                                                *
*   E1:  The first Edge.                                                     *
*   E2:  The second Edge.                                                    *
*                                                                            *
* RETURN VALUE:                                                              *
*   void                                                                     *
*****************************************************************************/
static void GMPlCrvtrAttachTwins(GMPlCrvtrEdgeStruct *E1,
				 GMPlCrvtrEdgeStruct *E2)
{
    if (E1 == NULL || E2 == NULL || E1 -> Twin || E2 -> Twin)
        return;
    if (E1 -> Source != E2 -> Target || E1 -> Target != E2 -> Source)
        return;
    E1 -> Twin = E2;
    E2 -> Twin = E1;
    
    GMPlCrvtrVertexAttach(E1 -> Source, E1);
    GMPlCrvtrVertexAttach(E2 -> Source, E2);
}

/*****************************************************************************
* DESCRIPTION:                                                               *
*   Attach the given Edge to the Vertex, if the Vertex coinsides with the    *
*                                                        source of the Edge. *
*                                                                            *
* PARAMETERS:                                                                *
*   Vertex:  The Vertex where the Edge must be attached.                     *
*   e:       The Edge that must be attached to the Vertex.                   *
*                                                                            *
* RETURN VALUE:                                                              *
*   void                                                                     *
*****************************************************************************/
static void  GMPlCrvtrVertexAttach(GMPlCrvtrVertexStruct *Vertex,
                                   GMPlCrvtrEdgeStruct *e)
{
    if (e == NULL || e -> Source != Vertex)
        return;

    GMPlCrvtrVertexValidateBoundaries(Vertex);
    GMPlCrvtrVertexAttachAsBoundary(Vertex, e);
}

/*****************************************************************************
* DESCRIPTION:                                                               *
*   Attach boundary Edge to the vertex.                                      *
*                                                                            *
* PARAMETERS:                                                                *
*   Vertex:  The Vertex, where the Edge must be attached.                    *
*   Edge:    The boundary Edge.                                              *
*                                                                            *
* RETURN VALUE:                                                              *
*   void                                                                     *
*****************************************************************************/
static void GMPlCrvtrVertexAttachAsBoundary(GMPlCrvtrVertexStruct *Vertex, 
                                            GMPlCrvtrEdgeStruct *Edge)
{
    GMPlCrvtrEdgeStruct 
        *e = Edge;

    Edge -> Source = Vertex;
    while (e && e -> Twin != NULL) {
        e = GMPlCrvtrEdgePrevOfS(e);
        if (Edge == e) /* stop if Vertex is manifold*/
            break;
    }
    if (e -> Twin == NULL && !GMPlCrvtrVertexIsBoundaryEdge(Vertex, e)) {
        if (Vertex -> Edges == NULL)
            Vertex -> Edges = GMPlCrvtrNewList(sizeof(GMPlCrvtrEdgeStruct));
        GMPlCrvtrPushFrontEmpty(Vertex -> Edges);
        Vertex -> Edges -> Tail -> Pnext -> Data = e;
    }
}

/*****************************************************************************
* DESCRIPTION:                                                               *
*   Find the previous Edge among the Edges, having the common source Vertex. *
*                                                                            *
* PARAMETERS:                                                                *
*   e:  Current Edge.                                                        *
*                                                                            *
* RETURN VALUE:                                                              *
*   GMPlCrvtrEdgeStruct *:   The pointer to the previous Edge.               *
*****************************************************************************/
static GMPlCrvtrEdgeStruct *GMPlCrvtrEdgePrevOfS(const GMPlCrvtrEdgeStruct *e)
{
    return GMPlCrvtrEdgeNextInFace(e -> Twin);
}

/*****************************************************************************
* DESCRIPTION:                                                               *
*   Find the next Edge in the same face.                                     *
*                                                                            *
* PARAMETERS:                                                                *
*   e:  The current Edge.                                                    *
*                                                                            *
* RETURN VALUE:                                                              *
*   GMPlCrvtrEdgeStruct *:  The pointer to next Edge in the face.            *
*****************************************************************************/
static GMPlCrvtrEdgeStruct *GMPlCrvtrEdgeNextInFace(const GMPlCrvtrEdgeStruct *e)
{
    if (!e)
        return NULL;
    if (e -> InFace -> Pnext == e -> InFace -> Tail)
        return (GMPlCrvtrEdgeStruct *) (e -> InFace -> Tail -> Pnext -> Data);
    return (GMPlCrvtrEdgeStruct *) (e -> InFace -> Pnext -> Data);
}

/*****************************************************************************
* DESCRIPTION:                                                               *
*   Organize the Edges of the Vertex, so that it consists of either one      *
*   Edge, if the Vertex is manifold, or the list of right boundary Edges.    *
*                                                                            *
* PARAMETERS:                                                                *
*   Vertex:  The Vertex.                                                     *
*                                                                            *
* RETURN VALUE:                                                              *
*   void                                                                     *
*****************************************************************************/
static void GMPlCrvtrVertexValidateBoundaries(GMPlCrvtrVertexStruct *Vertex)
{
    GMPlCrvtrDListStruct 
        *It = NULL;

    if (Vertex -> Edges == NULL) 
        return;

    for (It = Vertex -> Edges -> Tail -> Pnext; 
	 It != Vertex -> Edges -> Tail; 
	 ) {
        GMPlCrvtrEdgeStruct 
	    *e = (GMPlCrvtrEdgeStruct *) (It -> Data);

        /* Check for an invalid Edge. */
        if (e == NULL)
            continue;
        if (e -> Twin != NULL) {
            It -> Data = NULL;
            It = GMPlCrvtrDeleteAt(It);
            GMPlCrvtrVertexAttachAsBoundary(Vertex, e);
            if (GMPlCrvtrIsEmpty(Vertex -> Edges)) {
                GMPlCrvtrPushFrontEmpty(Vertex -> Edges);
                Vertex -> Edges -> Tail -> Pnext -> Data = e;
            }
        }
	else
	    It = It -> Pnext;
    }
}

/*****************************************************************************
* DESCRIPTION:                                                               *
*   Push the new empty entry into the list.                                  *
*                                                                            *
* PARAMETERS:                                                                *
*   List:  The list data structure.                                          *
*                                                                            *
* RETURN VALUE:                                                              *
*   void                                                                     *
*****************************************************************************/
static void GMPlCrvtrPushFrontEmpty(GMPlCrvtrDListStruct *DList)
{
    GMPlCrvtrDListStruct 
        *n = DList -> Tail -> Pnext;

    DList -> Tail -> Pnext = GMPlCrvtrNewElementAux(DList);
    DList -> Tail -> Pnext -> Pnext = n;
    DList -> Tail -> Pnext -> Pprev = DList -> Tail;
    n -> Pprev = DList -> Tail -> Pnext;
}

/*****************************************************************************
* DESCRIPTION:                                                               *
*   Is the given Edge on the boundary.                                       *
*                                                                            *
* PARAMETERS:                                                                *
*   Vertex: The source Vertex.                                               *
*   Edge:   The Edge that must be checked.                                   *
*                                                                            *
* RETURN VALUE:                                                              *
*   int:  TRUE if the Edge is on the boundary, FALSE otherwise.              *
*****************************************************************************/
static int GMPlCrvtrVertexIsBoundaryEdge(GMPlCrvtrVertexStruct *Vertex, 
					 GMPlCrvtrEdgeStruct *Edge)
{
    GMPlCrvtrDListStruct 
        *EI = NULL;

    if (!Edge || Edge -> Source != Vertex || Edge -> Twin)
        return FALSE;
    if (Vertex -> Edges == NULL)
        return FALSE;
    for (EI = Vertex -> Edges -> Tail -> Pnext; 
	 EI != Vertex -> Edges -> Tail; 
	 EI = EI -> Pnext) {
        GMPlCrvtrEdgeStruct 
            *e = (GMPlCrvtrEdgeStruct *)(EI -> Data);

        if (e == Edge)
            return TRUE;
    }
    return FALSE;
}

/*****************************************************************************
* DESCRIPTION:                                                               *
*   Delete the given entry of the list.                                      *
*                                                                            *
* PARAMETERS:                                                                *
*   Place:  The entry that must be deleted..                                 *
*                                                                            *
* RETURN VALUE:                                                              *
*   GMPlCrvtrDListStruct *:  The pointer to the next entry.                  *
*****************************************************************************/
static GMPlCrvtrDListStruct *GMPlCrvtrDeleteAt(GMPlCrvtrDListStruct *place)
{
    GMPlCrvtrDListStruct 
        *Next = place -> Pnext;

    place -> Pprev -> Pnext = Next;
    Next -> Pprev = place -> Pprev;
    place -> Tail = NULL;
    place -> Pnext = NULL;
    place -> Pprev = NULL;
    if (place -> Data) {
        IritFree(place -> Data);
        place -> Data = NULL;
    }
    IritFree(place);
    return Next;
}

/*****************************************************************************
* DESCRIPTION:                                                               *
*   Initialize the Vertex, from the given data.                              *
*                                                                            *
* PARAMETERS:                                                                *
*   Vertex:  The Vertex that must be initialized.                            *
*   v:       The data for Vertex initialization.                             *
*                                                                            *
* RETURN VALUE:                                                              *
*   void                                                                     *
*****************************************************************************/
static void GMPlCrvtrInitVertexAux(GMPlCrvtrVertexStruct *Vertex,
				   IPVertexStruct *v)
{
    PT_COPY(Vertex -> Vert -> Coord, v -> Coord);
    PT_COPY(Vertex -> Vert -> Normal, v -> Normal);
    PT_SAFE_NORMALIZE(Vertex -> Vert -> Normal);

    GMPlCrvtrInitAttributesAux(Vertex);
}

/*****************************************************************************
* DESCRIPTION:                                                               *
*   Allocates a new GMPlCrvtrFaceStruct data structure.                      *
*                                                                            *
* PARAMETERS:                                                                *
*   None                                                                     *
*                                                                            *
* RETURN VALUE:                                                              *
*   GMPlCrvtrFaceStruct *:  The pointer to the newly allocated face.         *
*****************************************************************************/
static GMPlCrvtrFaceStruct *GMPlCrvtrNewFace(void)
{
    GMPlCrvtrFaceStruct 
	*f = (GMPlCrvtrFaceStruct *) IritMalloc(sizeof(GMPlCrvtrFaceStruct));

    f -> Edges = NULL;
    return f;
}

/*****************************************************************************
* DESCRIPTION:                                                               *
*   Allocates a new GMPlCrvtrEdgeStruct data structure.                      *
*                                                                            *
* PARAMETERS:                                                                *
*   None                                                                     *
*                                                                            *
* RETURN VALUE:                                                              *
*   GMPlCrvtrEdgeStruct *:  The pointer to the newly allocated Edge.         *
*****************************************************************************/
static GMPlCrvtrEdgeStruct *GMPlCrvtrNewEdge(void)
{
    GMPlCrvtrEdgeStruct 
	*e = (GMPlCrvtrEdgeStruct *) IritMalloc(sizeof(GMPlCrvtrEdgeStruct));

    ZAP_MEM(e, sizeof(GMPlCrvtrEdgeStruct));
    return e;
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Allocates a new GMPlCrvtrMeshStruct data structure.                      M
*                                                                            *
* PARAMETERS:                                                                M
*   None                                                                     M
*                                                                            *
* RETURN VALUE:                                                              M
*   GMPlCrvtrMeshStruct *:  The pointer to the newly allocated Mesh.         M
*                                                                            *
* SEE ALSO:                                                                  M
*   GMPlCrvtrDeleteMesh                                                      M
*                                                                            *
* KEYWORDS:                                                                  M
*   GMPlCrvtrNewMesh                                                         M
*****************************************************************************/
GMPlCrvtrMeshStruct *GMPlCrvtrNewMesh(void)
{
    GMPlCrvtrMeshStruct 
        *m = (GMPlCrvtrMeshStruct *) IritMalloc(sizeof(GMPlCrvtrMeshStruct));

    ZAP_MEM(m, sizeof(GMPlCrvtrMeshStruct));
    return m;
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Estimates the Gaussian and Mean curvature values for the given           M
* triangular mesh and initializes the corresponding attributes: for each     M
* vertex: "K1Curv", "K2Curv", "D1", "D2" and "boundary" if the vertex is not M
* two-manifold.								     M
*                                                                            *
* PARAMETERS:                                                                M
*   PolyList:  The triangular two-manifold mesh data structure.              M
*   NumOfRings:  Number of rings around a vertex in the paraboloid fitting.  M
*                                                                            *
* RETURN VALUE:                                                              M
*   void                                                                     M
*                                                                            *
* SEE ALSO:                                                                  M
*   SymbEvalSrfCurvPrep, SymbEvalSrfCurvature                                M
*                                                                            *
* KEYWORDS:                                                                  M
*   GMPlCrvtrSetCurvatureAttr                                                M
*****************************************************************************/
void GMPlCrvtrSetCurvatureAttr(IPPolygonStruct *PolyList, int NumOfRings)
{
    GMPlCrvtrMeshStruct 
        *Mesh = GMPlCrvtrNewMesh();
    GMPlCrvtrAlgorithmStruct 
        *Alg  = GMPlCrvtrNewAlg();
    GMPlCrvtrDListStruct 
        *It = NULL;
    IPPolygonStruct 
        *PIt = NULL;

    if (!GMPlCrvtrInitMesh(Mesh, PolyList))
        return;

    GMPlCrvtrInitAlg(Alg, Mesh, GMPL_CRVTR_PARABOLOID_FIT, NumOfRings);

    GMPlCrvtrRunAlg(Alg);

    for (PIt = PolyList; PIt != NULL; PIt = PIt -> Pnext) {
        IPVertexStruct 
            *VCurr = PIt -> PVertex;

        while (VCurr) {
            GMPlCrvtrVertexStruct
                *Source = (GMPlCrvtrVertexStruct *)
		          (AttrGetIntAttrib(VCurr -> Attr, "UVertexAddress"));

	    AttrFreeAttributes(&VCurr -> Attr);
            VCurr -> Attr = AttrCopyAttributes(Source -> Vert -> Attr);

            VCurr = VCurr -> Pnext;
            if (VCurr == PIt -> PVertex)
                break;
        }
    }

    GMPlCrvtrDeleteMesh(Mesh);
    GMPlCrvtrDeleteAlg(Alg);
}

/*****************************************************************************
* DESCRIPTION:                                                               *
*   Compare two points first by x-coordinate, then by y and, finally by      *
*   z-coordinate (lexico-graphical order)                                    *
*                                                                            *
* PARAMETERS:                                                                *
*   Elem1: Address of the first element.                                     *
*   Elem2: Address of the second element.                                    *
*                                                                            *
* RETURN VALUE:                                                              *
*   int:  -1, 0, 1 based on which is larger/equal.                           *
*****************************************************************************/
static int GMPlCrvtrCompareVerticesByXYZAux(const void *Elem1,
					    const void *Elem2)
{
    VectorType Diff;
    VectorType V1, V2;

    PT_COPY(V1, ((* (IPVertexStruct **) (Elem1))) -> Coord);
    PT_COPY(V2, ((* (IPVertexStruct **) (Elem2))) -> Coord);

    PT_SUB(Diff, V1, V2);
    
    if (fabs(Diff[0]) < IRIT_UEPS && 
        fabs(Diff[1]) < IRIT_UEPS && 
        fabs(Diff[2]) < IRIT_UEPS) 
        return 0;
    if (Diff[0] > IRIT_UEPS)
        return 1;
    if (Diff[0] < -1 * IRIT_UEPS)
        return -1;
    if (Diff[1] > IRIT_UEPS)
        return 1;
    if (Diff[1] < -1 * IRIT_UEPS)
        return -1;
    if (Diff[2] > IRIT_UEPS)
        return 1;
    return -1;
}

/*****************************************************************************
* DESCRIPTION:                                                               *
*   Insert vertex into the Mesh.                                             *
*                                                                            *
* PARAMETERS:                                                                *
*   Mesh:           The Mesh data structure.                                 *
*   v:              The new Vertex data.                                     *
*                                                                            *
* RETURN VALUE:                                                              *
*   void:                                                                    *
*****************************************************************************/
static void GMPlCrvtrInsertVertexAux(IPVertexStruct *v)
{
    GMPlCrvtrVertexStruct 
	*Out = GMPlCrvtrNewVertexAux();

    GMPlCrvtrMeshInWork -> NumOfVertices++;

    GMPlCrvtrPushBackEmpty(GMPlCrvtrMeshInWork -> Vertices);
    GMPlCrvtrMeshInWork -> Vertices -> Tail -> Pprev -> Data = Out;
    GMPlCrvtrInitVertexAux(Out, v);
    GMPlCrvtrLastInitialized = Out;
    AttrSetIntAttrib(&(v -> Attr), "UVertexAddress", (int)(Out));
}

/*****************************************************************************
* DESCRIPTION:                                                               *
*    Stores the address of the unique GMPlCrvtrVertexStruct with             *
*    corresponding coordinate.                                               *
*                                                                            *
*                                                                            *
* PARAMETERS:                                                                *
*   First: the vertex that was met as the first with the same geometry       *
*   Vert:  current vertex                                                    *
*                                                                            *
* RETURN VALUE:                                                              *
*   void                                                                     *
*****************************************************************************/
static void GMPlCrvtrPutVertexIntoMirrorAux(IPVertexStruct *First,
					    IPVertexStruct *Vert)
{
    GMPlCrvtrVertexStruct
        *Inner = (GMPlCrvtrVertexStruct *) (AttrGetIntAttrib(First -> Attr,
							    "UVertexAddress"));
    if (GMPlCrvtrCompareVerticesByXYZAux(&First,
					 &(GMPlCrvtrLastInitialized -> Vert)))
        return;

    AttrSetIntAttrib(&(Vert -> Attr), "UVertexAddress", (int) (Inner));
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Initialize the data structure allocated by GMPlCrvtrNemMesh.             M
*                                                                            *
* PARAMETERS:                                                                M
*   Mesh:       The Mesh for initialization.                                 M
*   PolyList:   The list of polygons for initialization.                     M
*                                                                            *
* RETURN VALUE:                                                              M
*   int:  TRUE if successful, FALSE otherwise.                               M
*                                                                            *
* SEE ALSO:                                                                  M
*   GMPlCrvtrNemMesh                                                         M
*                                                                            *
* KEYWORDS:                                                                  M
*   GMPlCrvtrInitMesh                                                        M
*****************************************************************************/
int GMPlCrvtrInitMesh(GMPlCrvtrMeshStruct *Mesh, IPPolygonStruct *PolyList)
{
    IPPolygonStruct 
        *It = NULL;
    IPObjectStruct
        *Obj = IPGenPolyObject("temp", PolyList, NULL);

    /* Make sure we have only triangles. */
    for (It = Obj -> U.Pl; It != NULL; It = It -> Pnext) {
        int i = 0;
        IPVertexStruct
	    *V = It -> PVertex;

        do {
	    i++;
	    V = V -> Pnext;
	} while (V != NULL && V != It -> PVertex);


	if (i != 3) {
	    GEOM_FATAL_ERROR(GEOM_ERR_TRIANGLES_ONLY);
	    return FALSE;
	}
    }

    GMPlCrvtrMeshInWork = Mesh;
    Mesh -> Faces = GMPlCrvtrNewList(sizeof(GMPlCrvtrFaceStruct));
    IPOpenPolysToClosed(PolyList);

    BoolGenAdjacencies(Obj);

    Mesh -> Vertices = GMPlCrvtrNewList(sizeof(GMPlCrvtrVertexStruct));
    Mesh -> Edges = GMPlCrvtrNewList(sizeof(GMPlCrvtrEdgeStruct));

    GMPlCrvtrForEachVertex(Obj -> U.Pl, 
			   GMPlCrvtrInsertVertexAux, 
			   GMPlCrvtrPutVertexIntoMirrorAux);

    for (It = Obj -> U.Pl; It != NULL; It = It -> Pnext) {
        GMPlCrvtrDListStruct 
            *Vlist = GMPlCrvtrNewList(sizeof(GMPlCrvtrVertexStruct));
        IPVertexStruct	
            *v = NULL;
        GMPlCrvtrVertexStruct
                *Curr = NULL;
        for (v = It -> PVertex; v != NULL; v = v -> Pnext) {
            if (v == It -> PVertex && Vlist -> Tail -> Pnext != Vlist -> Tail)
                break;
            GMPlCrvtrPushBackEmpty(Vlist);
            Curr = (GMPlCrvtrVertexStruct *) (AttrGetIntAttrib(v -> Attr, 
							       "UVertexAddress"));
            Vlist -> Tail -> Pprev -> Data = Curr;
        }
        GMPlCrvtrMeshAddFace(Mesh, Vlist);
        ++Mesh -> NumOfFaces;

        for (Vlist = Vlist -> Tail -> Pnext; 
	     Vlist != Vlist -> Tail; 
	     Vlist = Vlist -> Pnext)
            Vlist -> Data = NULL;

        GMPlCrvtrDeleteList(Vlist);
    }

    Obj -> U.Pl = NULL;
    IPFreeObject(Obj);

    GMPlCrvtrMeshInWork = NULL;
    GMPlCrvtrLastInitialized = NULL;
    return TRUE;
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Perform different kinds of action for the list of unique vertices        M
*   (one per geometric location) and for the copies of these vertices.       M
*                                                                            *
* PARAMETERS:                                                                M
*   PolyList:  The given list of triangles.                                  M
*   ActionNew: The first type of operations (once per geomtric location).    M
*   ActionOld: The second type of operation (for the rest of the vertices in M
*              the given mesh).                                              M
*                                                                            *
* RETURN VALUE:                                                              M
*   void                                                                     M
*                                                                            *
* SEE ALSO:                                                                  M
*   GMPlCrvtrInitMesh                                                        M
*                                                                            *
* KEYWORDS:                                                                  M
*   GMPlCrvtrForEachVertex                                                   M
*****************************************************************************/
void GMPlCrvtrForEachVertex(IPPolygonStruct *PolyList, 
			    void (*ActionNew)(IPVertexStruct *Vertex), 
			    void (*ActionOld)(IPVertexStruct *Prev, 
					      IPVertexStruct *Current))
{
    int ListLen = 0, 
        i = 0;
    IPPolygonStruct *PIt;
    IPVertexStruct *Curr;
    
    /* Count vertices. */
    for (PIt = PolyList; PIt != NULL; PIt = PIt -> Pnext) {
        Curr = PIt -> PVertex;
        while (Curr) {
            ListLen++;

            Curr = Curr -> Pnext;
            if (Curr == PIt -> PVertex)
                break;
        }
    }

    /* Initialize array. */
    GMPlCrvtrVertexArray = (IPVertexStruct **)
			        IritMalloc(sizeof(IPVertexStruct *) * ListLen);
    i = 0;
    for (PIt = PolyList; PIt != NULL; PIt = PIt -> Pnext) {
        Curr = PIt -> PVertex;
        while (Curr) {
            GMPlCrvtrVertexArray[i] = Curr;
            i++;

            Curr = Curr -> Pnext;
            if (Curr == PIt -> PVertex)
                break;
        }
    }

    qsort(GMPlCrvtrVertexArray, ListLen, sizeof(IPVertexStruct *), 
	  GMPlCrvtrCompareVerticesByXYZAux);
    
    ActionNew(GMPlCrvtrVertexArray[0]);
    Curr = GMPlCrvtrVertexArray[0];
    for (i = 1; i < ListLen; ++i) {
        if (GMPlCrvtrCompareVerticesByXYZAux(&(GMPlCrvtrVertexArray[i - 1]), 
					     &(GMPlCrvtrVertexArray[i]))) {
            ActionNew(GMPlCrvtrVertexArray[i]);
            Curr = GMPlCrvtrVertexArray[i];
        }
	else
            ActionOld(Curr, GMPlCrvtrVertexArray[i]);
    }
    for (i = 0; i < ListLen; ++i) 
        GMPlCrvtrVertexArray[i] = NULL;

    IritFree(GMPlCrvtrVertexArray);

    GMPlCrvtrVertexArray = NULL;
}
