/*****************************************************************************
* Entry point of the program. Configuration and global data variables.       *
******************************************************************************
* (C) Gershon Elber, Technion, Israel Institute of Technology                *
******************************************************************************
* Written by:  David Shafrir & Alex Reicher       Ver 0.3, Sep. 2003         *
*****************************************************************************/
#include <irit_sm.h>
#include <misc_lib.h>
#include <geom_lib.h>
#include <rndr_lib.h>
#include "parser.h"
#include "config.h"

#ifdef __FreeBSD__
#include <floatingpoint.h>
#endif

/* Contains all configuration options. Subject to change by config file and  */
/* and command line parameters.                                              */
GLOBAL_DATA GlobalOptionsStruct Options;
STATIC_DATA RealType GlblLastTime,
    GlblZMin = 0,
    GlblZMax = 0;

static void ReverseVertexNormal(IPVertexStruct *v);
static void ReversePlaneNormal(IPPolygonStruct *Poly);
static void TakeTime(const char *msg);
static void ScanObjects(IRndrPtrType Rend,
			IPObjectStruct *Object,
			BoolType DoClipping);
static void LightListLoad(IRndrPtrType Rend, IPObjectStruct* Objects);
static void ComputeZDepth(IPVertexStruct *Vertex);

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Performs configuration management, preprocesses data files and renderes. M
*                                                                            *
* PARAMETERS:                                                                M
*   argc:    IN, command line parameters number.                             M
*   argv:    IN, command line parameters.                                    M
*                                                                            *
* RETURN VALUE:                                                              M
*   int:    Exit reason code.                                                M
*                                                                            *
* KEYWORDS:                                                                  M
*   main                                                                     M
*****************************************************************************/
int main(int argc, char *argv[])
{
#ifdef __FreeBSD__
    fpsetmask(FP_X_INV | FP_X_DZ);
#endif /* __FreeBSD__ */

    IPObjectStruct *Objects, *MatrixObjects;
    char *BaseDirectory,
	*MatrixFiles[] = {NULL, NULL};
    IRndrPtrType Rend;
    BoolType DoClipping;

    DoClipping = TRUE;
    GlblLastTime = IritCPUTime(FALSE);
    InitOptions();
    GetConfig(argv);
    GetOptions(argc, argv);
    BaseDirectory = argv[0];

    Objects = ParseFiles(Options.NFiles, Options.Files,
			 Options.HasTime, Options.Time );

    if (Options.MatricesFileName) {
        MatrixFiles[0] = Options.MatricesFileName;
        MatrixObjects = IPGetDataFiles(MatrixFiles, 1, TRUE, FALSE);
	fprintf(stderr, "\nirender: '-m' option not supported\\n");
    }

    Rend = IRndrInitialize(Options.XSize,
			   Options.YSize,
			   Options.FilterName ? 3 : 1,
			   Options.Transp,
			   Options.BackFace,
			   Options.BackGround,
			   Options.Ambient);
    if (Options.Stencil)
	IRndrClearStencil(Rend);

    IRndrSetShadeModel(Rend, (IRndrShadingType) Options.ShadeModel);
    IRndrSetViewPrsp(Rend,
		     IPWasViewMat ? IPViewMat : NULL,
		     IPWasPrspMat ? IPPrspMat : NULL);

    if (Options.NormalReverse) {
        IPForEachVertex(Objects, ReverseVertexNormal);
        IPForEachPoly(Objects, ReversePlaneNormal);
    }

    if (Options.ZNear || Options.ZFar) {
        IRndrSetZBounds(Rend, Options.ZNear, Options.ZFar);
    }
    else {
        /* Default: no clipping. */
        DoClipping = 0;
    }
    if (Options.PllMaxW != Options.PllMinW) {
        GlblZMin = FAREST_Z;
	GlblZMax = NEAREST_Z;
	/* Used for polylines width computation. */
	IPForEachVertex(Objects, ComputeZDepth);
    }
    TakeTime("Parsing and transform");

    IRndrSetFilter(Rend, Options.FilterName);
    LightListLoad(Rend, Objects);
    ScanObjects(Rend, Objects, DoClipping );
    TakeTime("Scan conversion");

    if (Options.ZDepth) {
        IRndrSaveFileDepth(Rend, BaseDirectory,
			   Options.OutFileName, Options.FileType);
    }
    else if (Options.Stencil) {
        IRndrSaveFileStencil(Rend, BaseDirectory,
			     Options.OutFileName, Options.FileType);
    }
    else {
        IRndrSaveFile(Rend, BaseDirectory,
		      Options.OutFileName, Options.FileType);
    }
    IRndrDestroy(Rend);
    TakeTime("Z-Buffer dump");

#ifdef __FreeBSD__
    fpresetsticky(FP_X_INV | FP_X_OFL | FP_X_UFL |
		  FP_X_DZ | FP_X_IMP | FP_X_DNML);
#endif /* __FreeBSD__ */

    return 0;
}
/*****************************************************************************
* DESCRIPTION:                                                               *
*   Prints the time the last action took.                                    *
*                                                                            *
* PARAMETERS:                                                                *
*   msg:       IN, the name of the action.                                   *
*                                                                            *
* RETURN VALUE:                                                              *
*   void                                                                     *
*****************************************************************************/
void TakeTime(const char *Msg)
{
    RealType
        Time = IritCPUTime(FALSE);

    if (Options.Verbose)
        fprintf(stderr, "\nirender: '%s' done in %0.2f s",
		Msg, Time - GlblLastTime );

    GlblLastTime = Time;
}

/*****************************************************************************
* DESCRIPTION:                                                               *
*   Set vertex normal in opposite direction.                                 *
*                                                                            *
* PARAMETERS:                                                                *
*   v:       IN OUT, pointer to vertex object.                               *
*                                                                            *
* RETURN VALUE:                                                              *
*   void                                                                     *
*****************************************************************************/
static void ReverseVertexNormal(IPVertexStruct *v)
{
    PT_SCALE(v -> Normal, -1.0);
}

/*****************************************************************************
* DESCRIPTION:                                                               *
*   Sets plane equation to have normal defined in opposite direction.        *
*                                                                            *
* PARAMETERS:                                                                *
*   Poly:       IN OUT, pointer to the polygon.                              *
*                                                                            *
* RETURN VALUE:                                                              *
*   void                                                                     *
*****************************************************************************/
static void ReversePlaneNormal(IPPolygonStruct *Poly)
{
    PT_SCALE(Poly -> Plane, -1.0);
    Poly -> Plane[3] = -Poly -> Plane[3];
}

/*****************************************************************************
* DESCRIPTION:                                                               *
*   Tringulates an Irit object.                                              *
*                                                                            *
* PARAMETERS:                                                                *
*   PObj:       IN OUT, pointer to the polygon.                              *
*                                                                            *
* RETURN VALUE:                                                              *
*   IPObjectStruct* : the result object.                                     *
*****************************************************************************/
IPObjectStruct* IObjectTriangulate(IPObjectStruct *PObj)
{
    int OldCirc;
    IPObjectStruct *Result,
        *PTmp = IPCopyObject(NULL, PObj, TRUE);

    /* Make sure all polygons are convex. */
    OldCirc = IPSetPolyListCirc(TRUE);
    IPOpenPolysToClosed(PTmp -> U.Pl);
    GMConvexPolyObject(PTmp);
    IPSetPolyListCirc(OldCirc);
    IPClosedPolysToOpen(PTmp -> U.Pl);

    /* Make sure all polygons are triangle. */
    Result = GMConvertPolysToTriangles(PTmp);
    Result -> Attr = AttrCopyAttributes(PObj -> Attr);
    IPFreeObject(PTmp);
    return Result;
}

/*****************************************************************************
* DESCRIPTION:                                                               *
*   A helper callback function to compute the scene min and max depth.       *
*                                                                            *
* PARAMETERS:                                                                *
*   Vertex:       IN, pointer to the vertex.                                 *
*                                                                            *
* RETURN VALUE:                                                              *
*   void                                                                     *
*****************************************************************************/
static void ComputeZDepth(IPVertexStruct *Vertex)
{
    PointType Result;

    PT_COPY(Result, Vertex -> Coord);
    if (IPWasViewMat)
        MatMultPtby4by4( Result, Result, IPViewMat );
    if (IPWasPrspMat )
        MatMultPtby4by4( Result, Result, IPPrspMat );

    if (Result[Z_AXIS] NEAR_THAN GlblZMin)
        GlblZMin = Result[Z_AXIS];
    if (Result[Z_AXIS] FARTHER_THAN GlblZMax)
        GlblZMax = Result[Z_AXIS];
}

/*****************************************************************************
* DESCRIPTION:                                                               *
*   Iterates the object list and clips out partially visible polygons.       *
*   Warning : very simplistic and naive.                                     *
*                                                                            *
* PARAMETERS:                                                                *
*   Object:           IN, pointer to the object list.                        *
*   ClippingPlanes:   IN, pointer to the 6 clipping plane.                   *
*                                                                            *
* RETURN VALUE:                                                              *
*   void                                                                     *
*****************************************************************************/
void IObjectClip(IPObjectStruct *Object, PlaneType *ClippingPlanes)
{
    IPPolygonStruct *Poly;
    IPVertexStruct *Vertex;
    IPObjectStruct
        *First = Object;
    int i;

    if (IP_IS_POLY_OBJ(Object) && !IP_IS_POLYLINE_OBJ(Object)) {
        for (Poly = Object -> U.Pl;
	     Poly != NULL;
	     Poly = Poly != NULL ? Poly -> Pnext : Object -> U.Pl) {
            for (Vertex = Poly -> PVertex;
		 Vertex != NULL;
		 Vertex = Vertex -> Pnext) {
                for (i = 4; i < 6 ; i++) {
                    if ((DOT_PROD(Vertex -> Coord, ClippingPlanes[i])
                                               + ClippingPlanes[i][3]) < 0) {
                        AttrSetRealAttrib(&Poly -> Attr, "_INVIS", 1);
                        /* Set it invisible: we can't simply remove it -    */
                        /* it will destroy texture consitency.              */
                    }
                }
            }
        }
    }
}

/*****************************************************************************
* DESCRIPTION:                                                               *
*   Iterates the object list and scans all polygons of them.                 *
*                                                                            *
* PARAMETERS:                                                                *
*   Rend:       IN, the render context.                                      *
*   Object:     IN, pointer to the object list.                              *
*                                                                            *
* RETURN VALUE:                                                              *
*   void                                                                     *
*****************************************************************************/
void ScanObjects(IRndrPtrType Rend, IPObjectStruct *Object, BoolType DoClipping)
{
    IPPolygonStruct *Poly;
    IPObjectStruct *Next, *ObjectTri;
    IPVertexStruct *Vertex;
    int IsPolyLine;
    PlaneType ClipPlanes[6];

    IRndrGetClippingPlanes(Rend, ClipPlanes );

    for (; Object; Object = Next) {
        Next = Object -> Pnext;
        if (IP_IS_POLY_OBJ(Object) ) {
            IsPolyLine = IP_IS_POLYLINE_OBJ(Object);
            /* Don't triangulate polylines. */
            ObjectTri = IsPolyLine ? Object : IObjectTriangulate(Object);
            if (DoClipping)
                IObjectClip(ObjectTri, ClipPlanes);
            IRndrBeginObject(Rend, ObjectTri);

            for (Poly = ObjectTri -> U.Pl; Poly; Poly = Poly -> Pnext) {
                if (IsPolyLine) {
		    RealType
			Width = AttrGetObjectWidth(Object);

		    if (IP_ATTR_IS_BAD_REAL(Width))
		        IRndrSetPllParams(Rend, Options.PllMinW,
					        Options.PllMaxW,
					  GlblZMin, GlblZMax);
		    else {
		        RealType
			    Width2 = Width * Options.PllMinW / Options.PllMaxW;

		        IRndrSetPllParams(Rend, Width2, Width,
					  GlblZMin, GlblZMax);
		    }

                    IRndrBeginPll(Rend);
                    for (Vertex = Poly -> PVertex;
			 Vertex != NULL;
			 Vertex = Vertex -> Pnext) {
                        IRndrPutPllVertex(Rend, Vertex);
                    }
                    IRndrEndPll(Rend);
                }
                else {
                    IRndrPutTriangle(Rend, Poly);
                }
            }
            IRndrEndObject(Rend);
        }
        IPFreeObject(Object);
    }
}
/*****************************************************************************
* DESCRIPTION:                                                               *
*   Iterates the object list to find light sources.                          *
*   And then add them to the rendering context.                              *
*                                                                            *
* PARAMETERS:                                                                *
*   Rend:       IN, the render context.                                      *
*   Object:     IN, pointer to the object list.                              *
*                                                                            *
* RETURN VALUE:                                                              *
*   void                                                                     *
*****************************************************************************/
void LightListLoad(IRndrPtrType Rend, IPObjectStruct* Objects)
{
    IPObjectStruct *o;
    char *t;
    int r, g, b,
        SeenLight = FALSE;
    IRndrLightType Type;
    PointType Where, Color;

    for (o = Objects; o; o = o -> Pnext) {
        if (!IP_ATTR_IS_BAD_INT(AttrGetObjectIntAttrib(o, "LIGHT_SOURCE"))) {
            t = AttrGetObjectStrAttrib(o, "TYPE");
            if (!t || stricmp(t, "POINT_POS")) {
                Type = IRNDR_LIGHT_VECTOR;
            }
            else {
                Type = IRNDR_LIGHT_POINT;
            }
            PT_COPY(Where, o -> U.Pt);
            if (Type == IRNDR_LIGHT_VECTOR) {
                PT_NORMALIZE(Where);
            }
            if (AttrGetObjectRGBColor(o, &r, &g, &b)) {
                Color[RED_CLR] = r;
                Color[GREEN_CLR] = g;
                Color[BLUE_CLR] = b;
                PT_SCALE(Color, 1.0 / 0xff);
            }
            else {
                Color[RED_CLR] = 1.0;
                Color[GREEN_CLR] = 1.0;
                Color[BLUE_CLR] = 1.0;
            }
            IRndrAddLightSource(Rend, Type, Where, Color);
            SeenLight = TRUE;
            if (!IP_ATTR_IS_BAD_INT(AttrGetObjectIntAttrib(o, "TWOLIGHT"))) {
                PT_SCALE(Where, -1);
                IRndrAddLightSource(Rend, Type, Where, Color);
            }
        }
    }
    if (!SeenLight) {
        PointType
	    Ones = { 1.0, 1.0, 1.0 },
            MinusOnes = { -1.0, -1.0, -1.0 };

        IRndrAddLightSource(Rend, IRNDR_LIGHT_VECTOR, Ones, Ones);
        IRndrAddLightSource(Rend, IRNDR_LIGHT_VECTOR, MinusOnes, Ones);
    }
}
