/*****************************************************************************
*   "Irit" - the 3d (not only polygonal) solid modeller.		     *
*									     *
* Written by:  Gershon Elber				Ver 0.2, Mar. 1990   *
******************************************************************************
* (C) Gershon Elber, Technion, Israel Institute of Technology                *
******************************************************************************
*   Module to evaluate the binary tree generated by the InptPrsr module.     *
*   All the objects are handled the same but the numerical one, which is     *
* moved as a RealType and not as an object (only internally within this	     *
* module) as it is frequently used and consumes much less memory this way.   *
*   Note this module is par of InptPrsr module and was splited only because  *
* of text file sizes problems...					     *
*****************************************************************************/

#include <stdio.h>
#include <ctype.h>
#include <math.h>
#include <string.h>
#include "program.h"
#include "allocate.h"
#include "attribut.h"
#include "ctrl-brk.h"
#include "dosintr.h"
#include "freeform.h"
#include "geom_lib.h"
#include "inptprsg.h"
#include "inptprsl.h"
#include "objects.h"
#include "overload.h"
#include "iritgrap.h"
#include "bsc_geom.h"

#ifdef SUN4
    static double natural_log(double x) { return log10(x) / log10(M_E); }
#else
#   define natural_log  log
#endif /* SUN4 */

#ifndef IRIT_QUIET_STRINGS
static void IritQueryFunctions(void);
static void IritQueryKeywords(void);
#endif /* !IRIT_QUIET_STRINGS */
static void IritQuery(RealType *RQuery);

/* Although the type of parameters is specified (for InptPrsrTypeCheck rtn)  */
/* All the parameters to the following dispatched functions are passed by    */
/* address.								     */

NumFuncTableType NumFuncTable[] = {
    { "ACOS",	 ARCCOS,   acos,		1,	{ NUMERIC_EXPR } },
    { "ASIN",	 ARCSIN,   asin,		1,	{ NUMERIC_EXPR } },
    { "ATAN2",	 ARCTAN2,  atan2,		2,	{ NUMERIC_EXPR, NUMERIC_EXPR } },
    { "ATAN",	 ARCTAN,   atan,		1,	{ NUMERIC_EXPR } },
    { "COS",	 COS,	   cos,			1,	{ NUMERIC_EXPR } },
    { "EXP",	 EXP,	   exp,			1,	{ NUMERIC_EXPR } },
    { "ABS",	 FABS,	   fabs,		1,	{ NUMERIC_EXPR } },
    { "FLOOR",	 FLOOR,	   floor,		1,	{ NUMERIC_EXPR } },
    { "FMOD",	 FMOD,	   fmod,		2,	{ NUMERIC_EXPR, NUMERIC_EXPR } },
    { "POWER",	 FPOWER,   pow,			2,	{ NUMERIC_EXPR, NUMERIC_EXPR } },
    { "LN",	 LN,	   natural_log,		1,	{ NUMERIC_EXPR } },
    { "LOG",	 LOG,	   log10,		1,	{ NUMERIC_EXPR } },
    { "SIN",	 SIN,	   sin,			1,	{ NUMERIC_EXPR } },
    { "SQRT",	 SQRT,	   sqrt,		1,	{ NUMERIC_EXPR } },
    { "TAN",	 TAN,	   tan,			1,	{ NUMERIC_EXPR } },
    { "CPOLY",	 CPOLY,	   PolyCountPolys,	1,	{ POLY_EXPR } },
    { "AREA",	 AREA,	   GMPolyObjectArea,	1,	{ POLY_EXPR } },
    { "VOLUME",  VOLUME,   GMPolyObjectVolume,	1, 	{ POLY_EXPR } },
    { "TIME",	 TIME,	   DosGetTime,		1,	{ NUMERIC_EXPR } },
    { "SIZEOF",  SIZEOF,   GetObjectSize,	1,	{ OLST_EXPR | POLY_EXPR | CURVE_EXPR | STRING_EXPR } },
    { "MESHSIZE",MESHSIZE, GetMeshSize,		2,	{ CURVE_EXPR | SURFACE_EXPR | TRIMSRF_EXPR | TRIVAR_EXPR | TRISRF_EXPR | MULTIVAR_EXPR, NUMERIC_EXPR } },
    { "THISOBJ", THISOBJ,  ThisObjectIs,	1,	{ STRING_EXPR } },
    { "RANDOM",  IRT_RNDM, IritRandom,		2,	{ NUMERIC_EXPR, NUMERIC_EXPR } },
    { "CLNTEXEC",CLNTEXEC, ClientExecute,	1,	{ STRING_EXPR } },
    { "DSTPTLN", DSTPTLN,  DistPointLine,	3,	{ POINT_EXPR, POINT_EXPR, VECTOR_EXPR } },
    { "DSTPTPLN",DSTPTPLN, DistPointPlane,	2,	{ POINT_EXPR, PLANE_EXPR } },
    { "DSTLNLN", DSTLNLN,  DistLineLine,	4,	{ POINT_EXPR, VECTOR_EXPR, POINT_EXPR, VECTOR_EXPR } },
    { "ZCOLLIDE",ZCOLLIDE, ComputeZCollisions,	4,	{ ANY_EXPR, ANY_EXPR, NUMERIC_EXPR, NUMERIC_EXPR } }
};
int NumFuncTableSize = sizeof(NumFuncTable) / sizeof(NumFuncTableType);

ObjFuncTableType ObjFuncTable[] = {
    { "POINT",	  POINT,	IPGenPTObject,		3,	{ NUMERIC_EXPR, NUMERIC_EXPR, NUMERIC_EXPR },		POINT_EXPR },
    { "VECTOR",	  VECTOR,	IPGenVECObject,		3,	{ NUMERIC_EXPR, NUMERIC_EXPR, NUMERIC_EXPR },		VECTOR_EXPR },
    { "PLANE",	  PLANE,	IPGenPLANEObject,	4,	{ NUMERIC_EXPR, NUMERIC_EXPR, NUMERIC_EXPR, NUMERIC_EXPR }, PLANE_EXPR },
    { "CTLPT",	  CTLPT,	InptEvalCtlPtFromParams, ANY_PARAM_NUM, { 0 },						CTLPT_EXPR },
    { "ROTX",	  ROTX,		GMGenMatObjectRotX,	1,	{ NUMERIC_EXPR },					MATRIX_EXPR },
    { "ROTY",	  ROTY,		GMGenMatObjectRotY,	1,	{ NUMERIC_EXPR },					MATRIX_EXPR },
    { "ROTZ",	  ROTZ,		GMGenMatObjectRotZ,	1,	{ NUMERIC_EXPR },					MATRIX_EXPR },
    { "ROTVEC",	  ROTVEC,	GMGenMatObjectRotVec,	2,	{ VECTOR_EXPR, NUMERIC_EXPR },				MATRIX_EXPR },
    { "ROTZ2V2",  ROTZ2V2,	GMGenMatObjectZ2Dir2,	2,	{ VECTOR_EXPR, VECTOR_EXPR },				MATRIX_EXPR },
    { "ROTZ2V",	  ROTZ2V,	GMGenMatObjectZ2Dir,	1,	{ VECTOR_EXPR },					MATRIX_EXPR },
    { "TRANS",	  TRANS,	GMGenMatObjectTrans,	1,	{ VECTOR_EXPR },					MATRIX_EXPR },
    { "SCALE",	  SCALE,	GMGenMatObjectScale,	1,	{ VECTOR_EXPR },					MATRIX_EXPR },
    { "BOX",	  BOX,		GenBOXObject,		4,	{ VECTOR_EXPR, NUMERIC_EXPR, NUMERIC_EXPR, NUMERIC_EXPR }, POLY_EXPR },
    { "GBOX",	  GBOX,		GenGBOXObject,		4,	{ VECTOR_EXPR, VECTOR_EXPR, VECTOR_EXPR, VECTOR_EXPR }, POLY_EXPR },
    { "CONE",	  CONE,		GenCONEObject,		4,	{ VECTOR_EXPR, VECTOR_EXPR, NUMERIC_EXPR, NUMERIC_EXPR }, POLY_EXPR | SURFACE_EXPR },
    { "CON2",	  CONE2,	GenCONE2Object,		5,	{ VECTOR_EXPR, VECTOR_EXPR, NUMERIC_EXPR, NUMERIC_EXPR, NUMERIC_EXPR }, POLY_EXPR | SURFACE_EXPR },
    { "CYLIN",	  CYLIN,	GenCYLINObject,		4,	{ VECTOR_EXPR, VECTOR_EXPR, NUMERIC_EXPR, NUMERIC_EXPR }, POLY_EXPR | SURFACE_EXPR },
    { "SPHERE",	  SPHERE,	GenSPHEREObject,	2,	{ VECTOR_EXPR, NUMERIC_EXPR },				POLY_EXPR | SURFACE_EXPR },
    { "TORUS",	  TORUS,	GenTORUSObject,		4,	{ VECTOR_EXPR, VECTOR_EXPR, NUMERIC_EXPR, NUMERIC_EXPR }, POLY_EXPR | SURFACE_EXPR },
    { "CIRCPOLY", CIRCPOLY,	GenPOLYDISKObject,	3,	{ VECTOR_EXPR, VECTOR_EXPR, NUMERIC_EXPR },		POLY_EXPR },
    { "POLY",     POLY,		GenPOLYGONObject,	2,	{ OLST_EXPR, NUMERIC_EXPR },				POLY_EXPR },
    { "CROSSEC",  CROSSEC,	GenCROSSECObject,	1,	{ POLY_CURVE_EXPR },					POLY_EXPR },
    { "SURFREV",  SURFREV,	GenSURFREVObject,	1,	{ POLY_CURVE_EXPR },					POLY_EXPR | SURFACE_EXPR },
    { "SURFREVAXS",SURFREVAXS,	GenSURFREVAxisObject,	2,	{ POLY_CURVE_EXPR, VECTOR_EXPR },			POLY_EXPR | SURFACE_EXPR },
    { "SURFREV2", SURFREV2,	GenSURFREV2Object,	3,	{ POLY_CURVE_EXPR, NUMERIC_EXPR, NUMERIC_EXPR },	POLY_EXPR | SURFACE_EXPR },
    { "SURFREVAX2",SURFREVAXS2, GenSURFREV2AxisObject,	4,	{ POLY_CURVE_EXPR, NUMERIC_EXPR, NUMERIC_EXPR, VECTOR_EXPR }, POLY_EXPR | SURFACE_EXPR },
    { "SURFPREV", SURFPREV,	GenSURFPREVObject,	1,	{ CURVE_EXPR },						SURFACE_EXPR },
    { "EXTRUDE",  EXTRUDE,	GenEXTRUDEObject,	3,	{ POLY_CURVE_EXPR | SURFACE_EXPR, VECTOR_EXPR, NUMERIC_EXPR }, POLY_EXPR | SURFACE_EXPR | TRIVAR_EXPR | OLST_EXPR },
    { "LIST",	  LIST,		InptEvalGenObjectList,	ANY_PARAM_NUM, { 0 },						OLST_EXPR },
    { "LOAD",     LOAD,		LoadObjectFromFile,	1,	{ STRING_EXPR },					ANY_EXPR },
    { "CONVEX",	  CONVEX,	GMConvexPolyObjectN,	1,	{ POLY_EXPR  | OLST_EXPR },				POLY_EXPR | OLST_EXPR },
    { "SPOWER",   SPOWER,	GenPowerSurfaceObject,	1,	{ OLST_EXPR },						SURFACE_EXPR },
    { "CPOWER",   CPOWER,	GenPowerCurveObject,	1,	{ OLST_EXPR },						CURVE_EXPR },
    { "SBEZIER",  SBEZIER,	GenBezierSurfaceObject, 1,	{ OLST_EXPR },						SURFACE_EXPR },
    { "CBEZIER",  CBEZIER,	GenBezierCurveObject,	1,	{ OLST_EXPR },						CURVE_EXPR },
    { "SBSPLINE", SBSPLINE,	GenBsplineSurfaceObject,4,	{ NUMERIC_EXPR, NUMERIC_EXPR, OLST_EXPR, OLST_EXPR },	SURFACE_EXPR },
    { "CBSPLINE", CBSPLINE,	GenBsplineCurveObject,	3,	{ NUMERIC_EXPR, OLST_EXPR, OLST_EXPR },			CURVE_EXPR },
    { "SEVAL",    SEVAL,	EvalSurfaceObject,	3,	{ TRIMSRF_EXPR | SURFACE_EXPR, NUMERIC_EXPR, NUMERIC_EXPR }, CTLPT_EXPR },
    { "CEVAL",    CEVAL,	EvalCurveObject,	2,	{ CURVE_EXPR, NUMERIC_EXPR },				CTLPT_EXPR },
    { "STANGENT", STANGENT,	TangentSurfaceObject,	4,	{ TRIMSRF_EXPR | SURFACE_EXPR, NUMERIC_EXPR, NUMERIC_EXPR, NUMERIC_EXPR }, VECTOR_EXPR },
    { "CTANGENT", CTANGENT,	TangentCurveObject,	2,	{ CURVE_EXPR, NUMERIC_EXPR },				VECTOR_EXPR },
    { "PNORMAL",  PNORMAL,	SetPolyVrtxNormal,	3,	{ POLY_EXPR, NUMERIC_EXPR, ANY_EXPR },			VECTOR_EXPR },
    { "PATTRIB",  PATTRIB,      SetGetPolyVrtxAttrib,   4,      { POLY_EXPR, NUMERIC_EXPR, STRING_EXPR, ANY_EXPR },	ANY_EXPR },
    { "SNORMAL",  SNORMAL,	NormalSurfaceObject,	3,	{ TRIMSRF_EXPR | SURFACE_EXPR, NUMERIC_EXPR, NUMERIC_EXPR }, VECTOR_EXPR },
    { "TSNORMAL", TSNORMAL,	NormalTriSrfObject,	4,	{ TRISRF_EXPR, NUMERIC_EXPR, NUMERIC_EXPR, NUMERIC_EXPR }, VECTOR_EXPR },
    { "CNORMAL",  CNORMAL,	NormalCurveObject,	2,	{ CURVE_EXPR, NUMERIC_EXPR },				VECTOR_EXPR },
    { "MDIVIDE",  MDIVIDE,	DivideMultivarObject,	3,	{ MULTIVAR_EXPR, NUMERIC_EXPR, NUMERIC_EXPR },		OLST_EXPR },
    { "TDIVIDE",  TDIVIDE,	DivideTrivarObject,	3,	{ TRIVAR_EXPR, NUMERIC_EXPR, NUMERIC_EXPR },		OLST_EXPR },
    { "SDIVIDE",  SDIVIDE,	DivideSurfaceObject,	3,	{ TRIMSRF_EXPR | SURFACE_EXPR, NUMERIC_EXPR, NUMERIC_EXPR }, OLST_EXPR },
    { "CDIVIDE",  CDIVIDE,	DivideCurveObject,	2,	{ CURVE_EXPR, NUMERIC_EXPR },				OLST_EXPR },
    { "MREGION",  MREGION,	RegionFromMultivarObject,4,	{ MULTIVAR_EXPR, NUMERIC_EXPR, NUMERIC_EXPR, NUMERIC_EXPR }, MULTIVAR_EXPR },
    { "TREGION",  TREGION,	RegionFromTrivarObject, 4,	{ TRIVAR_EXPR, NUMERIC_EXPR, NUMERIC_EXPR, NUMERIC_EXPR }, TRIVAR_EXPR },
    { "SREGION",  SREGION,	RegionFromSurfaceObject,4,	{ TRIMSRF_EXPR | SURFACE_EXPR, NUMERIC_EXPR, NUMERIC_EXPR, NUMERIC_EXPR }, SURFACE_EXPR },
    { "CREGION",  CREGION,	RegionFromCurveObject,	3,	{ CURVE_EXPR, NUMERIC_EXPR, NUMERIC_EXPR },		CURVE_EXPR },
    { "MREFINE",  MREFINE,	RefineMultivarObject,	4,	{ MULTIVAR_EXPR, NUMERIC_EXPR, NUMERIC_EXPR, OLST_EXPR }, MULTIVAR_EXPR },
    { "TREFINE",  TREFINE,	RefineTrivarObject,	4,	{ TRIVAR_EXPR, NUMERIC_EXPR, NUMERIC_EXPR, OLST_EXPR }, TRIVAR_EXPR },
    { "SREFINE",  SREFINE,	RefineSurfaceObject,	4,	{ SURFACE_EXPR, NUMERIC_EXPR, NUMERIC_EXPR, OLST_EXPR }, SURFACE_EXPR },
    { "CREFINE",  CREFINE,	RefineCurveObject,	3,	{ CURVE_EXPR, NUMERIC_EXPR, OLST_EXPR },		CURVE_EXPR },
    { "MRAISE",   MRAISE,	RaiseMultivarObject,	3,	{ MULTIVAR_EXPR, NUMERIC_EXPR, NUMERIC_EXPR },		MULTIVAR_EXPR },
    { "TRAISE",   TRAISE,	RaiseTrivarObject,	3,	{ TRIVAR_EXPR, NUMERIC_EXPR, NUMERIC_EXPR },		TRIVAR_EXPR },
    { "SRAISE",   SRAISE,	RaiseSurfaceObject,	3,	{ SURFACE_EXPR, NUMERIC_EXPR, NUMERIC_EXPR },		SURFACE_EXPR },
    { "CRAISE",   CRAISE,	RaiseCurveObject,	2,	{ CURVE_EXPR, NUMERIC_EXPR },				CURVE_EXPR },
    { "CREDUCE",  CREDUCE,	ReduceCurveObject,	2,	{ CURVE_EXPR, NUMERIC_EXPR },				CURVE_EXPR },
    { "CSURFACE", CSURFACE,	CurveFromSurface,	3,	{ SURFACE_EXPR | TRISRF_EXPR, NUMERIC_EXPR, NUMERIC_EXPR }, CURVE_EXPR },
    { "CMESH",    CMESH,	CurveFromSrfMesh,	3,	{ SURFACE_EXPR, NUMERIC_EXPR, NUMERIC_EXPR },		CURVE_EXPR },
    { "NTH",	  NTH,		GetNthList,		2,	{ OLST_EXPR, NUMERIC_EXPR },				ANY_EXPR },
    { "NREF",	  NREF,		RefNthList,		2,	{ OLST_EXPR, NUMERIC_EXPR },				ANY_EXPR },
    { "GPOLYGON", GPOLYGON,	Geometry2Polygons,	2,	{ OLST_GEOM_EXPR, NUMERIC_EXPR },			POLY_EXPR },
    { "GPOLYLINE",GPOLYLINE,	Geometry2Polylines,	2,	{ OLST_GEOM_EXPR, NUMERIC_EXPR },			POLY_EXPR },
    { "CIRCLE",   CIRCLE,	GenCircleCurveObject,	2,	{ VECTOR_EXPR, NUMERIC_EXPR },				CURVE_EXPR },
    { "PCIRCLE",  PCIRCLE,	GenPCircleCurveObject,	2,	{ VECTOR_EXPR, NUMERIC_EXPR },				CURVE_EXPR },
    { "ARC",	  ARC,		GenArcCurveObject,	3,	{ VECTOR_EXPR, VECTOR_EXPR, VECTOR_EXPR },		CURVE_EXPR },
    { "ARC360",	  ARC360,	GenArc2CurveObject,	4,	{ VECTOR_EXPR, NUMERIC_EXPR, NUMERIC_EXPR, NUMERIC_EXPR }, CURVE_EXPR },
    { "RULEDSRF", RULEDSRF,	GenRuledSrfObject,	2,	{ POLY_CURVE_EXPR, POLY_CURVE_EXPR },			SURFACE_EXPR | POLY_EXPR },
    { "RULEDTV",  RULEDTV,	GenRuledTVObject,	2,	{ SURFACE_EXPR, SURFACE_EXPR },				TRIVAR_EXPR },
    { "BOOLSUM",  BOOLSUM,	GenBoolSumSrfObject,	4,	{ CURVE_EXPR, CURVE_EXPR, CURVE_EXPR, CURVE_EXPR },	SURFACE_EXPR },
    { "BOOLONE",  BOOLONE,	GenBoolOneSrfObject,	1,	{ CURVE_EXPR },						SURFACE_EXPR },
    { "SFROMCRVS", SFROMCRVS,	GenSrfFromCrvsObject,	3,	{ OLST_EXPR, NUMERIC_EXPR, NUMERIC_EXPR },		SURFACE_EXPR },
    { "SWEEPSRF", SWEEPSRF,	GenSweepSrfObject,	3,	{ CURVE_EXPR | OLST_EXPR, CURVE_EXPR, CURVE_EXPR | VECTOR_EXPR | NUMERIC_EXPR }, SURFACE_EXPR },
    { "SWPSCLSRF", SWPSCLSRF,	GenSweepScaleSrfObject,	5,	{ CURVE_EXPR | OLST_EXPR, CURVE_EXPR, CURVE_EXPR | NUMERIC_EXPR, CURVE_EXPR | VECTOR_EXPR | NUMERIC_EXPR, NUMERIC_EXPR }, SURFACE_EXPR },
    { "OFFSET",	  OFFSET,	GenOffsetObject,	4,	{ POLY_EXPR | CURVE_EXPR | SURFACE_EXPR | TRIMSRF_EXPR, NUMERIC_EXPR | CURVE_EXPR, NUMERIC_EXPR, NUMERIC_EXPR }, POLY_EXPR | CURVE_EXPR | SURFACE_EXPR | TRIMSRF_EXPR },
    { "AOFFSET",  AOFFSET,	GenAOffsetObject,	5,	{ CURVE_EXPR, NUMERIC_EXPR | CURVE_EXPR, NUMERIC_EXPR, NUMERIC_EXPR, NUMERIC_EXPR }, CURVE_EXPR },
    { "LOFFSET",  LOFFSET,	GenLeastSqrOffsetObject,5,	{ CURVE_EXPR, NUMERIC_EXPR, NUMERIC_EXPR, NUMERIC_EXPR, NUMERIC_EXPR }, CURVE_EXPR },
    { "MOFFSET",  MOFFSET,	GenMatchingOffsetObject,3,	{ CURVE_EXPR, NUMERIC_EXPR, NUMERIC_EXPR },		CURVE_EXPR },
    { "TOFFSET",  TOFFSET,	TrimOffsetObject,	5,	{ CURVE_EXPR, CURVE_EXPR, NUMERIC_EXPR, NUMERIC_EXPR, NUMERIC_EXPR }, OLST_EXPR },
    { "COERCE",   COERCE,	CoerceIritObjectTo,	2,	{ ANY_EXPR, NUMERIC_EXPR },				ANY_EXPR },
    { "CEDITPT",  CEDITPT,	EditCrvControlPoint,	3,	{ CURVE_EXPR, CTLPT_EXPR, NUMERIC_EXPR },		CURVE_EXPR },
    { "SEDITPT",  SEDITPT,	EditSrfControlPoint,	4,	{ SURFACE_EXPR, CTLPT_EXPR, NUMERIC_EXPR, NUMERIC_EXPR }, SURFACE_EXPR },
    { "TEDITPT",  TEDITPT,	EditTVControlPoint,	5,	{ TRIVAR_EXPR, CTLPT_EXPR, NUMERIC_EXPR, NUMERIC_EXPR, NUMERIC_EXPR }, TRIVAR_EXPR },
    { "MERGEPOLY",MERGEPOLY,	GenObjectFromPolyList,	1,	{ OLST_EXPR },						POLY_EXPR },
    { "CMORPH",   CMORPH,	TwoCrvsMorphing,	4,	{ CURVE_EXPR, CURVE_EXPR, NUMERIC_EXPR, NUMERIC_EXPR },	CURVE_EXPR | OLST_EXPR },
    { "PMORPH",   PMORPH,	TwoPolysMorphing,	3,	{ POLY_EXPR, POLY_EXPR, NUMERIC_EXPR },			POLY_EXPR },
    { "SMORPH",   SMORPH,	TwoSrfsMorphing,	3,	{ SURFACE_EXPR, SURFACE_EXPR, NUMERIC_EXPR },		SURFACE_EXPR },
    { "TMORPH",   TMORPH,	TwoTVsMorphing,		3,	{ TRIVAR_EXPR, TRIVAR_EXPR, NUMERIC_EXPR },		TRIVAR_EXPR },
    { "BZR2BSP",  BZR2BSP,	CnvrtBezierToBspline,	1,	{ SURFACE_EXPR | CURVE_EXPR },				CURVE_EXPR | SURFACE_EXPR },
    { "BSP2BZR",  BSP2BZR,	CnvrtBsplineToBezier,	1,	{ SURFACE_EXPR | CURVE_EXPR },				CURVE_EXPR | SURFACE_EXPR },
    { "SMERGE",   SMERGE,	MergeSrfSrf,		4,	{ SURFACE_EXPR, SURFACE_EXPR, NUMERIC_EXPR, NUMERIC_EXPR }, SURFACE_EXPR },
    { "MMERGE",   MMERGE,	MergeMVMV,		4,	{ MULTIVAR_EXPR, MULTIVAR_EXPR, NUMERIC_EXPR, NUMERIC_EXPR },MULTIVAR_EXPR },
    { "CDERIVE",  CDERIVE,	DeriveCurveObject,	1,	{ CURVE_EXPR },						CURVE_EXPR },
    { "SDERIVE",  SDERIVE,	DeriveSurfaceObject,	2,	{ SURFACE_EXPR, NUMERIC_EXPR },				SURFACE_EXPR },
    { "TDERIVE",  TDERIVE,	DeriveTrivarObject,	2,	{ TRIVAR_EXPR, NUMERIC_EXPR },				TRIVAR_EXPR },
    { "MDERIVE",  MDERIVE,	DeriveMultivarObject,	2,	{ MULTIVAR_EXPR, NUMERIC_EXPR },			MULTIVAR_EXPR },
    { "TSDERIVE", TSDERIVE,	DeriveTriSrfObject,	2,	{ TRISRF_EXPR, NUMERIC_EXPR },				TRIVAR_EXPR },
    { "CINTEG",   CINTEG,	IntegrateCurveObject,	1,	{ CURVE_EXPR },						CURVE_EXPR },
    { "SINTEG",   SINTEG,	IntegrateSurfaceObject,	2,	{ SURFACE_EXPR, NUMERIC_EXPR },				SURFACE_EXPR },
    { "SNRMLSRF", SNRMLSRF,	SurfaceNormalObject,	1,	{ SURFACE_EXPR },					SURFACE_EXPR },
    { "CNRMLCRV", CNRMLCRV,	CurveNormalObject,	1,	{ CURVE_EXPR },						CURVE_EXPR },
    { "SYMBPROD", SYMBPROD,	TwoFreeformsProduct,	2,	{ CURVE_EXPR | SURFACE_EXPR | MULTIVAR_EXPR, CURVE_EXPR | SURFACE_EXPR | MULTIVAR_EXPR }, CURVE_EXPR | SURFACE_EXPR | MULTIVAR_EXPR },
    { "SYMBDPROD",SYMBDPROD,	TwoFreeformsDotProduct,	2,	{ VECTOR_EXPR | CURVE_EXPR | SURFACE_EXPR | MULTIVAR_EXPR, VECTOR_EXPR | CURVE_EXPR | SURFACE_EXPR | MULTIVAR_EXPR }, CURVE_EXPR | SURFACE_EXPR | MULTIVAR_EXPR },
    { "SYMBCPROD",SYMBCPROD,	TwoFreeformsCrossProduct,2,	{ CURVE_EXPR | SURFACE_EXPR | MULTIVAR_EXPR, CURVE_EXPR | SURFACE_EXPR | MULTIVAR_EXPR }, CURVE_EXPR | SURFACE_EXPR | MULTIVAR_EXPR },
    { "SYMBIPROD",SYMBIPROD,	TwoBasisInnerProduct,   3,	{ ANY_EXPR, NUMERIC_EXPR, NUMERIC_EXPR }, 		NUMERIC_EXPR },
    { "SYMBSUM",  SYMBSUM,	TwoFreeformsSum,	2,	{ CURVE_EXPR | SURFACE_EXPR | MULTIVAR_EXPR, CURVE_EXPR | SURFACE_EXPR | MULTIVAR_EXPR }, CURVE_EXPR | SURFACE_EXPR | MULTIVAR_EXPR },
    { "SYMBDIFF", SYMBDIFF,	TwoFreeformsDiff,	2,	{ CURVE_EXPR | SURFACE_EXPR | MULTIVAR_EXPR, CURVE_EXPR | SURFACE_EXPR | MULTIVAR_EXPR }, CURVE_EXPR | SURFACE_EXPR | MULTIVAR_EXPR },
    { "HOMOMAT",  HOMOMAT,	GenMatObjectGeneric,	1,	{ OLST_EXPR },						MATRIX_EXPR },
    { "PROJMAT",  PROJMAT,	GenMatProjectionMat,	3,	{ PLANE_EXPR, VECTOR_EXPR, NUMERIC_EXPR },		MATRIX_EXPR },
    { "RFLCTMAT", RFLCTMAT,	GenMatReflectionMat,	1,	{ PLANE_EXPR },						MATRIX_EXPR },
    { "MATPOSDIR",MATPOSDIR,	GenMatObjectPosDir,	3,	{ POINT_EXPR, VECTOR_EXPR, VECTOR_EXPR },		MATRIX_EXPR },
    { "MATDECOMP",MATDECOMP,	GetMatTransDecomp,	1,	{ MATRIX_EXPR },					OLST_EXPR },
    { "CINFLECT", CINFLECT,	CrvInflectionPts,	2,	{ CURVE_EXPR, NUMERIC_EXPR },				CURVE_EXPR | OLST_EXPR },
    { "CCRVTR",   CCRVTR,	CrvCurvaturePts,	2,	{ CURVE_EXPR, NUMERIC_EXPR },				CURVE_EXPR | OLST_EXPR },
    { "SCRVTR",   SCRVTR,	SrfCurvatureBounds,	3,	{ SURFACE_EXPR, NUMERIC_EXPR, NUMERIC_EXPR },		SURFACE_EXPR },
    { "PCRVTR",   PCRVTR,	PolyCurvatureApprox,	2,	{ POLY_EXPR, NUMERIC_EXPR },				POLY_EXPR },
    { "SGAUSS",   SGAUSS,	SrfGaussCurvature,	2,	{ SURFACE_EXPR, NUMERIC_EXPR },				SURFACE_EXPR },
    { "SMEAN",    SMEAN,	SrfMeanCurvature,	2,	{ SURFACE_EXPR, NUMERIC_EXPR },				SURFACE_EXPR },
    { "SUMBILIC", SUMBILIC,	SrfUmbilicPts,		3,	{ SURFACE_EXPR, NUMERIC_EXPR, NUMERIC_EXPR },		OLST_EXPR },
    { "CREPARAM", CREPARAM,	CrvReparametrization,	3,	{ CURVE_EXPR, NUMERIC_EXPR, NUMERIC_EXPR },		CURVE_EXPR },
    { "SREPARAM", SREPARAM,	SrfReparametrization,	4,	{ SURFACE_EXPR | TRIMSRF_EXPR, NUMERIC_EXPR, NUMERIC_EXPR, NUMERIC_EXPR }, SURFACE_EXPR | TRIMSRF_EXPR },
    { "TREPARAM", TREPARAM,	TrivReparametrization,  4,	{ TRIVAR_EXPR, NUMERIC_EXPR, NUMERIC_EXPR, NUMERIC_EXPR }, TRIVAR_EXPR },
    { "MREPARAM", MREPARAM,	MvarReparametrization,  4,	{ MULTIVAR_EXPR, NUMERIC_EXPR, NUMERIC_EXPR, NUMERIC_EXPR }, MULTIVAR_EXPR },
    { "EVOLUTE",  EVOLUTE,	FreeformEvolute,	1,	{ CURVE_EXPR | SURFACE_EXPR },				CURVE_EXPR | SURFACE_EXPR },
    { "CZEROS",   CZEROS,	CrvZeros,		3,	{ CURVE_EXPR, NUMERIC_EXPR, NUMERIC_EXPR },		OLST_EXPR },
    { "CEXTREMES",CEXTREMES,	CrvExtremes,		3,	{ CURVE_EXPR, NUMERIC_EXPR, NUMERIC_EXPR },		OLST_EXPR },
    { "PPINTER",  PPINTER,	PolyPolyIntersection,	2,	{ POLY_EXPR, POLY_EXPR },				POLY_EXPR },
    { "CCINTER",  CCINTER,	CrvCrvInter,		4,	{ CURVE_EXPR, CURVE_EXPR, NUMERIC_EXPR, NUMERIC_EXPR },	SURFACE_EXPR | OLST_EXPR },
    { "SSINTER",  SSINTER,	SrfSrfInter,		5,	{ SURFACE_EXPR, SURFACE_EXPR, NUMERIC_EXPR, NUMERIC_EXPR, NUMERIC_EXPR }, OLST_EXPR },
    { "NIL",      NIL,	        GetNilList,		0, 	{ 0 },							OLST_EXPR },
    { "COORD",    COORD,        GetObjectCoord,		2, 	{ ANY_EXPR, NUMERIC_EXPR },				ANY_EXPR },
    { "COMPOSE",  COMPOSE,      CrvComposition,		2, 	{ CURVE_EXPR | SURFACE_EXPR, CURVE_EXPR },		CURVE_EXPR },
    { "DECOMPOSE",DECOMPOSE,    CrvDecomposition,	1, 	{ CURVE_EXPR },						OLST_EXPR },
    { "PRISA",    PRISA,        SrfsPrisa,		6, 	{ SURFACE_EXPR | TRIMSRF_EXPR, NUMERIC_EXPR, NUMERIC_EXPR, NUMERIC_EXPR, VECTOR_EXPR, NUMERIC_EXPR, }, OLST_EXPR },
    { "CRVPTDST", CRVPTDIST,    CrvPointDist,		4, 	{ CURVE_EXPR, POINT_EXPR, NUMERIC_EXPR, NUMERIC_EXPR },	NUMERIC_EXPR | OLST_EXPR },
    { "CRVLNDST", CRVLNDIST,    CrvLineDist,		5, 	{ CURVE_EXPR, POINT_EXPR, VECTOR_EXPR, NUMERIC_EXPR, NUMERIC_EXPR },	  NUMERIC_EXPR | OLST_EXPR },
    { "ADAPISO",  ADAPISO,      SrfAdapIsoCurves,	5, 	{ SURFACE_EXPR, NUMERIC_EXPR, NUMERIC_EXPR, NUMERIC_EXPR, NUMERIC_EXPR }, OLST_EXPR | CURVE_EXPR },
    { "PDOMAIN",  PDOMAIN,	GetFreefromParamDomain,	1,	{ MULTIVAR_EXPR | TRISRF_EXPR | TRIVAR_EXPR | TRIMSRF_EXPR | SURFACE_EXPR | CURVE_EXPR }, OLST_EXPR },
    { "LINTERP",  LINTERP,	LineLeastSquarePtData,	1,	{ OLST_EXPR },						OLST_EXPR },
    { "CINTERP",  CINTERP,	CrvLeastSquarePtData,	5,	{ CURVE_EXPR | OLST_EXPR, NUMERIC_EXPR, NUMERIC_EXPR, NUMERIC_EXPR | OLST_EXPR, NUMERIC_EXPR }, CURVE_EXPR },
    { "SINTERP",  SINTERP,	SrfLeastSquarePtData,	6,	{ SURFACE_EXPR | OLST_EXPR, NUMERIC_EXPR, NUMERIC_EXPR, NUMERIC_EXPR, NUMERIC_EXPR, NUMERIC_EXPR }, SURFACE_EXPR | OLST_EXPR },
    { "CMULTIRES",CMULTIRES,	CurveMultiResDecomp,	3,	{ CURVE_EXPR, NUMERIC_EXPR, NUMERIC_EXPR },		CURVE_EXPR | OLST_EXPR },
    { "GETLINE",  GETLINE,	WndwInputStdinObject,	1,	{ NUMERIC_EXPR },					ANY_EXPR },
    { "FFEXTREME",FFEXTREME,	ExtremumControlPointVals,2,	{ CURVE_EXPR | SURFACE_EXPR, NUMERIC_EXPR },		CTLPT_EXPR },
    { "TRIMSRF",  TRIMSRF,      GenTrimmedSurface,      3,      { SURFACE_EXPR, CURVE_EXPR | OLST_EXPR, NUMERIC_EXPR },	TRIMSRF_EXPR },
    { "TRMSRFS",  TRMSRFS,      GenTrimmedSurfaces2,    2,      { SURFACE_EXPR, CURVE_EXPR | POLY_EXPR | OLST_EXPR },	OLST_EXPR | TRIMSRF_EXPR },
    { "STRIMSRF", SRFTRIMSRF,   GetSrfFromTrimmedSrf,   1,      { TRIMSRF_EXPR },					SURFACE_EXPR },
    { "CTRIMSRF", CRVTRIMCRV,   GetTrimCrvsFromTrimmedSrf,2,    { TRIMSRF_EXPR, NUMERIC_EXPR },				OLST_EXPR },
    { "TBEZIER",  TBEZIER,      GenBezierTrivarObject,  1,	{ OLST_EXPR },						TRIVAR_EXPR },
    { "TBSPLINE", TBSPLINE,     GenBsplineTrivarObject, 5,	{ NUMERIC_EXPR, NUMERIC_EXPR, NUMERIC_EXPR, OLST_EXPR, OLST_EXPR },	TRIVAR_EXPR },
    { "TSBEZIER", TSBEZIER,     GenBezierTriSrfObject,  2,	{ NUMERIC_EXPR, OLST_EXPR },				TRISRF_EXPR },
    { "TSBSPLINE",TSBSPLINE,    GenBsplineTriSrfObject, 4,	{ NUMERIC_EXPR, NUMERIC_EXPR, OLST_EXPR, OLST_EXPR },	TRISRF_EXPR },
    { "TSGREGORY",TSGREGORY,    GenGregoryTriSrfObject, 2,	{ NUMERIC_EXPR, OLST_EXPR },				TRISRF_EXPR },
    { "TEVAL",    TEVAL,        EvalTrivarObject,	4,	{ TRIVAR_EXPR, NUMERIC_EXPR, NUMERIC_EXPR, NUMERIC_EXPR }, CTLPT_EXPR },
    { "TSEVAL",   TSEVAL,       EvalTriSrfObject,	4,	{ TRISRF_EXPR, NUMERIC_EXPR, NUMERIC_EXPR, NUMERIC_EXPR }, CTLPT_EXPR },
    { "TEXTGEOM", TEXTGEOM,     GMMakeTextGeometry,	3,	{ STRING_EXPR, VECTOR_EXPR, NUMERIC_EXPR },		ANY_EXPR },
    { "STRIVAR",  STRIVAR,	SurfaceFromTrivar,	3,	{ TRIVAR_EXPR, NUMERIC_EXPR, NUMERIC_EXPR },		SURFACE_EXPR },
    { "SMESH",    SMESH,	SurfaceFromTrivMesh,	3,	{ TRIVAR_EXPR, NUMERIC_EXPR, NUMERIC_EXPR },		SURFACE_EXPR },
    { "TINTERP",  TINTERP,	InterpolateTrivar,	7,	{ TRIVAR_EXPR | OLST_EXPR, NUMERIC_EXPR, NUMERIC_EXPR, NUMERIC_EXPR, NUMERIC_EXPR, NUMERIC_EXPR, NUMERIC_EXPR }, TRIVAR_EXPR },
    { "CLNTREAD", CLNTREAD,	ClientRead,		2,	{ NUMERIC_EXPR, NUMERIC_EXPR },				ANY_EXPR },
    { "CLNTCRSR", CLNTCRSR,	ClientCursor,		1,	{ NUMERIC_EXPR },					OLST_EXPR },
    { "MOMENT",   MOMENT,	ComputeCrvMoments,	2,	{ CURVE_EXPR, NUMERIC_EXPR },				POINT_EXPR | VECTOR_EXPR },
    { "TFROMSRFS",TFROMSRFS,	GenTVFromSrfsObject,	2,	{ OLST_EXPR, NUMERIC_EXPR },				TRIVAR_EXPR },
    { "SFOCAL",   SFOCAL,	SrfIsoFocalSrf,		2,	{ SURFACE_EXPR, NUMERIC_EXPR },				SURFACE_EXPR },
    { "HERMITE",  HERMITE,	GenHermiteObject,	4,	{ CURVE_EXPR | POINT_EXPR | VECTOR_EXPR, CURVE_EXPR | POINT_EXPR | VECTOR_EXPR,
								  CURVE_EXPR | POINT_EXPR | VECTOR_EXPR, CURVE_EXPR | POINT_EXPR | VECTOR_EXPR }, CURVE_EXPR | SURFACE_EXPR },
    { "BLHERMITE",BLHERMITE,	GenBlendHermiteObject,	6,	{ CURVE_EXPR, CURVE_EXPR, CURVE_EXPR, CURVE_EXPR, CURVE_EXPR, CURVE_EXPR }, SURFACE_EXPR },
    { "BLSHERMITE",BLSHERMITE,	GenBlendHermiteOnSrfObject, 6,	{ SURFACE_EXPR, CURVE_EXPR, CURVE_EXPR, NUMERIC_EXPR, NUMERIC_EXPR, NUMERIC_EXPR | CURVE_EXPR }, SURFACE_EXPR },
    { "FFMATCH",  FFMATCH,      MatchTwoCurves,		7,      { CURVE_EXPR, CURVE_EXPR, NUMERIC_EXPR, NUMERIC_EXPR, NUMERIC_EXPR, NUMERIC_EXPR, NUMERIC_EXPR },	CURVE_EXPR },
    { "CONTOUR",  CONTOUR,      ContourFreeform,	ANY_PARAM_NUM, { SURFACE_EXPR, PLANE_EXPR, SURFACE_EXPR, OLST_EXPR }, POLY_EXPR },
    { "PRINTER",  PRINTER,      PolygonRayIntersect,	3, 	{ POLY_EXPR, POINT_EXPR, NUMERIC_EXPR },		NUMERIC_EXPR },
    { "SRINTER",  SRINTER,      SurfaceRayIntersect,	3, 	{ SURFACE_EXPR, POINT_EXPR, VECTOR_EXPR },		POINT_EXPR },
    { "PLN3PTS",  PLN3PTS,	PlaneFrom3Points,	3,	{ POINT_EXPR, POINT_EXPR, POINT_EXPR },			PLANE_EXPR },
    { "PTPTLN",   PTPTLN,	PointFromPointLine,	3,	{ POINT_EXPR, POINT_EXPR, VECTOR_EXPR },		POINT_EXPR },
    { "PTLNPLN",  PTLNPLN,	PointFromLinePlane,	3,	{ POINT_EXPR, VECTOR_EXPR, PLANE_EXPR },		POINT_EXPR },
    { "PTSLNLN",  PTSLNLN,	TwoPointsFromLineLine,	4,	{ POINT_EXPR, VECTOR_EXPR, POINT_EXPR, VECTOR_EXPR },	OLST_EXPR },
    { "PTSCRCR",  PTSCRCR,	TwoPointsFromCircCirc,	6,	{ POINT_EXPR, VECTOR_EXPR, NUMERIC_EXPR, POINT_EXPR, VECTOR_EXPR, NUMERIC_EXPR }, OLST_EXPR },
    { "TNSCRCR",  TNSCRCR,	TwoTangentsFromCircCirc,5,	{ POINT_EXPR, NUMERIC_EXPR, POINT_EXPR, NUMERIC_EXPR, NUMERIC_EXPR }, OLST_EXPR },
    { "PT3BARY",  PT3BARY,	BaryCentric3Pts,	4,	{ POINT_EXPR, POINT_EXPR, POINT_EXPR, POINT_EXPR },	POINT_EXPR },
    { "FFSPLIT",  FFSPLIT,	FreeFormSplitScalar,    1,	{ CURVE_EXPR | SURFACE_EXPR | MULTIVAR_EXPR },		OLST_EXPR },
    { "FFMERGE",  FFMERGE,	FreeFormMergeScalar,    2,	{ OLST_EXPR, NUMERIC_EXPR },				CURVE_EXPR | SURFACE_EXPR | MULTIVAR_EXPR },
    { "FFPTTYPE", FFPTTYPE,	FreeFormPointType,      1,	{ CURVE_EXPR | SURFACE_EXPR | TRIMSRF_EXPR | TRIVAR_EXPR | TRISRF_EXPR }, NUMERIC_EXPR },
    { "FFORDER",  FFORDER,	FreeFormOrder,          1,	{ CURVE_EXPR | SURFACE_EXPR | TRIMSRF_EXPR | TRIVAR_EXPR | TRISRF_EXPR }, OLST_EXPR },
    { "FFMSIZE",  FFMSIZE,	FreeFormMeshSize,       1,	{ CURVE_EXPR | SURFACE_EXPR | TRIMSRF_EXPR | TRIVAR_EXPR | TRISRF_EXPR }, OLST_EXPR },
    { "FFMESH",   FFMESH,	MeshGeometry2Polylines, 1,	{ CURVE_EXPR | SURFACE_EXPR | TRIMSRF_EXPR | TRIVAR_EXPR | TRISRF_EXPR }, POLY_EXPR },
    { "FFKNTVEC", FFKNTVEC,	FreeFormKnotVector,     1,	{ CURVE_EXPR | SURFACE_EXPR | TRIMSRF_EXPR | TRIVAR_EXPR | TRISRF_EXPR }, OLST_EXPR },
    { "FFCTLPTS", FFCTLPTS,	FreeFormControlPoints,  1,	{ CURVE_EXPR | SURFACE_EXPR | TRIMSRF_EXPR | TRIVAR_EXPR | TRISRF_EXPR }, OLST_EXPR },
    { "FFPOLES",  FFPOLES,	FreeFormPoles,		1,	{ CURVE_EXPR | SURFACE_EXPR | TRIMSRF_EXPR | TRIVAR_EXPR | TRISRF_EXPR }, NUMERIC_EXPR },
    { "CNVXHULL", CNVXHULL,     ComputeConvexHull,	2,	{ CURVE_EXPR | SURFACE_EXPR | OLST_EXPR | POLY_EXPR, NUMERIC_EXPR }, CURVE_EXPR | SURFACE_EXPR | OLST_EXPR | POLY_EXPR },
    { "MSCIRC",   MSCIRC,       ComputeMinSpanCirc,	1,	{ POLY_EXPR | OLST_EXPR },				CURVE_EXPR },
    { "CRVPTTAN", CRVPTTAN,     CrvPointTangents,	3,	{ CURVE_EXPR, POINT_EXPR, NUMERIC_EXPR },		OLST_EXPR },
    { "CRV2TANS", CRV2TANS,     CrvTwoTangents,		2,	{ CURVE_EXPR, NUMERIC_EXPR },				OLST_EXPR },
    { "CRC2CRVTAN",CRC2CRVTAN,  CircTangentsTwoCrvs,	4,	{ CURVE_EXPR, CURVE_EXPR, NUMERIC_EXPR, NUMERIC_EXPR },	OLST_EXPR },
    { "SRF3TANS", SRF3TANS,     SrfThreeTangents,	4,	{ OLST_EXPR, NUMERIC_EXPR, NUMERIC_EXPR, NUMERIC_EXPR },	OLST_EXPR },
    { "INSTANCE", INSTANCE,     GenerateInstance,	2,	{ STRING_EXPR, MATRIX_EXPR },				INSTANCE_EXPR },
    { "SVISIBLE", SVISIBLE,     VisibConeDecomposition,	3,	{ SURFACE_EXPR, NUMERIC_EXPR, NUMERIC_EXPR },		OLST_EXPR },
    { "GETATTR",  GETATTR,      GetObjectAttrib,	2,	{ ANY_EXPR, STRING_EXPR },				OLST_EXPR },
    { "PTHMSPR",  PTHMSPR,      PointCoverOfHemiSphere,	1,	{ NUMERIC_EXPR },					OLST_EXPR },
    { "PDECIMATE",PDECIMATE,    GenDecimatedObject,     3,      { POLY_EXPR, NUMERIC_EXPR, NUMERIC_EXPR },		POLY_EXPR },
    { "FFPTDIST", FFPTDIST,     FreeformPointDistrib,   3,      { CURVE_EXPR | SURFACE_EXPR, NUMERIC_EXPR, NUMERIC_EXPR }, OLST_EXPR },
    { "CENVOFF",  CENVOFF,      CurveEnvelopeOffset,    3,      { CURVE_EXPR, NUMERIC_EXPR, NUMERIC_EXPR },		SURFACE_EXPR },
    { "CBISECTOR",CBISECTOR,    CurveBisectorSkel,      5,      { CURVE_EXPR | OLST_EXPR, NUMERIC_EXPR, NUMERIC_EXPR, NUMERIC_EXPR, NUMERIC_EXPR },	CURVE_EXPR },
    { "SBISECTOR",SBISECTOR,    SurfaceBisectorSkel,    2,      { SURFACE_EXPR, POINT_EXPR },				SURFACE_EXPR },
    { "MBISECTOR",MBISECTOR,    MultivarBisector,       5,      { MULTIVAR_EXPR, MULTIVAR_EXPR, NUMERIC_EXPR, NUMERIC_EXPR, NUMERIC_EXPR }, MULTIVAR_EXPR },
    { "SPRBISECT",SPRBISECT,    SphericalBisector,      2,      { OLST_EXPR, NUMERIC_EXPR },				OLST_EXPR | CURVE_EXPR | SURFACE_EXPR },
    { "BSCTPLNPT",BSCTPLNPT,    PlanePointBisector,     2,      { POINT_EXPR, NUMERIC_EXPR },				OLST_EXPR | SURFACE_EXPR },
    { "BSCTCYLPT",BSCTCYLPT,    CylinPointBisector,     5,      { POINT_EXPR, VECTOR_EXPR, NUMERIC_EXPR, POINT_EXPR, NUMERIC_EXPR }, OLST_EXPR | SURFACE_EXPR },
    { "BSCTCONPT",BSCTCONPT,    ConePointBisector,      5,      { POINT_EXPR, VECTOR_EXPR, NUMERIC_EXPR, POINT_EXPR, NUMERIC_EXPR }, OLST_EXPR | SURFACE_EXPR },
    { "BSCTSPRPT",BSCTSPRPT,    SpherePointBisector,    3,      { POINT_EXPR, NUMERIC_EXPR, POINT_EXPR },		OLST_EXPR | SURFACE_EXPR },
    { "BSCTTRSPT",BSCTTRSPT,    TorusPointBisector,     5,      { POINT_EXPR, VECTOR_EXPR, NUMERIC_EXPR, NUMERIC_EXPR, POINT_EXPR }, OLST_EXPR | SURFACE_EXPR },
    { "BSCTPLNLN",BSCTPLNLN,    PlaneLineBisector,      2,      { VECTOR_EXPR, NUMERIC_EXPR },				OLST_EXPR | SURFACE_EXPR },
    { "BSCTCONLN",BSCTCONLN,    ConeLineBisector,       4,      { VECTOR_EXPR, NUMERIC_EXPR, VECTOR_EXPR, NUMERIC_EXPR }, OLST_EXPR | SURFACE_EXPR },
    { "BSCTSPRLN",BSCTSPRLN,    SphereLineBisector,     3,      { POINT_EXPR, NUMERIC_EXPR, NUMERIC_EXPR },		OLST_EXPR | SURFACE_EXPR },
    { "BSCTSPRPL",BSCTSPRPL,    SpherePlaneBisector,    3,      { POINT_EXPR, NUMERIC_EXPR, NUMERIC_EXPR },		OLST_EXPR | SURFACE_EXPR },
    { "BSCTCYLPL",BSCTCYLPL,    CylinPlaneBisector,     4,      { POINT_EXPR, VECTOR_EXPR, NUMERIC_EXPR, NUMERIC_EXPR }, OLST_EXPR | SURFACE_EXPR },
    { "BSCTCONPL",BSCTCONPL,    ConePlaneBisector,      4,      { POINT_EXPR, VECTOR_EXPR, NUMERIC_EXPR, NUMERIC_EXPR }, OLST_EXPR | SURFACE_EXPR },

    { "BSCTCONCON",BSCTCONCON,  ConeConeBisector,       5,      { VECTOR_EXPR, NUMERIC_EXPR, VECTOR_EXPR, NUMERIC_EXPR, NUMERIC_EXPR }, OLST_EXPR | SURFACE_EXPR },
    { "BSCTCONCN2",BSCTCONCN2,  ConeConeBisector2,      6,      { POINT_EXPR, VECTOR_EXPR, NUMERIC_EXPR, POINT_EXPR, VECTOR_EXPR, NUMERIC_EXPR }, SURFACE_EXPR },
    { "BSCTCONSPR",BSCTCONSPR,  ConeSphereBisector,     6,      { POINT_EXPR, VECTOR_EXPR, NUMERIC_EXPR, POINT_EXPR, NUMERIC_EXPR, NUMERIC_EXPR }, SURFACE_EXPR | SURFACE_EXPR },
    { "BSCTCYLSPR",BSCTCYLSPR,  CylinSphereBisector,    6,      { POINT_EXPR, VECTOR_EXPR, NUMERIC_EXPR, POINT_EXPR, NUMERIC_EXPR, NUMERIC_EXPR }, SURFACE_EXPR | SURFACE_EXPR },
    { "BSCTSPRSPR",BSCTSPRSPR,  SphereSphereBisector,   4,      { POINT_EXPR, NUMERIC_EXPR, POINT_EXPR, NUMERIC_EXPR },	OLST_EXPR | SURFACE_EXPR },
    { "BSCTTRSSPR",BSCTTRSSPR,  TorusSphereBisector,    6,      { POINT_EXPR, VECTOR_EXPR, NUMERIC_EXPR, NUMERIC_EXPR, POINT_EXPR, NUMERIC_EXPR }, SURFACE_EXPR },
    { "BSCTCYLCYL",BSCTCYLCYL,  CylinCylinBisector,     6,      { POINT_EXPR, VECTOR_EXPR, NUMERIC_EXPR, POINT_EXPR, VECTOR_EXPR, NUMERIC_EXPR }, SURFACE_EXPR },
    { "BSCTCONCYL",BSCTCONCYL,  ConeCylinBisector,      6,      { POINT_EXPR, VECTOR_EXPR, NUMERIC_EXPR, POINT_EXPR, VECTOR_EXPR, NUMERIC_EXPR }, SURFACE_EXPR },
    { "SKEL2DINT",SKEL2DINT,    Skel2D2PrimsInter,      6,      { POINT_EXPR | VECTOR_EXPR | CTLPT_EXPR | CURVE_EXPR,
								  POINT_EXPR | VECTOR_EXPR | CTLPT_EXPR | CURVE_EXPR,
								  POINT_EXPR | VECTOR_EXPR | CTLPT_EXPR | CURVE_EXPR, NUMERIC_EXPR, NUMERIC_EXPR, NUMERIC_EXPR }, OLST_EXPR },
    { "BBOX",	  BBOXOBJ,	ComputeBBOXObject,	1,	{ OLST_GEOM_EXPR },					OLST_EXPR },
    { "ORTHOTOMC",ORTHOTOMC,	FreeFormOrthotomic,	3,	{ CURVE_EXPR | SURFACE_EXPR, POINT_EXPR, NUMERIC_EXPR }, CURVE_EXPR | SURFACE_EXPR },
    { "BOUNDARY", BOUNDARY,	FreeFormBoundary,	1,	{ POLY_EXPR | SURFACE_EXPR | TRIMSRF_EXPR | TRIVAR_EXPR }, POLY_EXPR | CURVE_EXPR | SURFACE_EXPR },
    { "SILHOUETTE",SILHOUETTE,	FreeFormSilhouette,	3,	{ POLY_EXPR | SURFACE_EXPR, VECTOR_EXPR, NUMERIC_EXPR }, POLY_EXPR },
    { "SASPCTGRPH",SASPCTGRPH,	FreeFormTopoAspectGraph,1,	{ SURFACE_EXPR },					POLY_EXPR },
    { "SSILINFL", SSILINFL,	FreeFormSilhInflections,4,	{ SURFACE_EXPR, VECTOR_EXPR, NUMERIC_EXPR, NUMERIC_EXPR }, POLY_EXPR },
    { "POLARSIL", POLARSIL,	FreeFormPolarSilhouette,3,	{ SURFACE_EXPR, VECTOR_EXPR, NUMERIC_EXPR },		POLY_EXPR },
    { "ISOCLINE", ISOCLINE,	FreeFormIsocline,	4,	{ SURFACE_EXPR, VECTOR_EXPR, NUMERIC_EXPR, NUMERIC_EXPR }, POLY_EXPR },
    { "TCRVTR",   TCRVTR,	TrivIsoContourCurvature,3,	{ TRIVAR_EXPR, POINT_EXPR, NUMERIC_EXPR }, ANY_EXPR },
    { "MRCHCUBE", MRCHCUBE,	MarchCubeVolume,	4,	{ OLST_EXPR, POINT_EXPR, NUMERIC_EXPR, NUMERIC_EXPR },  POLY_EXPR },
    { "TRIANGL",  TRIANGL,	ConvertPolysTriangles,  2,	{ POLY_EXPR, NUMERIC_EXPR },				POLY_EXPR },
    { "MAXEDGELEN", MAXEDGELEN, LimitTrianglesEdgeLen,  2,      { POLY_EXPR, NUMERIC_EXPR },				POLY_EXPR },
    { "PTS2PLLN", PTS2PLLN,	ConvertPointsToPolys,   2,	{ OLST_EXPR, NUMERIC_EXPR },				POLY_EXPR },
    { "COVERPT",  COVERPT,	PointCoverPolyObj,	3,	{ POLY_EXPR, NUMERIC_EXPR, VECTOR_EXPR },	POLY_EXPR },
    { "COVERISO", COVERISO,	IsoCoverTVObj,		7,	{ TRIVAR_EXPR, NUMERIC_EXPR, NUMERIC_EXPR, VECTOR_EXPR, NUMERIC_EXPR, NUMERIC_EXPR, VECTOR_EXPR },	CURVE_EXPR },
    { "TVLOAD",   TVLOAD,	LoadVolumeIntoTV,	4,	{ STRING_EXPR, NUMERIC_EXPR, VECTOR_EXPR, VECTOR_EXPR }, TRIVAR_EXPR },
    { "DUALITY",  DUALITY,      FreeFormDuality,	1,	{ CURVE_EXPR | SURFACE_EXPR },				CURVE_EXPR | SURFACE_EXPR },
    { "ALGSUM",	  ALGSUM,	TwoCrvsAlgebraicSum,	2,	{ CURVE_EXPR, CURVE_EXPR },				CURVE_EXPR },
    { "SWUNGASUM",SWUNGASUM,	TwoCrvsSwungAlgSum,	2,	{ CURVE_EXPR, CURVE_EXPR },				CURVE_EXPR },
    { "MPROMOTE", MPROMOTE,	PromoteMultiVar,	2,	{ MULTIVAR_EXPR, OLST_EXPR },				MULTIVAR_EXPR },
    { "MREVERSE", MREVERSE,	ReverseMultiVar,	3,	{ MULTIVAR_EXPR, NUMERIC_EXPR, NUMERIC_EXPR },		MULTIVAR_EXPR },
    { "MFROMMV",  MFROMMV,	MultiVarFromMultiVar,	3,	{ MULTIVAR_EXPR, NUMERIC_EXPR, NUMERIC_EXPR },		MULTIVAR_EXPR },
    { "MFROMMESH",MFROMMESH,	MultiVarFromMVMesh,     3,	{ MULTIVAR_EXPR, NUMERIC_EXPR, NUMERIC_EXPR },		MULTIVAR_EXPR },
    { "MEVAL",    MEVAL,        EvalMultiVarObject,	2,	{ MULTIVAR_EXPR, OLST_EXPR },				CTLPT_EXPR },
    { "MZERO",    MZERO,        EvalMultiVarZeros,	3,	{ OLST_EXPR, NUMERIC_EXPR, NUMERIC_EXPR },		CTLPT_EXPR },
    { "SPARABOLC",SPARABOLC,    ParabolicCurvesOnSrf,	2,	{ SURFACE_EXPR,  NUMERIC_EXPR },			OLST_EXPR },
    { "RRINTER",  RRINTER,      RuledRuledInter,        6,      { CURVE_EXPR, CURVE_EXPR, CURVE_EXPR, CURVE_EXPR, NUMERIC_EXPR, NUMERIC_EXPR }, OLST_EXPR },
    { "GGINTER",  GGINTER,      RingRingInter,          6,      { CURVE_EXPR, CURVE_EXPR, CURVE_EXPR, CURVE_EXPR, NUMERIC_EXPR, NUMERIC_EXPR }, OLST_EXPR },
    { "SREVERSE", SREVERSE,	SurfaceReverse,		1,	{ SURFACE_EXPR | TRIMSRF_EXPR },			SURFACE_EXPR | TRIMSRF_EXPR },
    { "CONICSEC", CONICSEC,	GenConicSection,	4,	{ OLST_EXPR, NUMERIC_EXPR, NUMERIC_EXPR, NUMERIC_EXPR }, OLST_EXPR | CURVE_EXPR | SURFACE_EXPR },
    { "QUADRIC",  QUADRIC,	GenQuadric,	        1,	{ OLST_EXPR },						SURFACE_EXPR },
    { "IRITSTATE",IRITSTATE,	SetIritState,		2,	{ STRING_EXPR, NUMERIC_EXPR | STRING_EXPR },		NUMERIC_EXPR | STRING_EXPR },
    { "ISGEOM",   ISGEOM,	WasIsGeometry,		3,	{ ANY_EXPR, NUMERIC_EXPR, NUMERIC_EXPR },		OLST_EXPR },
    { "SMOOTHNRML",SMOOTHNRML,  ComputeSmoothPolyNormals,2,     { OLST_EXPR | POLY_EXPR, NUMERIC_EXPR },		OLST_EXPR | POLY_EXPR },
    { "FIXPLNRML",FIXPLNRML,    FixPolyNormals,		2,      { OLST_EXPR | POLY_EXPR, NUMERIC_EXPR },		OLST_EXPR | POLY_EXPR },
    { "FIXPLGEOM",FIXPLGEOM,    FixPolyGeometry,	3,      { OLST_EXPR | POLY_EXPR, NUMERIC_EXPR, NUMERIC_EXPR },	OLST_EXPR | POLY_EXPR },
    { "BLOSSOM",  BLOSSOM,      BlossomEvaluation,      2,      { CURVE_EXPR, OLST_EXPR },				CTLPT_EXPR },
    { "KNOTCLEAN",KNOTCLEAN,    CleanRefinedKnots,	2,      { CURVE_EXPR, NUMERIC_EXPR },				CURVE_EXPR },
    { "KNOTREMOVE",KNOTREMOVE,  RemoveKnots,		4,      { CURVE_EXPR, NUMERIC_EXPR, NUMERIC_EXPR, NUMERIC_EXPR }, CURVE_EXPR },
    { "RAYTRAPS", RAYTRAPS,     MultivarRayTrap,	3,      { OLST_EXPR, NUMERIC_EXPR, NUMERIC_EXPR },		OLST_EXPR },
    { "CMOEBIUS", CMOEBIUS,     MoebiusCurveTrans,	2,      { CURVE_EXPR, NUMERIC_EXPR },				CURVE_EXPR },
    { "SMOEBIUS", SMOEBIUS,     MoebiusSurfaceTrans,	3,      { SURFACE_EXPR, NUMERIC_EXPR, NUMERIC_EXPR },		SURFACE_EXPR },
    { "RFLCTLN",  RFLCTLN,      ReflectionLines,	4,      { SURFACE_EXPR, VECTOR_EXPR, ANY_EXPR, NUMERIC_EXPR },	OLST_EXPR },
    { "CAREA",    CAREA,        CrvAreaIntergal,	1,      { CURVE_EXPR },						CURVE_EXPR },
    { "SVOLUME",  SVOLUME,      SrfVolumeIntergal,	3,      { SURFACE_EXPR, NUMERIC_EXPR, NUMERIC_EXPR },		SURFACE_EXPR | NUMERIC_EXPR },
    { "SMOMENTS", SMOMENTS,     SrfMomentIntergal,	5,      { SURFACE_EXPR, NUMERIC_EXPR, NUMERIC_EXPR, NUMERIC_EXPR, NUMERIC_EXPR }, SURFACE_EXPR | NUMERIC_EXPR },
    { "TVZRJACOB",TVZRJACOB,	TrivZeroJacobian,	4,	{ TRIVAR_EXPR, NUMERIC_EXPR, NUMERIC_EXPR, NUMERIC_EXPR | OLST_EXPR }, POLY_EXPR },
    { "CRVKERNEL",CRVKERNEL,	CrvKernel,		5,	{ CURVE_EXPR, NUMERIC_EXPR, NUMERIC_EXPR, NUMERIC_EXPR | OLST_EXPR, NUMERIC_EXPR },POLY_EXPR | TRIVAR_EXPR },
    { "SRFKERNEL",SRFKERNEL,	SrfKernel,		3,	{ SURFACE_EXPR, NUMERIC_EXPR, NUMERIC_EXPR },		POLY_EXPR },
    { "CRVDIAMTR",CRVDIAMTR,	CrvDiameter,		4,	{ CURVE_EXPR, NUMERIC_EXPR, NUMERIC_EXPR, NUMERIC_EXPR },POINT_EXPR | OLST_EXPR },
    { "SACCESS",  SACCESS,	SurfaceAccessibility,	6,	{ SURFACE_EXPR, ANY_EXPR, SURFACE_EXPR, ANY_EXPR, NUMERIC_EXPR, NUMERIC_EXPR }, POLY_EXPR },
    { "SFLECNODAL",SFLECNODAL,  SurfaceFlecnodals,      5,      { SURFACE_EXPR, NUMERIC_EXPR, NUMERIC_EXPR, NUMERIC_EXPR, NUMERIC_EXPR }, POLY_EXPR },
    { "SRF2TANS", SRF2TANS,     SurfaceBiTangents,      5,      { SURFACE_EXPR | OLST_EXPR, NUMERIC_EXPR, NUMERIC_EXPR, NUMERIC_EXPR, NUMERIC_EXPR }, POLY_EXPR },
    { "NRMLCONE", NRMLCONE,	EvalSrfNormalsCone,	1,	{ SURFACE_EXPR },					OLST_EXPR },
    { "PTREGISTER",PTREGISTER,	RegisterPointSet,	4,	{ OLST_EXPR, OLST_EXPR | SURFACE_EXPR, NUMERIC_EXPR, NUMERIC_EXPR }, MATRIX_EXPR },
    { "GETNAME",  GETNAME,	GetSubObjectName,	2,	{ OLST_EXPR, NUMERIC_EXPR },				STRING_EXPR },
    { "BUMPMAP",  SBUMPMAP,	EvalBumpMapForSrf,	5,	{ SURFACE_EXPR, ANY_EXPR, NUMERIC_EXPR, NUMERIC_EXPR, NUMERIC_EXPR }, ANY_EXPR },
    { "MVEXPLICIT",MVEXPLICIT, 	InptEvalMVExplicit,	1,	{ STRING_EXPR },					MULTIVAR_EXPR },
    { "ANALYFIT", ANALYFIT, 	AnalyticSrfFit,		4,	{ OLST_EXPR, OLST_EXPR, NUMERIC_EXPR, NUMERIC_EXPR },	OLST_EXPR },
    { "PLANECLIP",PLANECLIP, 	PolyPlaneClipping,	2,	{ POLY_EXPR, PLANE_EXPR },				OLST_EXPR },
    { "SRFFFORM", SRFFFORM, 	SrfFundamentalForms,	2,	{ SURFACE_EXPR, NUMERIC_EXPR },				OLST_EXPR },
    { "SCRVTREVAL",SCRVTREVAL,  EvalSurfaceCurvature,   4,      { SURFACE_EXPR, NUMERIC_EXPR, NUMERIC_EXPR, NUMERIC_EXPR }, OLST_EXPR },
    { "SASYMPEVAL",SASYMPEVAL,  EvalSurfaceAsympDir,    4,      { SURFACE_EXPR, NUMERIC_EXPR, NUMERIC_EXPR, NUMERIC_EXPR }, OLST_EXPR },
    { "CARCLEN",  CARCLEN,	CrvArcLenApprox,        3,      { CURVE_EXPR, NUMERIC_EXPR, NUMERIC_EXPR },		CURVE_EXPR },
    { "TEXTWARP", TEXTWARP,     TextWarpThroughSrf,     6,      { SURFACE_EXPR, STRING_EXPR, NUMERIC_EXPR, NUMERIC_EXPR, NUMERIC_EXPR, NUMERIC_EXPR }, OLST_EXPR },
    { "SRAYCLIP", SRAYCLIP,     BezierRayClipping,      3,	{ POINT_EXPR, VECTOR_EXPR, SURFACE_EXPR },		OLST_EXPR },
    { "PPROPFTCH",PPROPFTCH,	PolyPropFetch,		3,	{ POLY_EXPR, NUMERIC_EXPR, OLST_EXPR },			POLY_EXPR },
    { "FITPMODEL",FITPMODEL,    FitPrim2PolyModel,      4,      { POLY_EXPR, NUMERIC_EXPR, NUMERIC_EXPR, NUMERIC_EXPR }, OLST_EXPR },
    { "UVPOLY",   UVPOLY,       SetUVsToPolys,          2,      { POLY_EXPR, OLST_EXPR }, 				POLY_EXPR },
    { "MPOWER",   MPOWER,       GenPowerMVObject,       2,      { OLST_EXPR, OLST_EXPR }, 				MULTIVAR_EXPR },
    { "MBEZIER",  MBEZIER,      GenBezierMVObject,      2,      { OLST_EXPR, OLST_EXPR }, 				MULTIVAR_EXPR },
    { "MBSPLINE", MBSPLINE,     GenBsplineMVObject,     4,      { OLST_EXPR, OLST_EXPR, OLST_EXPR, OLST_EXPR }, 	MULTIVAR_EXPR },
    { "CBIARCS",  CBIARCS,      CrvBiArcApprox,         3,      { CURVE_EXPR, NUMERIC_EXPR, NUMERIC_EXPR },	 	OLST_EXPR },
    { "SPLITLST", SPLITLST,	SplitLnkList2ObjList,   1,      { ANY_EXPR },                                           OLST_EXPR },
    { "DIST2FF",  DIST2FF,	DistanceTwoFreeforms,   3,      { CURVE_EXPR | SURFACE_EXPR, CURVE_EXPR | SURFACE_EXPR, NUMERIC_EXPR }, SURFACE_EXPR | MULTIVAR_EXPR }
};
int ObjFuncTableSize = sizeof(ObjFuncTable) / sizeof(ObjFuncTableType);

GenFuncTableType GenFuncTable[] = {
    { "EXIT",	EXIT,		IritExit0,		0,	{ 0 } },
    { "RESET",  IRITRESET,	IritReset,		0,	{ 0 } },
    { "VIEWOBJ",VIEWOBJ,	WndwViewObject,		1,	{ ANY_EXPR } },
    { "VIEWSET",VIEWSET,	WndwSetCrntDisplay,	1,	{ NUMERIC_EXPR } },
    { "CHDIR",	CHDIR,		DosChangeDir,		1,	{ STRING_EXPR } },
    { "INCLUDE",INCLUDE,	InptPrsrFileInclude,	1,	{ STRING_EXPR } },
    { "SAVE",	SAVE,		SaveObjectInFile,	2,	{ STRING_EXPR, ANY_EXPR } },
    { "FREE",	FREEOBJ,	FreeIritObject,		1,	{ ANY_EXPR } },
    { "FNFREE",	FREEFUNC,	InptEvalFreeFunc,	1,	{ STRING_EXPR } },
    { "IF",	IFCOND, 	InptEvalIfCondition,	ANY_PARAM_NUM, { 0 } },
    { "FOR",	FORLOOP, 	InptEvalForLoop,	4,	{ NUMERIC_EXPR,	NUMERIC_EXPR, NUMERIC_EXPR, ANY_EXPR } },
    { "WHILE",  WHILELOOP,      InptEvalWhileLoop,      2,      { NUMERIC_EXPR, ANY_EXPR } },
    { "HELP",	PRHELP,		InptEvalPrintHelp,	1,	{ STRING_EXPR } },
    { "VARLIST",VARLIST, 	PrintIritObjectList,	0,	{ 0 } },
    { "SYSTEM",	SYSTEM,		DosSystem,		1,	{ STRING_EXPR } },
    { "LOGFILE",LOGFILE, 	WndwLogPrint,		1,	{ STRING_EXPR | NUMERIC_EXPR } },
    { "COLOR",	COLOR,		SetObjectAttrColor, 	2,	{ OLST_GEOM_EXPR, NUMERIC_EXPR } },
    { "AWIDTH",	AWIDTH,		SetObjectAttrWidth, 	2,	{ OLST_GEOM_EXPR, NUMERIC_EXPR } },
    { "ADWIDTH",ADWIDTH,	SetObjectAttrDWidth, 	2,	{ OLST_GEOM_EXPR, NUMERIC_EXPR } },
    { "SNOC",	SNOC,		SnocList,		2,	{ ANY_EXPR, OLST_EXPR } },
    { "ATTRIB",	ATTRIB,		SetObjectAttrib,	3,	{ ANY_EXPR, STRING_EXPR, ANY_EXPR } },
    { "ATTRPROP",ATTRPROP,	SetObjectAttribProp,	3,	{ ANY_EXPR, STRING_EXPR, ANY_EXPR } },
    { "CPATTR",	CPATTR,		CopyObjectAllAttribs,	2,	{ ANY_EXPR, ANY_EXPR } },
    { "RMATTR",	RMATTR,		RemoveObjectAttrib,	2,	{ ANY_EXPR, STRING_EXPR } },
    { "FFCOMPAT",FFCOMPAT,	MakeFreeFormCompatible,	2,	{ CURVE_EXPR | SURFACE_EXPR | TRIVAR_EXPR, CURVE_EXPR | SURFACE_EXPR | TRIVAR_EXPR } },
    { "MSLEEP",  MSLEEP,	MilisecondSleep,	1,	{ NUMERIC_EXPR } },
    { "PRINTF", IRITPRINT,	IritObjectPrintf,	2,	{ STRING_EXPR, OLST_EXPR } },
    { "PRINTFILE",IRITPRNTFL,	IritObjectPrintfSetFile,1,	{ STRING_EXPR } },
    { "ERROR",  IRITERROR,      (void (*)()) IritNonFatalError,	1, { STRING_EXPR } },
    { "CLNTWRITE",CLNTWRITE,	ClientWrite,		2,	{ NUMERIC_EXPR, ANY_EXPR } },
    { "CLNTCLOSE",CLNTCLOSE,	ClientClose,		2,	{ NUMERIC_EXPR, NUMERIC_EXPR } },
    { "SETNAME",SETNAME,	SetSubObjectName,	3,	{ OLST_EXPR, NUMERIC_EXPR, STRING_EXPR } },
    { "EXEC",   EXEC,		ExecOneLine,		1,	{ STRING_EXPR } },
    { "ANIMEVAL",ANIMEVAL,	EvalAnimationTime,	2,	{ NUMERIC_EXPR, ANY_EXPR } },
    { "IQUERY", IQUERY,		IritQuery,		1,	{ NUMERIC_EXPR } },
    { "INSERTPOLY",INSERTPOLY,	(void (*)()) InsertPolyToPoly,	2, { POLY_EXPR, POLY_EXPR } }

#ifdef DEBUG_IRIT_MALLOC
    , { "IDYNMEM",IDYNMEM,	IritCheckMarkDynMemory,	1,	{ NUMERIC_EXPR } }
#endif /* DEBUG_IRIT_MALLOC */
};
int GenFuncTableSize = sizeof(GenFuncTable) / sizeof(GenFuncTableType);

ConstantTableType ConstantTable[] = {
    { "PI",	M_PI },

    { "ON",	1.0 },
    { "TRUE",	1.0 },
    { "OFF",	0.0 },
    { "FALSE",	0.0 },

    { "COL",    (double) CAGD_CONST_U_DIR },
    { "ROW",    (double) CAGD_CONST_V_DIR },
    { "DEPTH",  (double) TRIV_CONST_W_DIR },

    { "CLIENTS_ALL", (double) IP_CLNT_BROADCAST_ALL_HANDLES },

    { "KV_OPEN", (double) KV_UNIFORM_OPEN },
    { "KV_DISC_OPEN", (double) KV_UNIFORM_DISCONT_OPEN },
    { "KV_FLOAT", (double) KV_UNIFORM_FLOAT },
    { "KV_PERIODIC", (double) KV_UNIFORM_PERIODIC },

    { "PARAM_NIELFOL", (double) CAGD_NIELSON_FOLEY_PARAM },
    { "PARAM_CHORD", (double) CAGD_CHORD_LEN_PARAM },
    { "PARAM_CENTRIP", (double) CAGD_CENTRIPETAL_PARAM },
    { "PARAM_UNIFORM", (double) CAGD_UNIFORM_PARAM },

    { "E1", (double) CAGD_PT_E1_TYPE },
    { "E2", (double) CAGD_PT_E2_TYPE },
    { "E3", (double) CAGD_PT_E3_TYPE },
    { "E4", (double) CAGD_PT_E4_TYPE },
    { "E5", (double) CAGD_PT_E5_TYPE },
    { "E6", (double) CAGD_PT_E6_TYPE },
    { "E7", (double) CAGD_PT_E7_TYPE },
    { "E8", (double) CAGD_PT_E8_TYPE },
    { "E9", (double) CAGD_PT_E9_TYPE },

    { "P1", (double) CAGD_PT_P1_TYPE },
    { "P2", (double) CAGD_PT_P2_TYPE },
    { "P3", (double) CAGD_PT_P3_TYPE },
    { "P4", (double) CAGD_PT_P4_TYPE },
    { "P5", (double) CAGD_PT_P5_TYPE },
    { "P6", (double) CAGD_PT_P6_TYPE },
    { "P7", (double) CAGD_PT_P7_TYPE },
    { "P8", (double) CAGD_PT_P8_TYPE },
    { "P9", (double) CAGD_PT_P9_TYPE },

    { "UNDEF_TYPE",   (double) IP_OBJ_UNDEF },
    { "POLY_TYPE",    (double) IP_OBJ_POLY },
    { "NUMERIC_TYPE", (double) IP_OBJ_NUMERIC },
    { "POINT_TYPE",   (double) IP_OBJ_POINT },
    { "VECTOR_TYPE",  (double) IP_OBJ_VECTOR },
    { "PLANE_TYPE",   (double) IP_OBJ_PLANE },
    { "MATRIX_TYPE",  (double) IP_OBJ_MATRIX },
    { "CURVE_TYPE",   (double) IP_OBJ_CURVE },
    { "SURFACE_TYPE", (double) IP_OBJ_SURFACE },
    { "STRING_TYPE",  (double) IP_OBJ_STRING },
    { "CTLPT_TYPE",   (double) IP_OBJ_CTLPT },
    { "LIST_TYPE",    (double) IP_OBJ_LIST_OBJ },
    { "TRIVAR_TYPE",  (double) IP_OBJ_TRIVAR },
    { "TRISRF_TYPE",  (double) IP_OBJ_TRISRF },
    { "TRIMSRF_TYPE", (double) IP_OBJ_TRIMSRF },
    { "MODEL_TYPE",   (double) IP_OBJ_MODEL },
    { "MULTIVAR_TYPE",(double) IP_OBJ_MULTIVAR },

    { "BEZIER_TYPE", (double) FF_BEZIER_TYPE },
    { "BSPLINE_TYPE",(double) FF_BSPLINE_TYPE },
    { "POWER_TYPE",  (double) FF_POWER_TYPE },
    { "GREGORY_TYPE",(double) FF_GREGORY_TYPE },

    { "GEOM_CONST",      (double) CAGD_GEOM_CONST },
    { "GEOM_LINEAR",     (double) CAGD_GEOM_LINEAR },
    { "GEOM_CIRCULAR",   (double) CAGD_GEOM_CIRCULAR },
    { "GEOM_PLANAR",     (double) CAGD_GEOM_PLANAR },
    { "GEOM_SPHERICAL",  (double) CAGD_GEOM_SPHERICAL },
    { "GEOM_SRF_OF_REV", (double) CAGD_GEOM_SRF_OF_REV },
    { "GEOM_EXTRUSION",  (double) CAGD_GEOM_EXTRUSION },
    { "GEOM_RULED_SRF",  (double) CAGD_GEOM_RULED_SRF },
    { "GEOM_DEVELOP_SRF",(double) CAGD_GEOM_DEVELOP_SRF },
    { "GEOM_SWEEP",      (double) CAGD_GEOM_SWEEP_SRF },

    { "BLACK",  (double) IG_IRIT_BLACK },
    { "BLUE",	(double) IG_IRIT_BLUE },
    { "GREEN",	(double) IG_IRIT_GREEN },
    { "CYAN",	(double) IG_IRIT_CYAN },
    { "RED",	(double) IG_IRIT_RED },
    { "MAGENTA",(double) IG_IRIT_MAGENTA },
    { "YELLOW", (double) IG_IRIT_YELLOW },
    { "WHITE",  (double) IG_IRIT_WHITE },

    { "MSDOS",  (double) MACHINE_MSDOS },
    { "SGI",    (double) MACHINE_SGI },
    { "HP",     (double) MACHINE_HP },
    { "SUN",    (double) MACHINE_SUN },
    { "APOLLO", (double) MACHINE_APOLLO },
    { "UNIX",   (double) MACHINE_UNIX },
    { "IBMOS2", (double) MACHINE_IBMOS2 },
    { "IBMNT",  (double) MACHINE_IBMNT },
    { "AMIGA",  (double) MACHINE_AMIGA },
};
int ConstantTableSize = sizeof(ConstantTable) / sizeof(ConstantTableType);

UserDefinedFuncDefType
    *UserDefinedFuncList = NULL;

#ifndef IRIT_QUIET_STRINGS

/*****************************************************************************
* DESCRIPTION:                                                               *
*   Prints to stdout the definition of all available functions in IRIT.      *
*                                                                            *
* PARAMETERS:                                                                *
*   None                                                                     *
*                                                                            *
* RETURN VALUE:                                                              *
*   void                                                                     *
*****************************************************************************/
static void IritQueryFunctions(void)
{
    int i, l;
    UserDefinedFuncDefType *UserFunc;

    for (i = 0; i < NumFuncTableSize; i++) {
	NumFuncTableType
	    *NumEntry = &NumFuncTable[i];
	
	printf("%sDEFFUNC 00002 %-10s %3d [", GUI_PROMPT,
	       NumEntry -> FuncName, NumEntry -> NumOfParam);
	if (NumEntry -> NumOfParam != ANY_PARAM_NUM) {
	    for (l = 0; l < NumEntry -> NumOfParam; l++)
		printf("%s%05x",
		       l == 0 ? "" : " ", NumEntry -> ParamObjType[l]);
	}
	printf("]\n");
    }

    for (i = 0; i < ObjFuncTableSize; i++) {
	ObjFuncTableType
	    *ObjEntry = &ObjFuncTable[i];
	
	printf("%sDEFFUNC %05x %-10s %3d [", GUI_PROMPT, ObjEntry -> RetType,
	       ObjEntry -> FuncName, ObjEntry -> NumOfParam);
	if (ObjEntry -> NumOfParam != ANY_PARAM_NUM) {
	    for (l = 0; l < ObjEntry -> NumOfParam; l++)
		printf("%s%05x",
		       l == 0 ? "" : " ", ObjEntry -> ParamObjType[l]);
	}
	printf("]\n");
    }

    for (i = 0; i < GenFuncTableSize; i++) {
	GenFuncTableType
	    *GenEntry = &GenFuncTable[i];
	
	printf("%sDEFFUNC 00000 %-10s %3d [", GUI_PROMPT,
	       GenEntry -> FuncName, GenEntry -> NumOfParam);
	if (GenEntry -> NumOfParam != ANY_PARAM_NUM) {
	    for (l = 0; l < GenEntry -> NumOfParam; l++)
		printf("%s%05x",
		       l == 0 ? "" : " ", GenEntry -> ParamObjType[l]);
	}
	printf("]\n");
    }

    for (UserFunc = UserDefinedFuncList;
	 UserFunc != NULL;
	 UserFunc = UserFunc -> Pnext) {
	int n = IPObjListLen(UserFunc -> Params);

	printf("%sDEFUFUNC %05x %-10s %3d [", GUI_PROMPT,
	       UserFunc -> IsFunction ? ANY_EXPR : 0,
	       UserFunc -> FuncName, n);
	for (l = 0; l < n; l++)
	    printf("%s%05x", l == 0 ? "" : " ", ANY_EXPR);
	printf("]\n");
    }

    for (i = 0; i < ConstantTableSize; i++) {
	ConstantTableType
	    *GenEntry = &ConstantTable[i];
	
	printf("%sDEFCONST %-10s %-17.15f\n", GUI_PROMPT,
	       GenEntry -> FuncName, GenEntry -> Value);
    }
}

/*****************************************************************************
* DESCRIPTION:                                                               *
*   Prints to stdout the definition of all available functions in IRIT.      *
*                                                                            *
* PARAMETERS:                                                                *
*   None                                                                     *
*                                                                            *
* RETURN VALUE:                                                              *
*   void                                                                     *
*****************************************************************************/
static void IritQueryKeywords(void)
{
    int i;

    for (i = 0; i < NumFuncTableSize; i++)
	printf("%s\n", NumFuncTable[i].FuncName);

    for (i = 0; i < ObjFuncTableSize; i++)
	printf("%s\n", ObjFuncTable[i].FuncName);

    for (i = 0; i < GenFuncTableSize; i++)
	printf("%s\n", GenFuncTable[i].FuncName);

    for (i = 0; i < ConstantTableSize; i++)
	printf("%s\n", ConstantTable[i].FuncName);
}

#endif /* !IRIT_QUIET_STRINGS */

/*****************************************************************************
* DESCRIPTION:                                                               *
*   Prints to stdout queries on the IRIT state.                              *
*                                                                            *
* PARAMETERS:                                                                *
*   RealType *:  Type of query.                                              *
*                                                                            *
* RETURN VALUE:                                                              *
*   void                                                                     *
*****************************************************************************/
static void IritQuery(RealType *RQuery)
{
#ifndef IRIT_QUIET_STRINGS
    int Query = REAL_PTR_TO_INT(RQuery);

    switch (Query) {
	case 1:
	    IritQueryFunctions();
            break;
	case 2:
	    IritQueryKeywords();
            break;
    }
#endif /* !IRIT_QUIET_STRINGS */
}
