/*****************************************************************************
* Writew one image out.							     *
******************************************************************************
* (C) Gershon Elber, Technion, Israel Institute of Technology                *
******************************************************************************
* Written by:  Gershon Elber			       Ver 1.0,	May. 1999    *
*****************************************************************************/

#include <math.h>
#include "irit_sm.h"
#include "misc_lib.h"

#ifdef __WINNT__
    #include <io.h>
    #include <fcntl.h>
#endif /* __WINNT__ */

#ifdef HAVE_URT_RLE
#define NO_DECLARE_MALLOC /* For rle.h */
#define USE_PROTOTYPES
#define USE_STDLIB_H
#include <rle.h>
#include <rle_raw.h>
STATIC_DATA rle_hdr *Header;
STATIC_DATA rle_pixel **Rows;
#endif /* HAVE_URT_RLE */

STATIC_DATA IrtImgImageType
    IType = IRIT_IMAGE_PPM6_TYPE;
STATIC_DATA FILE
    *PPMFile = NULL;
STATIC_DATA int PPMWidth, PPMHeight, PPMLine;
STATIC_DATA IrtImgPixelStruct **PixelData;

static void PPMOpenFile(char **argv, char *FName, int XSize, int YSize);
static void PPMPutLine(ByteType *Alpha, IrtImgPixelStruct *Pixels);
static void PPMCloseFile(void);
static void RLEOpenFile(char **argv,
			char *FName,
			int Alpha,
			int XSize,
			int YSize);
static void RLEPutLine(ByteType *Alpha, IrtImgPixelStruct *Pixels);
static void RLECloseFile(void);
static void RLEDummyLink(void);

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Sets image type.                                                         M
*                                                                            *
* PARAMETERS:                                                                M
*   Image:  A string describing the image type.                              M
*                                                                            *
* RETURN VALUE:                                                              M
*   IrtImgImageType:  Returns the detected type.                             M
*                                                                            *
* KEYWORDS:                                                                  M
*   IrtImgWriteSetType                                                       M
*****************************************************************************/
IrtImgImageType IrtImgWriteSetType(char *Image)
{
    if (Image == NULL || strlen(Image) == 0) {
	IRIT_FATAL_ERROR("Empty image file name to write to.");
	return IType;
    }

    if (stricmp(Image, "rle") == 0)
        IType = IRIT_IMAGE_RLE_TYPE;
    else if (strnicmp(Image, "ppm", 3) == 0) {
	switch (Image[3]) {
	    case '3':
		IType = IRIT_IMAGE_PPM3_TYPE;
	        break;
	    default:
	    case '6':
		IType = IRIT_IMAGE_PPM6_TYPE;
	        break;
	}
    }
    else
        IType = IRIT_IMAGE_UNKNOWN_TYPE;

    return IType;
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Opens output image file.                                                 M
*                                                                            *
* PARAMETERS:                                                                M
*   argv:     Pointer to the name of this program.                           M
*   FName:    Filename to open, or NULL for stdout.			     M
*   Alpha:    Do we have aan alpha channel.				     M
*   XSize:    X dimension of the image.                                      M
*   YSize:    Y dimension of the image.                                      M
*                                                                            *
* RETURN VALUE:                                                              M
*   void                                                                     M
*                                                                            *
* KEYWORDS:                                                                  M
*   IrtImgWriteOpenFile, output image                                        M
*****************************************************************************/
void IrtImgWriteOpenFile(char **argv,
			 char *FName,
			 int Alpha,
			 int XSize,
			 int YSize)
{
    switch (IType) {
	default:
	case IRIT_IMAGE_RLE_TYPE:
	    RLEOpenFile(argv, FName, Alpha, XSize, YSize);
	    break;
	case IRIT_IMAGE_PPM3_TYPE:
	case IRIT_IMAGE_PPM6_TYPE:
	    PPMOpenFile(argv, FName, XSize, YSize);
	    break;
    }
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Outputs given line of color pixels and alpha correction values to the    M
*   ouput image file.                                                        M
*                                                                            *
* PARAMETERS:                                                                M
*   Alpha:  array of alpha values.                                           M
*   Pixels: array of color pixels.                                           M
*                                                                            *
* RETURN VALUE:                                                              M
*   void                                                                     M
*                                                                            *
* KEYWORDS:                                                                  M
*   IrtImgWritePutLine, output image                                         M
*****************************************************************************/
void IrtImgWritePutLine(ByteType *Alpha, IrtImgPixelStruct *Pixels)
{
    switch (IType) {
	default:
	case IRIT_IMAGE_RLE_TYPE:
	    RLEPutLine(Alpha, Pixels);
	    break;
	case IRIT_IMAGE_PPM3_TYPE:
	case IRIT_IMAGE_PPM6_TYPE:
	    PPMPutLine(Alpha, Pixels);
	    break;
    }
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Closes output image file.                                                M
*                                                                            *
* PARAMETERS:                                                                M
*   None                                                                     M
*                                                                            *
* RETURN VALUE:                                                              M
*   void                                                                     M
*                                                                            *
* KEYWORDS:                                                                  M
*   IrtImgWriteCloseFile, output image                                       M
*****************************************************************************/
void IrtImgWriteCloseFile(void)
{
    switch (IType) {
	default:
	case IRIT_IMAGE_RLE_TYPE:
	    RLECloseFile();
	    break;
	case IRIT_IMAGE_PPM3_TYPE:
	case IRIT_IMAGE_PPM6_TYPE:
	    PPMCloseFile();
	    break;
    }
}

/*****************************************************************************
* DESCRIPTION:                                                               *
*   Opens output image file in PPM format.                                   *
*   Creates file header. Note we allocate full image size since ppm expects  *
* lines top to bottom and we scan convert bottom to top.                     *
*                                                                            *
* PARAMETERS:                                                                *
*   argv:     Pointer to the name of this program.                           *
*   FName:    Filename to open, or NULL for stdout.			     *
*   XSize:    X dimension of the image.                                      *
*   YSize:    Y dimension of the image.                                      *
*                                                                            *
* RETURN VALUE:                                                              *
*   void                                                                     *
*****************************************************************************/
static void PPMOpenFile(char **argv, char *FName, int XSize, int YSize)
{
    int i;

    if (PPMFile != stdout && PPMFile != NULL)
        fclose(PPMFile);

    if (FName == NULL)
        PPMFile = stdout;
    else
	PPMFile = fopen(FName, "w");
    PPMWidth = XSize;
    PPMHeight = YSize;
    PPMLine = 0;

    PixelData = (IrtImgPixelStruct **) IritMalloc(PPMHeight *
						  sizeof(IrtImgPixelStruct *));

    for (i = 0; i < PPMHeight; i++) {
	PixelData[i] =
	    (IrtImgPixelStruct *) IritMalloc((PPMWidth + 4) *
					     sizeof(IrtImgPixelStruct));
    }

    if (PPMFile != NULL)
        fprintf(PPMFile,
		"%s\n# IRIT Irender image file 24 bit rgb\n%d %d\n%d\n",
		IType == IRIT_IMAGE_PPM3_TYPE ? "P3" : "P6",
		PPMWidth, PPMHeight, 255);
}

/*****************************************************************************
* DESCRIPTION:                                                               *
*   Outputs given line of color pixels and alpha correction values to the    *
*   ouput image file.                                                        *
*                                                                            *
* PARAMETERS:                                                                *
*   Alpha:  Array of alpha values.                                           *
*   Pixels: Array of color pixels.                                           *
*                                                                            *
* RETURN VALUE:                                                              *
*   void                                                                     *
*****************************************************************************/
static void PPMPutLine(ByteType *Alpha, IrtImgPixelStruct *Pixels)
{
    if (PPMLine < PPMHeight)
	GEN_COPY(PixelData[PPMLine++], Pixels,
		 PPMWidth * sizeof(IrtImgPixelStruct));
}

/*****************************************************************************
* DESCRIPTION:                                                               *
*   Closes output image file.                                                *
*                                                                            *
* PARAMETERS:                                                                *
*   None                                                                     *
*                                                                            *
* RETURN VALUE:                                                              *
*   void                                                                     *
*****************************************************************************/
static void PPMCloseFile(void)
{
    int i;

    for (i = 0; i < PPMHeight; i++) {
	if (IType == IRIT_IMAGE_PPM3_TYPE) {
	    int j;
	    IrtImgPixelStruct
		*Pixels = PixelData[PPMHeight - i - 1];

	    for (j = 0; j < PPMWidth; j++, Pixels++) {
		fprintf(PPMFile, j % 6 == 0 && j > 0 ? "\n%3d %3d %3d "
						     : "%3d %3d %3d ",
			Pixels -> r, Pixels -> g, Pixels -> b);
	    }
	    if (j % 6 != 0)
		fprintf(PPMFile, "\n");
	}
	else {
#ifdef __WINNT__
	    if (i == 0)
		setmode(_fileno(PPMFile), O_BINARY);
#endif /* __WINNT__ */
	    fwrite(PixelData[PPMHeight - i - 1],
		   PPMWidth * sizeof(IrtImgPixelStruct), 1, PPMFile);
	}

	IritFree(PixelData[PPMHeight - i - 1]);
    }
    IritFree(PixelData);

    if (PPMFile != NULL)
	fclose(PPMFile);

    PPMFile = NULL;
}

/*****************************************************************************
* DESCRIPTION:                                                               *
*   Opens output image file in RLE format.                                   *
*   Creates file header.                                                     *
*                                                                            *
* PARAMETERS:                                                                *
*   argv:      Pointer to the command line parameters.                       *
*   FName:     Filename to open, or NULL for stdout.			     *
*   Alpha:     Do we have Alpha channel.				     *
*   XSize:     X dimension of the image.                                     *
*   YSize:     Y dimension of the image.                                     *
*                                                                            *
* RETURN VALUE:                                                              *
*   void                                                                     *
*****************************************************************************/
static void RLEOpenFile(char **argv,
			char *FName,
			int Alpha,
			int XSize,
			int YSize)
{
#ifdef HAVE_URT_RLE
#ifdef URT_OLD_COMPAT
    Header = &rle_dflt_hdr;
    Header -> rle_file = rle_open_f("", FName, "w");
#else
    Header = rle_hdr_init((rle_hdr *) NULL);
    if (argv != NULL)
	rle_names(Header, *argv, NULL, 0);
    Header -> rle_file = rle_open_f(Header -> cmd, FName, "w");
#endif /* URT_OLD_COMAP */
    if (argv != NULL)
	rle_addhist(argv, (rle_hdr *) NULL, Header);
    if (Alpha) {
        RLE_SET_BIT(*Header, RLE_ALPHA);
	Header -> alpha = 1;
    }
    Header -> xmax = --XSize;
    Header -> ymax = --YSize;
    rle_put_setup(Header);
    rle_row_alloc(Header, &Rows);
#else
    fprintf(stderr, IRIT_EXP_STR("Utah raster tool kit is not supported\n"));
    exit(1);
#endif /* HAVE_URT_RLE */
}

/*****************************************************************************
* DESCRIPTION:                                                               *
*   Outputs given line of color pixels and alpha correction values to the    *
*   ouput image file.                                                        *
*                                                                            *
* PARAMETERS:                                                                *
*   Alpha:  Array of alpha values.                                           *
*   Pixels: Array of color pixels.                                           *
*                                                                            *
* RETURN VALUE:                                                              *
*   void                                                                     *
*****************************************************************************/
static void RLEPutLine(ByteType *Alpha, IrtImgPixelStruct *Pixels)
{
#ifdef HAVE_URT_RLE
    int x;

    for (x = 0; x <= Header -> xmax; x++) {
        IrtImgPixelStruct
	    *p = &Pixels[x];

	if (Alpha != NULL)
	    Rows[RLE_ALPHA][x] = Alpha[x];
        Rows[RLE_RED][x] = p -> r;
        Rows[RLE_GREEN][x] = p -> g;
        Rows[RLE_BLUE][x] = p -> b;
    }
    rle_putrow(Rows, Header -> xmax + 1, Header);
#endif /* HAVE_URT_RLE */
}

/*****************************************************************************
* DESCRIPTION:                                                               *
*   Closes output image file.                                                *
*                                                                            *
* PARAMETERS:                                                                *
*   None                                                                     *
*                                                                            *
* RETURN VALUE:                                                              *
*   void                                                                     *
*****************************************************************************/
static void RLECloseFile(void)
{
#ifdef HAVE_URT_RLE
    rle_puteof(Header);
    rle_close_f(Header -> rle_file);
#endif /* HAVE_URT_RLE */
}

/*****************************************************************************
* DESCRIPTION:                                                               *
*   Force the link of necessary rle functions so readimag.c in misc_lib      *
* would function, even if shared library.                                    *
*                                                                            *
* PARAMETERS:                                                                *
*   None                                                                     *
*                                                                            *
* RETURN VALUE:                                                              *
*   void                                                                     *
*****************************************************************************/
static void RLEDummyLink(void)
{
#ifdef HAVE_URT_RLE
    rle_open_f_noexit(NULL, NULL, NULL);
    rle_get_setup(NULL);
    rle_row_alloc(NULL, NULL);
    rle_getrow(NULL, NULL);
#endif /* HAVE_URT_RLE */
}
