/******************************************************************************
* Mvar_aux.c - auxiliary routine to interface to multi-variate rep.	      *
*******************************************************************************
* (C) Gershon Elber, Technion, Israel Institute of Technology                 *
*******************************************************************************
* Written by Gershon Elber, May. 97.					      *
******************************************************************************/

#include "mvar_loc.h"

/*****************************************************************************
* DESCRIPTION:                                                               M
* Given a multi-variate, returns its parametric domain.  The Min/Max arrays  M
* are assumed to of sufficiently large enough space to hold all dimensions,  M
* if Axis == -1.							     M
*                                                                            *
* PARAMETERS:                                                                M
*   MV:      Multi variate function to consider.                             M
*   Min:     Minimum domains of MV will be placed herein.                    M
*   Max:     Maximum domains of MV will be placed herein.                    M
*   Axis:    axis to extract or -1 for all axes.			     M
*                                                                            *
* RETURN VALUE:                                                              M
*   void                                                                     M
*                                                                            *
* KEYWORDS:                                                                  M
*   MvarMVDomain, multi-variates                                             M
*****************************************************************************/
void MvarMVDomain(MvarMVStruct *MV, CagdRType *Min, CagdRType *Max, int Axis)
{
    int i;

    if (Axis >= MV -> Dim)
	MVAR_FATAL_ERROR(MVAR_ERR_INVALID_AXIS);

    switch (MV -> GType) {
	case MVAR_BEZIER_TYPE:
	case MVAR_POWER_TYPE:
	    if (Axis == -1) {
		for (i = 0; i < MV -> Dim; i++) {
		    Min[i] = 0.0;
		    Max[i] = 1.0;
		}
	    }
	    else {
		*Min = 0.0;
		*Max = 1.0;
	    }
	    break;
	case MVAR_BSPLINE_TYPE:
	     if (Axis == -1) {
		for (i = 0; i < MV -> Dim; i++) {
		    int Order = MV -> Orders[i],
		        Len = MV -> Lengths[i];
		    CagdRType
		        *KV =  MV -> KnotVectors[i];

		    Min[i] = KV[Order - 1];
		    Max[i] = KV[Len];
		}
	    }
	    else {
		int Order = MV -> Orders[Axis],
		    Len = MV -> Lengths[Axis];
		CagdRType
		    *KV =  MV -> KnotVectors[Axis];

		*Min = KV[Order - 1];
		*Max = KV[Len];
	    }
	    break;
	default:
	    MVAR_FATAL_ERROR(MVAR_ERR_UNDEF_GEOM);
	    break;
    }
}

/*****************************************************************************
* DESCRIPTION:                                                               M
* Given a multi-variate and a domain - validate it.			     M
*                                                                            *
* PARAMETERS:                                                                M
*   MV:       To make sure t is in its Dir domain.                           M
*   t:        Parameter value to verify.                                     M
*   Dir:      Direction. Either U or V or W.                                 M
*                                                                            *
* RETURN VALUE:                                                              M
*   CagdBType:    TRUE if in domain, FALSE otherwise.                        M
*                                                                            *
* KEYWORDS:                                                                  M
*   MvarParamInDomain, multi-variates                                        M
*****************************************************************************/
CagdBType MvarParamInDomain(MvarMVStruct *MV, CagdRType t, MvarMVDirType Dir)
{
    CagdRType Min, Max;

    MvarMVDomain(MV, &Min, &Max, Dir);

    return t >= Min && t <= Max;
}

/*****************************************************************************
* DESCRIPTION:                                                               M
* Given a multi-variate and a domain - validate it.			     M
*                                                                            *
* PARAMETERS:                                                                M
*   MV:       To make sure (u, v, w) is in its domain.                       M
*   Params:   Array of real valued parameters of size Dim to verify if this  M
*	      point is in MV's parametric domain.                            M
*                                                                            *
* RETURN VALUE:                                                              M
*   CagdBType:   TRUE if in domain, FALSE otherwise.                         M
*                                                                            *
* KEYWORDS:                                                                  M
*   MvarParamsInDomain, multi-variates                                       M
*****************************************************************************/
CagdBType MvarParamsInDomain(MvarMVStruct *MV, CagdRType *Params)
{
    int i;

    switch (MV -> GType) {
	case MVAR_BEZIER_TYPE:
	case MVAR_POWER_TYPE:
	    for (i = 0; i < MV -> Dim; i++) {
		if (Params[i] < 0.0 || Params[i] > 1.0)
		    return FALSE;
	    }
	    break;
	case MVAR_BSPLINE_TYPE:
	    for (i = 0; i < MV -> Dim; i++) {
		int Order = MV -> Orders[i],
		    Len = MV -> Lengths[i];
		CagdRType
		    *KV =  MV -> KnotVectors[i];

		if (Params[i] < KV[Order - 1] || Params[i] > KV[Len])
		    return FALSE;
	    }
	    break;
	default:
	    MVAR_FATAL_ERROR(MVAR_ERR_UNDEF_GEOM);
	    break;
    }

    return TRUE;
}

/*****************************************************************************
* DESCRIPTION:                                                               M
* Given a multi-variate, returns a sub-region of it.                         M
*                                                                            *
* PARAMETERS:                                                                M
*   MV:        To extract a sub-region from.                                  M
*   t1, t2:    Domain to extract from MV, in parametric direction Dir.       M
*   Dir:       Direction to extract the sub-region. Either U or V or W.      M
*                                                                            *
* RETURN VALUE:                                                              M
*   MvarMVStruct *:   A sub-region of MV from t1 to t2 in direction Dir.     M
*                                                                            *
* KEYWORDS:                                                                  M
*   MvarMVRegionFromMV, multi-variates                                       M
*****************************************************************************/
MvarMVStruct *MvarMVRegionFromMV(MvarMVStruct *MV,
				 CagdRType t1,
				 CagdRType t2,
				 MvarMVDirType Dir)
{
    CagdRType TMin, TMax;
    MvarMVStruct *MVs;
    CagdBType
	NewMV = FALSE,
	BezMV = FALSE;

    switch (MV -> GType) {
	case MVAR_BEZIER_TYPE:
	    BezMV = TRUE;
	    break;
	case MVAR_BSPLINE_TYPE:
	    /* Might want to check for openend conditions here. */
	    break;
	default:
	    MVAR_FATAL_ERROR(MVAR_ERR_UNDEF_MVAR);
	    return NULL;
    }

    MvarMVDomain(MV, &TMin, &TMax, Dir);

    if (t1 > t2)
	SWAP(CagdRType, t1, t2);

    if (!APX_EQ(t1, TMin)) {
	MVs = MvarMVSubdivAtParam(MV, t1, Dir);
	MV = MVs -> Pnext;
	MVs -> Pnext = NULL;
	MvarMVFree(MVs);			   /* Free the first region. */
	NewMV = TRUE;
    }

    if (APX_EQ(t2, TMax))
	return NewMV ? MV : MvarMVCopy(MV);
    else {
	if (BezMV)
	    t2 = (t2 - t1) / (TMax - t1);

	MVs = MvarMVSubdivAtParam(MV, t2, Dir);

	if (NewMV)
	    MvarMVFree(MV);

    	MvarMVFree(MVs -> Pnext);		  /* Free the second region. */
    	MVs -> Pnext = NULL;
	return MVs;				/* Returns the first region. */
    }
}

/*****************************************************************************
* DESCRIPTION:                                                               M
* Computes a bounding box for a multi-variate freeform function.	     M
*                                                                            *
* PARAMETERS:                                                                M
*   MV:       To compute a bounding box for.                                 M
*   BBox:     Where bounding information is to be saved.                     M
*                                                                            *
* RETURN VALUE:                                                              M
*   void                                                                     M
*                                                                            *
* KEYWORDS:                                                                  M
*   MvarMVBBox, bbox, bounding box                                           M
*****************************************************************************/
void MvarMVBBox(MvarMVStruct *MV, CagdBBoxStruct *BBox)
{
    CagdPointsBBox(MV -> Points, MVAR_CTL_MESH_LENGTH(MV), BBox);
}

/*****************************************************************************
* DESCRIPTION:                                                               M
* Computes a bounding box for a list of multi-variate freeform function.     M
*                                                                            *
* PARAMETERS:                                                                M
*   MVs:    To compute a bounding box for.                                   M
*   BBox:   Where bounding information is to be saved.                       M
*                                                                            *
* RETURN VALUE:                                                              M
*   void                                                                     M
*                                                                            *
* KEYWORDS:                                                                  M
*   MvarMVListBBox, bbox, bounding box                                       M
*****************************************************************************/
void MvarMVListBBox(MvarMVStruct *MVs, CagdBBoxStruct *BBox)
{
    CAGD_RESET_BBOX(BBox);

    for ( ; MVs != NULL; MVs = MVs -> Pnext) {
	CagdBBoxStruct TmpBBox;

	MvarMVBBox(MVs, &TmpBBox);
	CagdMergeBBox(BBox, &TmpBBox);
    }
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Increment the index of the control mesh of the multivariate function     M
* by one.                                                                    M
*                                                                            *
* PARAMETERS:                                                                M
*   MV:         To increment Indices to its control mesh.                    M
*   Indices:    To increment one step.                                       M
*                                                                            *
* RETURN VALUE:                                                              M
*   int:        TRUE if Indices are in domain, FALSE if done.                M
*                                                                            *
* KEYWORDS:                                                                  M
*   MvarIncrementMeshIndices                                                 M
*****************************************************************************/
int MvarIncrementMeshIndices(MvarMVStruct *MV, int *Indices)
{
    int i;

    for (i = 0; i < MV -> Dim; i++) {
	if (++(Indices[i]) < MV -> Lengths[i])
	    return TRUE;
	Indices[i] = 0;
    }

    return FALSE;
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Increment the index of the control mesh of the multivariate function     M
* by one, skipping axis Dir.                                                 M
*                                                                            *
* PARAMETERS:                                                                M
*   MV:         To increment Indices to its control mesh.                    M
*   Indices:    To increment one step.                                       M
*   Dir:	To skip in the incrementation.                               M
*                                                                            *
* RETURN VALUE:                                                              M
*   int:        TRUE if Indices are in domain, FALSE if done.                M
*                                                                            *
* KEYWORDS:                                                                  M
*   MvarIncSkipMeshIndices                                                   M
*****************************************************************************/
int MvarIncSkipMeshIndices(MvarMVStruct *MV, int *Indices, int Dir)
{
    int i;

    for (i = 0; i < MV -> Dim; i++) {
	if (i == Dir)
	    continue;
	if (++(Indices[i]) < MV -> Lengths[i])
	    return TRUE;
	Indices[i] = 0;
    }

    return FALSE;
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Increment the index of the control mesh of the multivariate function     M
* by one, with given lower and upper bounds.                                 M
*                                                                            *
* PARAMETERS:                                                                M
*   MV:         To increment Indices to its control mesh.                    M
*   Indices:    To increment one step.                                       M
*   LowerBound:	Minimal values to assume.                                    M
*   UpperBound:	Maximal values to assume.                                    M
*                                                                            *
* RETURN VALUE:                                                              M
*   int:        TRUE if Indices are in domain, FALSE if done.                M
*                                                                            *
* KEYWORDS:                                                                  M
*   MvarIncBoundMeshIndices                                                  M
*****************************************************************************/
int MvarIncBoundMeshIndices(MvarMVStruct *MV,
			    int *Indices,
			    int *LowerBound,
			    int *UpperBound)
{
    int i;

    for (i = 0; i < MV -> Dim; i++) {
	if (++(Indices[i]) < UpperBound[i])
	    return TRUE;
	Indices[i] = LowerBound[i];
    }

    return FALSE;
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Given indices into the control mesh, return the index in the vector      M
* representation Points of that single point.                                M
*                                                                            *
* PARAMETERS:                                                                M
*   MV:         Whose indices are for.                                       M
*   Indices:    To compute the exact point location in MV -> Points          M
*                                                                            *
* RETURN VALUE:                                                              M
*   int:        Index of point whose indices are Indices in MV -> Points.    M
*                                                                            *
* KEYWORDS:                                                                  M
*   MvarGetPointsMeshIndices                                                 M
*****************************************************************************/
int MvarGetPointsMeshIndices(MvarMVStruct *MV, int *Indices)
{
    int i,
	Index = 0;

    for (i = 0; i < MV -> Dim; i++)
	Index += MV -> SubSpaces[i] * Indices[i];

    return Index;
}

/*****************************************************************************
* DESCRIPTION:                                                               M
* Merges two multivariates in the requested direction Dir.		     M
*  It is assumed that last edge of MV1 is identical to first edge of MV2.    M
*  It is assumed that both MVs have open end conditions and share the same   M
* orders and knot sequences in all axes, but the merged axes which can have  M
* different knots.							     M
*                                                                            *
* PARAMETERS:                                                                M
*   MV1:         To connect to MV2's starting boundary at its end.	     M
*   MV2:         To connect to MV1's end boundary at its start. 	     M
*   Dir:         Direction the merge should take place.			     M
*   Discont:     If TRUE, assumes the merged "edge" is discontinuous.        M
*                                                                            *
* RETURN VALUE:                                                              M
*   MvarMVStruct *:     The merged multivariate.                             M
*                                                                            *
* SEE ALSO:                                                                  M
*   CagdMergeSrfSrf                                                          M
*                                                                            *
* KEYWORDS:                                                                  M
*   MvarMergeMVMV, merge, multivariate                                       M
*****************************************************************************/
MvarMVStruct *MvarMergeMVMV(MvarMVStruct *MV1,
			    MvarMVStruct *MV2,
			    MvarMVDirType Dir,
			    CagdBType Discont)
{
    CagdBType IsNotRational,
	MergedEdgeDiscont = Discont ||
			    MV1 -> Orders[Dir] == 1 || MV2 -> Orders[Dir] == 1;
    int i, MaxCoord, *Lengths, *MergedIndices, *Indices1, *Indices2,
	*LowerBound, *UpperBound;
    CagdRType **MergedPoints, **Points1, **Points2;
    MvarMVStruct *MergedMV;

    if (MV1 -> Dim != MV2 -> Dim ||
	MV1 -> GType != MV2 -> GType ||
	MV1 -> PType != MV2 -> PType)  {
	MVAR_FATAL_ERROR(MVAR_ERR_FAIL_CMPT);
	return NULL;
    }

    /* Verify multivariate geometric types. */
    switch (MV1 -> GType) {
	case MVAR_BEZIER_TYPE:
	    MV1 = MvarCnvrtBezier2BsplineMV(MV1);
	    MV2 = MvarCnvrtBezier2BsplineMV(MV2);
	    break;
	case MVAR_BSPLINE_TYPE:
	    MV1 = MvarMVCopy(MV1);
	    MV2 = MvarMVCopy(MV2);
	    break;
	default:
	    MVAR_FATAL_ERROR(MVAR_ERR_UNDEF_GEOM);
	    return NULL;
    }

    IsNotRational = !MVAR_IS_RATIONAL_MV(MV1);
    MaxCoord = CAGD_NUM_OF_PT_COORD(MV1 -> PType),

    Lengths = (int *) IritMalloc(sizeof(int) * MV1 -> Dim);
    for (i = 0; i < MV1 -> Dim; i++) {
	if (i == Dir)
	    Lengths[i] = MV1 -> Lengths[i] + MV2 -> Lengths[i] - 1 +
							    MergedEdgeDiscont;
	else if (MV1 -> Lengths[i] == MV2 -> Lengths[i])
	    Lengths[i] = MV1 -> Lengths[i];
	else {
	    MvarMVFree(MV1);
	    MvarMVFree(MV2);
	    MVAR_FATAL_ERROR(MVAR_ERR_MVS_INCOMPATIBLE);
	    return NULL;
	}
    }

    MergedMV = MvarBspMVNew(MV1 -> Dim, Lengths, MV1 -> Orders, MV1 -> PType);
    IritFree((VoidPtr) Lengths);

    /* Update knot vectors. We assume open end condition here... */
    for (i = 0; i < MV1 -> Dim; i++) {
	CAGD_GEN_COPY(MergedMV -> KnotVectors[i], MV1 -> KnotVectors[i],
		      (MV1 -> Lengths[i] + MV1 -> Orders[i])
		                                          * sizeof(CagdRType));
	if (i == Dir) {
	    /* Append the second knot vector and affine transform so it      */
	    /* directly continues the first with Degree knots at connection. */
	    CAGD_GEN_COPY(&MergedMV -> KnotVectors[i][MV1 -> Lengths[i] +
						      MV1 -> Orders[i] - 1 +
						      MergedEdgeDiscont],
			  &MV2 -> KnotVectors[i][MV2 -> Orders[i]],
			  MV2 -> Lengths[i] * sizeof(CagdRType));
	    BspKnotAffineTrans(&MergedMV -> KnotVectors[i][MV1 -> Lengths[i] +
							   MV1 -> Orders[i] - 1],
			       MV2 -> Lengths[i],
			       MergedMV -> KnotVectors[i][MV1 -> Lengths[i] +
							  MV1 -> Orders[i] - 2] -
			           MV2 -> KnotVectors[i][0],
			       1.0);
	}
    }

    MergedPoints = MergedMV -> Points;
    Points1 = MV1 -> Points;
    Points2 = MV2 -> Points;

    MergedIndices = (int *) IritMalloc(sizeof(int) * MergedMV -> Dim);
    LowerBound = (int *) IritMalloc(sizeof(int) * MergedMV -> Dim);
    UpperBound = (int *) IritMalloc(sizeof(int) * MergedMV -> Dim);
    Indices1 = (int *) IritMalloc(sizeof(int) * MV1 -> Dim);
    Indices2 = (int *) IritMalloc(sizeof(int) * MV2 -> Dim);

    ZAP_MEM(LowerBound, sizeof(int) * MergedMV -> Dim);
    CAGD_GEN_COPY(UpperBound, MergedMV -> Lengths,
		  MergedMV -> Dim * sizeof(int));

    /* Copy the first control mesh into the merged multivariate. */
    UpperBound[Dir] = MV1 -> Lengths[Dir];
    ZAP_MEM(MergedIndices, sizeof(int) * MergedMV -> Dim);
    ZAP_MEM(Indices1, sizeof(int) * MV1 -> Dim);
    do {
	int MergedIndex = MvarGetPointsMeshIndices(MergedMV, MergedIndices),
	    Index1 = MvarGetPointsMeshIndices(MV1, Indices1);

	for (i = IsNotRational; i <= MaxCoord; i++)
	    MergedPoints[i][MergedIndex] = Points1[i][Index1];

	MvarIncBoundMeshIndices(MergedMV, MergedIndices,
				LowerBound, UpperBound);
    }
    while (MvarIncrementMeshIndices(MV1, Indices1));

    /* Copy the second control mesh into the merged multivariate. */
    LowerBound[Dir] = MV1 -> Lengths[Dir] - 1 + MergedEdgeDiscont;
    UpperBound[Dir] = MergedMV -> Lengths[Dir];
    CAGD_GEN_COPY(MergedIndices, LowerBound, sizeof(int) * MergedMV -> Dim);
    ZAP_MEM(Indices2, sizeof(int) * MV2 -> Dim);
    do {
	int MergedIndex = MvarGetPointsMeshIndices(MergedMV, MergedIndices),
	    Index2 = MvarGetPointsMeshIndices(MV2, Indices2);

	for (i = IsNotRational; i <= MaxCoord; i++)
	    MergedPoints[i][MergedIndex] = Points2[i][Index2];

	MvarIncBoundMeshIndices(MergedMV, MergedIndices,
				LowerBound, UpperBound);
    }
    while (MvarIncrementMeshIndices(MV2, Indices2));

    MvarMVFree(MV1);
    MvarMVFree(MV2);

    IritFree((VoidPtr) MergedIndices);
    IritFree((VoidPtr) LowerBound);
    IritFree((VoidPtr) UpperBound);
    IritFree((VoidPtr) Indices1);
    IritFree((VoidPtr) Indices2);

    return MergedMV;
}
