/******************************************************************************
* Mvar_Der.c - Compute derivatives of multi-variates.			      *
*******************************************************************************
* (C) Gershon Elber, Technion, Israel Institute of Technology                 *
*******************************************************************************
* Written by Gershon Elber, May. 97.					      *
******************************************************************************/

#include <ctype.h>
#include <stdio.h>
#include <string.h>
#include "mvar_loc.h"

/*****************************************************************************
* DESCRIPTION:                                                               M
* Given a multi-variate, computes its partial derivative multi-variate in    M
* direction Dir.							     M
*                                                                            *
* PARAMETERS:                                                                M
*   MV:       Multi-Variate to differentiate.                                M
*   Dir:      Direction of differentiation.			             M
*                                                                            *
* RETURN VALUE:                                                              M
*   MvarMVStruct *:   Differentiated multi-variate in direction Dir.         M
*                                                                            *
* KEYWORDS:                                                                  M
*   MvarMVDerive, multi-variates                                             M
*****************************************************************************/
MvarMVStruct *MvarMVDerive(MvarMVStruct *MV, MvarMVDirType Dir)
{
    switch (MV -> GType) {
	case MVAR_BEZIER_TYPE:
	    return MvarBzrMVDerive(MV, Dir);
	case MVAR_BSPLINE_TYPE:
	    return MvarBspMVDerive(MV, Dir);
	default:
	    MVAR_FATAL_ERROR(MVAR_ERR_UNDEF_GEOM);
	    return NULL;
    }
}

/*****************************************************************************
* DESCRIPTION:                                                               M
* Given a Bezier multi-variate, computes its partial derivative		     M
* multi-variate in direction Dir.					     M
*   Let old control polygon be P(i), i = 0 to k-1, and Q(i) be new one.	     M
*   Then:								     M
* Q(i) = (k - 1) * (P(i+1) - P(i)), i = 0 to k-2.			     V
*                                                                            *
* PARAMETERS:                                                                M
*   MV:       Multi-Variate to differentiate.                                M
*   Dir:      Direction of differentiation.                                  M
*                                                                            *
* RETURN VALUE:                                                              M
*   MvarMVStruct *:   Differentiated multi-variate in direction Dir. A       M
*		      Bezier multi-variate.				     M
*                                                                            *
* KEYWORDS:                                                                  M
*   MvarBzrMVDerive, multi-variates                                          M
*****************************************************************************/
MvarMVStruct *MvarBzrMVDerive(MvarMVStruct *MV, MvarMVDirType Dir)
{
    CagdBType
	IsNotRational = !MVAR_IS_RATIONAL_MV(MV);
    int l,
	Length = MV -> Lengths[Dir],
	*Indices = (int *) IritMalloc(sizeof(int) * MV -> Dim),
	MaxCoord = CAGD_NUM_OF_PT_COORD(MV -> PType);
    CagdRType **DPoints;
    MvarMVStruct
        *DerivedMV = NULL;

    if (!IsNotRational) {
	MVAR_FATAL_ERROR(MVAR_ERR_RATIONAL_NO_SUPPORT);
	return NULL;
    }

    ZAP_MEM(Indices, sizeof(int) * MV -> Dim);

    if (Length > 1)
        MV -> Lengths[Dir]--;
    DerivedMV = MvarBzrMVNew(MV -> Dim, MV -> Lengths, MV -> PType);
    if (Length > 1)
        MV -> Lengths[Dir]++;

    DPoints = DerivedMV -> Points;

    do {
	int DIndex = MvarGetPointsMeshIndices(DerivedMV, Indices),
	    Index = MvarGetPointsMeshIndices(MV, Indices),
	    NextIndex = Index + MVAR_NEXT_DIM(MV, Dir);

	for (l = IsNotRational; l <= MaxCoord; l++)
	    DPoints[l][DIndex] = Length < 2 ? 0.0
			                   : (Length - 1) *
						 (MV -> Points[l][NextIndex] -
						  MV -> Points[l][Index]);
    }
    while (MvarIncrementMeshIndices(DerivedMV, Indices));

    IritFree(Indices);

    return DerivedMV;
}

/*****************************************************************************
* DESCRIPTION:                                                               M
* Given a Bspline multi-variate, computes its partial derivative             M
* multi-variate in direction Dir.					     M
*   Let old control polygon be P(i), i = 0 to k-1, and Q(i) be new one.	     M
*   Then:								     M
* Q(i) = (k - 1) * (P(i+1) - P(i)) / (Kv(i + k) - Kv(i + 1)), i = 0 to k-2.  V
*                                                                            *
* PARAMETERS:                                                                M
*   MV:       Multi-Variate to differentiate.                                M
*   Dir:      Direction of differentiation.                                  M
*                                                                            *
* RETURN VALUE:                                                              M
*   MvarMVStruct *:   Differentiated multi-variate in direction Dir. A       M
*		      Bspline multi-variate.				     M
*                                                                            *
* KEYWORDS:                                                                  M
*   MvarBspMVDerive, multi-variates                                          M
*****************************************************************************/
MvarMVStruct *MvarBspMVDerive(MvarMVStruct *MV, MvarMVDirType Dir)
{
    CagdBType
	IsNotRational = !MVAR_IS_RATIONAL_MV(MV);
    int i,
	NewLength, NewOrder,
	Length = MV -> Lengths[Dir],
	Order = MV -> Orders[Dir],
	*Indices = (int *) IritMalloc(sizeof(int) * MV -> Dim),
	MaxCoord = CAGD_NUM_OF_PT_COORD(MV -> PType);
    CagdRType **DPoints,
	*KV = MV -> KnotVectors[Dir],
	**Points = MV -> Points;
    MvarMVStruct
        *DerivedMV = NULL;

    if (!IsNotRational) {
	MVAR_FATAL_ERROR(MVAR_ERR_RATIONAL_NO_SUPPORT);
	return NULL;
    }

    ZAP_MEM(Indices, sizeof(int) * MV -> Dim);

    NewLength = Order < 2 ? Length : Length - 1;
    NewOrder = MAX(Order - 1, 1);

    MV -> Lengths[Dir] = NewLength;
    MV -> Orders[Dir] = NewOrder;
    DerivedMV = MvarBspMVNew(MV -> Dim, MV -> Lengths, MV -> Orders,
			     MV -> PType);
    MV -> Lengths[Dir] = Length;
    MV -> Orders[Dir] = Order;

    for (i = 0; i < MV -> Dim; i++) {
	if (i == Dir)
	    CAGD_GEN_COPY(DerivedMV -> KnotVectors[i],
			  &MV -> KnotVectors[i][Order < 2 ? 0 : 1],
			  sizeof(CagdRType) * (NewLength + NewOrder));
	else
	    CAGD_GEN_COPY(DerivedMV -> KnotVectors[i], MV -> KnotVectors[i],
			  sizeof(CagdRType) * (MV -> Lengths[i] +
					       MV -> Orders[i]));
    }

    DPoints = DerivedMV -> Points;

    do {
	int l,
	    DIndex = MvarGetPointsMeshIndices(DerivedMV, Indices),
	    Index = MvarGetPointsMeshIndices(MV, Indices),
	    NextIndex = Index + MVAR_NEXT_DIM(MV, Dir);
	CagdRType
	    Denom = KV[Indices[Dir] + Order] - KV[Indices[Dir] + 1];

	if (APX_EQ(Denom, 0.0))
	    Denom = IRIT_UEPS;

	for (l = IsNotRational; l <= MaxCoord; l++) {
	    DPoints[l][DIndex] =
	        Order < 2 ? 0.0 : (Order - 1) *
				      (Points[l][NextIndex] -
				       Points[l][Index]) / Denom;
	}
    }
    while (MvarIncrementMeshIndices(DerivedMV, Indices));

    IritFree(Indices);

    return DerivedMV;
}

