/******************************************************************************
* MVar_Pll.c - process polylines out of MvarPlnStruct.			      *
*******************************************************************************
* (C) Gershon Elber, Technion, Israel Institute of Technology                 *
*******************************************************************************
* Written by Gershon Elber, Aug 2003.					      *
******************************************************************************/

#include "irit_sm.h"
#include "mvar_loc.h"

#define TEST_CLOSEST_DIST(P1, P2, IsStart1, IsStart2) { \
		RealType DstSqr; \
		if ((DstSqr = MvarPtDistSqrTwoPoints(P1, \
					             P2)) < TolSqr) { \
	            TolSqr = DstSqr; \
		    *Start1 = IsStart1; \
		    *Start2 = IsStart2; \
		    P1PrevMin = P1Prev; \
		    P2PrevMin = P2Prev; \
	        } \
	    }

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Returns the last element in the list.                                    M
*                                                                            *
* PARAMETERS:                                                                M
*   Pts:    List of points to fetch last one.                                M
*                                                                            *
* RETURN VALUE:                                                              M
*   MvarPtStruct *:   Last element in list.                                  M
*                                                                            *
* KEYWORDS:                                                                  M
*   MvarGetLastPt                                                            M
*****************************************************************************/
MvarPtStruct *MvarGetLastPt(MvarPtStruct *Pts)
{
    if (Pts == NULL)
	return NULL;

    while (Pts -> Pnext != NULL)
	Pts = Pts -> Pnext;

    return Pts;
}

/*****************************************************************************
* DESCRIPTION:                                                               *
*   Searchs for the two closest vertices/end points in the given list of     *
* polylines, remove them from list and returned them in Pl1/2.		     *
*                                                                            *
* PARAMETERS:                                                                *
*   PlList:    A list of polylines to search.                                *
*   MaxTol:    Maximum distance that is still considered o.k. to connect.    *
*   Pl1, Pl2:  The two matched points as the two polylines holding them.     *
*   Start1, Start2:  TRUE if start of polyline, FALSE if end of polyline.    *
*                                                                            *
* RETURN VALUE:                                                              *
*   int:      TRUE if found a match closer than MaxTol, FALSE otherwise.     *
*****************************************************************************/
static int MVGetClosestVertices(MvarPolyStruct **PlList,
				RealType MaxTol,
				MvarPolyStruct **Pl1,
				MvarPolyStruct **Pl2,
				int *Start1,
				int *Start2)
{
    MvarPolyStruct *P1, *P1Prev, *P2, *P2Prev,
	*P1PrevMin = NULL,
	*P2PrevMin = NULL;
    RealType
        TolSqr = IRIT_INFNTY;

    for (P1Prev = NULL, P1 = *PlList;
	 P1 -> Pnext != NULL;
	 P1Prev = P1, P1 = P1 -> Pnext) {
        MvarPtStruct
	    *P1Start = P1 -> Pl,
	    *P1End = MvarGetLastPt(P1Start);

	for (P2Prev = P1, P2 = P1 -> Pnext;
	     P2 != NULL;
	     P2Prev = P2, P2 = P2 -> Pnext) {
	    MvarPtStruct
	        *P2Start = P2 -> Pl,
	        *P2End = MvarGetLastPt(P2Start);

	    if (P1Start == P1End) {
	        if (P2Start == P2End) {
		    TEST_CLOSEST_DIST(P1Start, P2Start, 1, 1);
		}
		else {
		    TEST_CLOSEST_DIST(P1Start, P2Start, 1, 1);
		    TEST_CLOSEST_DIST(P1Start, P2End,   1, 0);
		}
	    }
	    else {
	        if (P2Start == P2End) {
		    TEST_CLOSEST_DIST(P1Start, P2Start, 1, 1);
		    TEST_CLOSEST_DIST(P1End,   P2Start, 0, 1);
		}
		else {
		    TEST_CLOSEST_DIST(P1Start, P2Start, 1, 1);
		    TEST_CLOSEST_DIST(P1End,   P2Start, 0, 1);
		    TEST_CLOSEST_DIST(P1Start, P2End,   1, 0);
		    TEST_CLOSEST_DIST(P1End,   P2End,   0, 0);
		}
	    }
	}
    }

    if (TolSqr >= MaxTol)
        return FALSE;
    else {
	/* Update the two matched polylines. */
	*Pl1 = P1PrevMin == NULL ? *PlList : P1PrevMin -> Pnext;
	*Pl2 = P2PrevMin -> Pnext;

	/* And remove them from the global list. */
	P2PrevMin -> Pnext = P2PrevMin -> Pnext -> Pnext;
	if (P1PrevMin == NULL) {
	    *PlList = (*PlList) -> Pnext;
	}
	else
	    P1PrevMin -> Pnext = P1PrevMin -> Pnext -> Pnext;

        return TRUE;
    }    
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   A comparison function to examine if the given two points are the same.   M
*                                                                            *
* PARAMETERS:                                                                M
*   P1, P2:  Tow multivariate points to compare.                             M
*   Eps: N.S.F.I.                                                            M
*                                                                            *
* RETURN VALUE:                                                              M
*   int:      0 if identical, -1 or +1 if first point is less than /greater  M
*	      than second point, in lexicographic order over dimensions.     M
*                                                                            *
* SEE ALSO:                                                                  M
*   MvarPtDistTwoPoints, MvarPtDistSqrTwoPoints                              M
*                                                                            *
* KEYWORDS:                                                                  M
*   MvarPtCmpTwoPoints                                                       M
*****************************************************************************/
int MvarPtCmpTwoPoints(MvarPtStruct *P1, MvarPtStruct *P2, CagdRType Eps)
{
    int i,
	Dim = P1 -> Dim;

    if (Dim != P2 -> Dim)
	return FALSE;

    for (i = 0; i < Dim; i++) {
	if (!APX_EQ_EPS(P1 -> Pt[i], P2 -> Pt[i], Eps))
	    return SIGN(P1 -> Pt[i] - P2 -> Pt[i]);
    }

    return TRUE;
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Compute the Eucildean distance between two multivariate points.          M
*                                                                            *
* PARAMETERS:                                                                M
*   P1, P2:  Two points to compute the distance between.                     M
*                                                                            *
* RETURN VALUE:                                                              M
*   CagdRType:    Distance computed.                                         M
*                                                                            *
* SEE ALSO:                                                                  M
*   MvarPtCmpTwoPoints, MvarPtDistSqrTwoPoints                               M
*                                                                            *
* KEYWORDS:                                                                  M
*   MvarPtDistTwoPoints                                                      M
*****************************************************************************/
CagdRType MvarPtDistTwoPoints(MvarPtStruct *P1, MvarPtStruct *P2)
{
    return sqrt(MvarPtDistSqrTwoPoints(P1, P2));
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Compute the Eucildean distance between two multivariate points.          M
*                                                                            *
* PARAMETERS:                                                                M
*   P1, P2:  Two points to compute the distance between.                     M
*                                                                            *
* RETURN VALUE:                                                              M
*   CagdRType:    Distance computed.                                         M
*                                                                            *
* SEE ALSO:                                                                  M
*   MvarPtCmpTwoPoints, MvarPtDistTwoPoints                                  M
*                                                                            *
* KEYWORDS:                                                                  M
*   MvarPtDistSqrTwoPoints                                                   M
*****************************************************************************/
CagdRType MvarPtDistSqrTwoPoints(MvarPtStruct *P1, MvarPtStruct *P2)
{
    int i,
	Dim = P1 -> Dim;
    CagdRType
	Dist = 0.0;

    if (Dim != P2 -> Dim)
	return 0.0;

    for (i = 0; i < Dim; i++) {
	Dist += SQR(P1 -> Pt[i] - P2 -> Pt[i]);
    }

    return Dist;
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Merges seperated multivariate polylines into longer ones, in place, as   M
* possible.								     M
    Given a list of multivariate polylines, matches end points and merged as M
* possible multivariate polylines with common end points, in place.	     M
*                                                                            *
* PARAMETERS:                                                                M
*   Polys:      Multivariate polylines to merge, in place.                   M
*   Eps:	Epslion of similarity to merge multivariate polylines at.    M
*                                                                            *
* RETURN VALUE:                                                              M
*   MvarPolyStruct *:  Merged as possible multivariate polylines.            M
*                                                                            *
* SEE ALSO:                                                                  M
*   GMMergePolylines				                             M
*                                                                            *
* KEYWORDS:                                                                  M
*   MvarPolyMergePolylines, merge, polyline                                  M
*****************************************************************************/
MvarPolyStruct *MvarPolyMergePolylines(MvarPolyStruct *Polys, RealType Eps)
{
    MvarPolyStruct *Pl, *Pl2, *Pl2Prev;

    for (Pl = Polys; Pl != NULL; ) {
	int HasChanged = FALSE;
	MvarPtStruct
	    *P1 = Pl -> Pl,
	    *P2 = MvarGetLastPt(P1);

	for (Pl2Prev = Pl, Pl2 = Pl -> Pnext; Pl2 != NULL && !HasChanged; ) { 
	    MvarPtStruct *P,
		*Pl2P1 = Pl2 -> Pl,
		*Pl2P2 = MvarGetLastPt(Pl2P1);
	    int FoundMatch = TRUE;

	    if (MvarPtCmpTwoPoints(P1, Pl2P1, Eps)) {
		Pl -> Pl = MvarPolyReverseList(Pl -> Pl);
		P = MvarGetLastPt(Pl -> Pl);
		P -> Pnext = Pl2 -> Pl -> Pnext;
	    }
	    else if (MvarPtCmpTwoPoints(P1, Pl2P2, Eps)) {
		Pl -> Pl = MvarPolyReverseList(Pl -> Pl);
		Pl2 -> Pl = MvarPolyReverseList(Pl2 -> Pl);
		P = MvarGetLastPt(Pl -> Pl);
		P -> Pnext = Pl2 -> Pl -> Pnext;
	    }
	    else if (MvarPtCmpTwoPoints(P2, Pl2P1, Eps)) {
		P2 -> Pnext = Pl2P1 -> Pnext;
	    }
	    else if (MvarPtCmpTwoPoints(P2, Pl2P2, Eps)) {
		Pl2 -> Pl = MvarPolyReverseList(Pl2 -> Pl);
		P2 -> Pnext = Pl2 -> Pl -> Pnext;
	    }
	    else
	        FoundMatch = FALSE;

	    if (FoundMatch) {
		Pl2Prev -> Pnext = Pl2 -> Pnext;

		Pl2 -> Pl -> Pnext = NULL;
		MvarPolyFree(Pl2);
		Pl2 = Pl2Prev -> Pnext;

		HasChanged = TRUE;
	    }
	    else {
	        Pl2Prev = Pl2;
		Pl2 = Pl2 -> Pnext;
	    }
	}

	if (!HasChanged)
	    Pl = Pl -> Pnext;
    }

    return Polys;
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Connect the list of multivariate points into multivariate polylines by   M
* connecting the closest multivariate point pairs, until the distances       M
* between adjacent multivariate points/polylines is more than MaxTol.        M
*                                                                            *
* PARAMETERS:                                                                M
*   PtsList:      Point list to connect into multivariate polylines.         M
*   MaxTol:       Maximum distance allowed to connect multivariate points.   M
*                                                                            *
* RETURN VALUE:                                                              M
*   MvarPolyStruct *:    Connected multivariate polylines, upto MaxTol       M
*			 tolerance.					     M
*                                                                            *
* SEE ALSO:                                                                  M
*   GMMatchPointListIntoPolylines		                             M
*                                                                            *
* KEYWORDS:                                                                  M
*   MvarMatchPointListIntoPolylines                                          M
*****************************************************************************/
MvarPolyStruct *MvarMatchPointListIntoPolylines(MvarPtStruct *PtsList,
						RealType MaxTol)
{
    int Start1, Start2;
    MvarPolyStruct *PllList, *Pl1, *Pl2;
    MvarPtStruct *Pt;

    /* Convert the poly list to polyline linked list with one point in   */
    /* each polyline so we can start and match-connect them.		 */
    for (PllList = NULL, Pt = PtsList; Pt != NULL; Pt = Pt -> Pnext) {
        Pl1 = MvarPolyNew(MvarPtCopy(Pt));
	LIST_PUSH(Pl1, PllList);
    }

    /* Get the two closest points and connect them until no more data. */
    while (PllList -> Pnext != NULL &&
	   MVGetClosestVertices(&PllList, MaxTol,
				&Pl1, &Pl2, &Start1, &Start2)) {
        if (Start1)
	    Pl1 -> Pl = MvarPolyReverseList(Pl1 -> Pl);
        if (!Start2)
	    Pl2 -> Pl = MvarPolyReverseList(Pl2 -> Pl);
	MvarGetLastPt(Pl1 -> Pl) -> Pnext = Pl2 -> Pl;
	Pl2 -> Pl = NULL;
	MvarPolyFree(Pl2);
	LIST_PUSH(Pl1, PllList);
    }

    return PllList;
}
