/******************************************************************************
* Mvar_Sub.c - Computes subdivision of multi-variates.			      *
*******************************************************************************
* (C) Gershon Elber, Technion, Israel Institute of Technology                 *
*******************************************************************************
* Written by Gershon Elber, May. 97.					      *
******************************************************************************/

#include "mvar_loc.h"

/*****************************************************************************
* DESCRIPTION:                                                               M
* Given a multi-variate, subdivides it at parameter value t in direction     M
* Dir.									     M
*                                                                            *
* PARAMETERS:                                                                M
*   MV:       Multi-Variate to subdivide.                                    M
*   t:        Parameter to subdivide at.                                     M
*   Dir:      Direction of subdivision.                                      M
*                                                                            *
* RETURN VALUE:                                                              M
*   MvarMVStruct *: A list of two multi-variates, result of the subdivision. M
*                                                                            *
* SEE ALSO:                                                                  M
*   MvarBspMVSubdivAtParam, MvarBzrMVSubdivAtParam		             M
*                                                                            *
* KEYWORDS:                                                                  M
*   MvarMVSubdivAtParam, multi-variates                                      M
*****************************************************************************/
MvarMVStruct *MvarMVSubdivAtParam(MvarMVStruct *MV,
				  CagdRType t,
				  MvarMVDirType Dir)
{
    switch (MV -> GType) {
	case MVAR_BEZIER_TYPE:
	    return MvarBzrMVSubdivAtParam(MV, t, Dir);
	case MVAR_BSPLINE_TYPE:
	    return MvarBspMVSubdivAtParam(MV, t, Dir);
	default:
	    MVAR_FATAL_ERROR(MVAR_ERR_UNDEF_GEOM);
	    break;
    }

    return NULL;
}

/*****************************************************************************
* DESCRIPTION:                                                               M
* Given a Bezier multi-variate, subdivides it at parameter value t in        M
* direction Dir.							     M
*                                                                            *
* PARAMETERS:                                                                M
*   MV:       Bezier Multi-Variate to subdivide.                             M
*   t:        Parameter to subdivide at.                                     M
*   Dir:      Direction of subdivision.                                      M
*                                                                            *
* RETURN VALUE:                                                              M
*   MvarMVStruct *: A list of two multi-variates, result of the subdivision. M
*                                                                            *
* SEE ALSO:                                                                  M
*   MvarBspMVSubdivAtParam, MvarMVSubdivAtParam			             M
*                                                                            *
* KEYWORDS:                                                                  M
*   MvarBzrMVSubdivAtParam, multi-variates                                   M
*****************************************************************************/
MvarMVStruct *MvarBzrMVSubdivAtParam(MvarMVStruct *MV,
				     CagdRType t,
				     MvarMVDirType Dir)
{
    CagdBType
	IsNotRational = !MVAR_IS_RATIONAL_MV(MV);
    int i, j, l, *RIndices,
	MaxCoord = CAGD_NUM_OF_PT_COORD(MV -> PType),
	Length = MV -> Lengths[Dir],
	Len = MVAR_CTL_MESH_LENGTH(MV);
    CagdRType **LPoints, **RPoints,
	t1 = 1.0 - t;
    MvarMVStruct *RMV, *LMV;

    if (Dir < 0 || Dir >= MV -> Dim) {
	MVAR_FATAL_ERROR(MVAR_ERR_DIR_NOT_VALID);
	return NULL;
    }

    if (!MVAR_IS_BEZIER_MV(MV)) {
	MVAR_FATAL_ERROR(MVAR_ERR_UNDEF_MVAR);
	return NULL;
    }

    LMV = MvarBzrMVNew(MV -> Dim, MV -> Lengths, MV -> PType);
    RMV = MvarBzrMVNew(MV -> Dim, MV -> Lengths, MV -> PType);
    LPoints = LMV -> Points;
    RPoints = RMV -> Points;

    /* Copy MV into RMV/LMV, so we can apply the recursive algo. to it.      */
    for (j = IsNotRational; j <= MaxCoord; j++) {
        GEN_COPY(RPoints[j], MV -> Points[j], Len * sizeof(CagdRType));
        GEN_COPY(LPoints[j], MV -> Points[j], Len * sizeof(CagdRType));
    }

    /* Do the control mesh's subdivision. */

    /* Update the control mesh. */
    RIndices = (int *) IritMalloc(sizeof(int) * MV -> Dim);
    ZAP_MEM(RIndices, sizeof(int) * MV -> Dim);

    do {
        for (i = 1; i < Length; i++) {
	    int LIndex, RIndex, RIndex1;

	    for (l = 0; l < Length - i; l++) {
	        RIndices[Dir] = l;
		RIndex = MvarGetPointsMeshIndices(RMV, RIndices);
	        RIndices[Dir]++;
		RIndex1 = MvarGetPointsMeshIndices(RMV, RIndices);

	        for (j = IsNotRational; j <= MaxCoord; j++)
		    RPoints[j][RIndex] = RPoints[j][RIndex] * t1 +
		                         RPoints[j][RIndex1] * t;
	    }

	    /* Copy temporary result to LMV: */
	    RIndices[Dir] = 0;
	    RIndex = MvarGetPointsMeshIndices(RMV, RIndices);
	    RIndices[Dir] = i;
	    LIndex = MvarGetPointsMeshIndices(LMV, RIndices);
	    for (j = IsNotRational; j <= MaxCoord; j++)
	        LPoints[j][LIndex] = RPoints[j][RIndex];
	}
    }
    while (MvarIncSkipMeshIndices(RMV, RIndices, Dir));

    IritFree((VoidPtr) RIndices);

    LMV -> Pnext = RMV;

    CAGD_PROPAGATE_ATTR(RMV, MV);
    CAGD_PROPAGATE_ATTR(LMV, MV);

    return LMV;
}

/*****************************************************************************
* DESCRIPTION:                                                               M
* Given a Bspline multi-variate, subdivides it at parameter value t in       M
* direction Dir.							     M
*                                                                            *
* PARAMETERS:                                                                M
*   MV:       BsplineMulti-Variate to subdivide.                             M
*   t:        Parameter to subdivide at.                                     M
*   Dir:      Direction of subdivision.                                      M
*                                                                            *
* RETURN VALUE:                                                              M
*   MvarMVStruct *: A list of two multi-variates, result of the subdivision. M
*                                                                            *
* SEE ALSO:                                                                  M
*   MvarMVSubdivAtParam, MvarBzrMVSubdivAtParam		                     M
*                                                                            *
* KEYWORDS:                                                                  M
*   MvarBspMVSubdivAtParam, multi-variates                                   M
*****************************************************************************/
MvarMVStruct *MvarBspMVSubdivAtParam(MvarMVStruct *MV,
				     CagdRType t,
				     MvarMVDirType Dir)
{
    CagdBType
	IsNotRational = !MVAR_IS_RATIONAL_MV(MV);
    int i, j, KVLen, Index1, Index2, *Indices, Mult, RLength, LLength,
	MaxCoord = CAGD_NUM_OF_PT_COORD(MV -> PType),
	Order = MV -> Orders[Dir],
	Length = MV -> Lengths[Dir];
    CagdRType *RefKV;
    MvarMVStruct *RMV, *LMV;
    BspKnotAlphaCoeffType *A;

    if (Dir < 0 || Dir >= MV -> Dim) {
	MVAR_FATAL_ERROR(MVAR_ERR_DIR_NOT_VALID);
	return NULL;
    }

    if (MVAR_IS_BSPLINE_MV(MV)) {
	RefKV = MV -> KnotVectors[Dir];
	KVLen = Order + Length;
	Index1 = BspKnotLastIndexL(RefKV, KVLen, t);
	if (Index1 + 1 < Order)
	    Index1 = Order - 1;
	Index2 = BspKnotFirstIndexG(RefKV, KVLen, t);
	if (Index2 > Length)
	    Index2 = Length;
	Mult = Order - 1 - (Index2 - Index1 - 1);

	MV -> Lengths[Dir] = Index1 + 1;
	LMV = MvarBspMVNew(MV -> Dim, MV -> Lengths, MV -> Orders, MV -> PType);
	MV -> Lengths[Dir] = Length - Index2 + Order;
	RMV = MvarBspMVNew(MV -> Dim, MV -> Lengths, MV -> Orders, MV -> PType);
	MV -> Lengths[Dir] = Length;

	/* Update the new knot vectors. */
	for (i = 0; i < MV -> Dim; i++) {
	    if (i == Dir) {
		CAGD_GEN_COPY(LMV -> KnotVectors[i],
			      MV -> KnotVectors[i],
			      sizeof(CagdRType) * (Index1 + 1));

		/* Close the knot vector with multiplicity Order: */
		for (j = Index1 + 1; j <= Index1 + Order; j++)
		    LMV -> KnotVectors[i][j] = t;

		CAGD_GEN_COPY(&RMV -> KnotVectors[i][Order],
			      &MV -> KnotVectors[i][Index2],
			      sizeof(CagdRType) *
				  (Length + Order - Index2));

		/* Make sure knot vector starts with multiplicity Order: */
		for (j = 0; j < Order; j++)
		    RMV -> KnotVectors[i][j] = t;
	    }
	    else {
		/* And copy the other direction(s)' knot vectors. */
		CAGD_GEN_COPY(LMV -> KnotVectors[i],
			      MV -> KnotVectors[i],
			      sizeof(CagdRType) * (MV -> Orders[i] +
						   MV -> Lengths[i]));
		CAGD_GEN_COPY(RMV -> KnotVectors[i],
			      MV -> KnotVectors[i],
			      sizeof(CagdRType) * (MV -> Orders[i] +
						   MV -> Lengths[i]));
	    }
	}
    }
    else {
	MVAR_FATAL_ERROR(MVAR_ERR_UNDEF_MVAR);
	return NULL;
    }

    /* Do the control mesh subdivision. */
    if (Mult > 0) {
	CagdRType Min, Max,
	    *NewKV = (CagdRType *) IritMalloc(sizeof(CagdRType) * Mult);

	MvarMVDomain(MV, &Min, &Max, Dir);

	CAGD_DOMAIN_T_VERIFY(t, Min, Max);
	if (t == Max)
	    t -= CAGD_DOMAIN_IRIT_EPS;
	for (i = 0; i < Mult; i++)
	    NewKV[i] = t;
	A = BspKnotEvalAlphaCoefMerge(Order, RefKV, Length, NewKV, Mult,
				      FALSE);
	IritFree((VoidPtr) NewKV);
    }
    else
        A = BspKnotEvalAlphaCoefMerge(Order, RefKV, Length, NULL, 0,
				      FALSE);

    /* Note that Mult can be negative in cases where original multiplicity   */
    /* was order or more and we need to compensate here, since Alpha matrix  */
    /* will be just a unit matrix then.					     */
    Mult = Mult >= 0 ? 0 : -Mult;

    /* Update the control mesh. */
    Indices = (int *) IritMalloc(sizeof(int) * MV -> Dim);

    LLength = LMV -> Lengths[Dir];
    RLength = RMV -> Lengths[Dir];

    /* Do the left hand side. */
    ZAP_MEM(Indices, sizeof(int) * MV -> Dim);

    do {
	int Index = MvarGetPointsMeshIndices(MV, Indices),
	    LIndex = MvarGetPointsMeshIndices(LMV, Indices);

	for (j = IsNotRational; j <= MaxCoord; j++) {
	    CagdRType
	        *LPts = &LMV -> Points[j][LIndex],
	        *Pts = &MV -> Points[j][Index];

	    for (i = 0;
		 i < LLength;
		 i++, LPts += MVAR_NEXT_DIM(LMV, Dir)) {
		CAGD_ALPHA_BLEND_STEP(A, i, Pts, IRIT_INFNTY,
				      LPts, MVAR_NEXT_DIM(MV, Dir));
	    }
	}
    }
    while (MvarIncSkipMeshIndices(RMV, Indices, Dir));

    /* Do the right hand side. */
    ZAP_MEM(Indices, sizeof(int) * MV -> Dim);

    do {
	int Index = MvarGetPointsMeshIndices(MV, Indices),
	    RIndex = MvarGetPointsMeshIndices(RMV, Indices),
	    Offset = LLength - 1;

	for (j = IsNotRational; j <= MaxCoord; j++) {
	    CagdRType
	        *RPts = &RMV -> Points[j][RIndex],
	        *Pts = &MV -> Points[j][Index];

	    for (i = Offset + Mult;
		 i < RLength + Offset + Mult;
		 i++, RPts += MVAR_NEXT_DIM(RMV, Dir)) {
		CAGD_ALPHA_BLEND_STEP(A, i, Pts, IRIT_INFNTY,
				      RPts, MVAR_NEXT_DIM(MV, Dir));
	    }
	}
    }
    while (MvarIncSkipMeshIndices(RMV, Indices, Dir));

    IritFree((VoidPtr) Indices);

    BspKnotFreeAlphaCoef(A);

    LMV -> Pnext = RMV;
    return LMV;
}

#ifdef DEBUG_ALPHA_MAT

/*****************************************************************************
* DESCRIPTION:                                                               *
* Prints the content of the alpha matrix.				     *
*                                                                            *
* PARAMETERS:                                                                *
*   A:        Alpha matrix to print.                                         *
*                                                                            *
* RETURN VALUE:                                                              *
*   void                                                                     *
*****************************************************************************/
static void PrintAlphaMat(BspKnotAlphaCoeffType *A)
{
    int i, j;

    fprintf(stderr, "Order = %d, Length = %d\n", A -> Order, A -> Length);
    for (i = 0; i < A -> Length; i++) {
	for (j = 0; j < A -> RefLength; j++)
	    fprintf(stderr, " %9.5g", A -> Rows[i][j]);
	fprintf(stderr, "\n");
    }

    fprintf(stderr, "    ");
    for (j = 0; j < A -> RefLength; j++)
	    fprintf(stderr, " %3d %3d |", A -> ColIndex[j], A -> ColLength[j]);
	fprintf(stderr, "\n");
}

#endif /* DEBUG_ALPHA_MAT */
