/******************************************************************************
* MvarDist.c - Computes ditsances between multi-variates.		      *
*******************************************************************************
* (C) Gershon Elber, Technion, Israel Institute of Technology                 *
*******************************************************************************
* Written by Gershon Elber, May. 97.					      *
******************************************************************************/

#include "mvar_loc.h"

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Given a curve and a surface, creates a multivariate scalar field         M
* representing the distance function square, between them.		     M
*                                                                            *
* PARAMETERS:                                                                M
*   Crv1, Srf2:  The two entities, Crv1(t) and Srf2(u, v), to form their     M
*                distance function square between them as a multivariate     M
*                function.						     M
*   DistType:    0 for distance vector function,			     M
*		 1 for distance square scalar function,			     M
*		 2 for distance vector projected on the normal of Crv1,	     M
*		 3 for distance vector projected on the normal of Srf2.	     M
*		 In cases 2 and 3 the normal field is not normalized.        M
*                                                                            *
* RETURN VALUE:                                                              M
*   MvarMVStruct *:  The distance function square d2(t, u, v) of the         M
*                    distance from Crv1(t) to Srf2(u, v).                    M
*                                                                            *
* SEE ALSO:                                                                  M
*   SymbSrfDistCrvCrv, SymbCrvCrvInter, SymbSrfDistFindPoints,		     M
*   MvarMVDistSrfSrf							     m
*                                                                            *
* KEYWORDS:                                                                  M
*   MvarMVDistCrvSrf, curve surface distance                                 M
*****************************************************************************/
MvarMVStruct *MvarMVDistCrvSrf(CagdCrvStruct *Crv1,
			       CagdSrfStruct *Srf2,
			       int DistType)
{
    MvarMVStruct *TMV, *DiffMV, *RetMV, *MVNrml,
	*MV1 = MvarCrvToMV(Crv1),
	*MV2 = MvarSrfToMV(Srf2);

    TMV = MvarPromoteMVToMV2(MV1, 3, 0);
    MvarMVFree(MV1);
    MV1 = TMV;

    TMV = MvarPromoteMVToMV2(MV2, 3, 1);
    MvarMVFree(MV2);
    MV2 = TMV;

    DiffMV = MvarMVSub(MV1, MV2);

    switch (DistType) {
	case 0:
	    RetMV = DiffMV;
    	    break;
	case 1:
	default:
	    RetMV = MvarMVDotProd(DiffMV, DiffMV);
	    MvarMVFree(DiffMV);
	    break;
	case 2:
	case 3:
	    if (DistType == 2) {
		CagdCrvStruct
		    *CNrml = SymbCrv3DRadiusNormal(Crv1);

		TMV = MvarCrvToMV(CNrml),
		CagdCrvFree(CNrml);
		MVNrml = MvarPromoteMVToMV2(TMV, 3, 0);
		MvarMVFree(TMV);
	    }
	    else {
		CagdSrfStruct
		    *SNrml = SymbSrfNormalSrf(Srf2);

		TMV = MvarSrfToMV(SNrml);
		CagdSrfFree(SNrml);
		MVNrml = MvarPromoteMVToMV2(TMV, 3, 1);
		MvarMVFree(TMV);
	    }

	    RetMV = MvarMVDotProd(DiffMV, MVNrml);
	    MvarMVFree(DiffMV);
	    MvarMVFree(MVNrml);
	    break;
    }
	
    MvarMVFree(MV1);
    MvarMVFree(MV2);

    return RetMV;
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Given two surfaces, creates a multivariate scalar field representing the M
* distance function square, between them.				     M
*                                                                            *
* PARAMETERS:                                                                M
*   Srf1, Srf2:  The two surfaces, Srf1(u, v) and Srf2(r, t), to form their  M
*                distance function square between them as a multivariate     M
*		 function.						     M
*   DistType:    0 for distance vector function,			     M
*		 1 for distance square scalar function,			     M
*		 2 for distance vector projected on the normal of Srf1,	     M
*		 3 for distance vector projected on the normal of Srf2.	     M
*		 In cases 2 and 3 the normal field is not normalized.        M
*                                                                            *
* RETURN VALUE:                                                              M
*   MvarMVStruct *:  The distance function square d2(u, v, r, t) of the      M
*                    distance from Srf1(u, v) to Srf2(r, t).                 M
*                                                                            *
* SEE ALSO:                                                                  M
*   SymbSrfDistCrvCrv, SymbCrvCrvInter, SymbSrfDistFindPoints,		     M
*   MvarMVDistCrvSrf							     M
*                                                                            *
* KEYWORDS:                                                                  M
*   MvarMVDistSrfSrf, surface surface distance                               M
*****************************************************************************/
MvarMVStruct *MvarMVDistSrfSrf(CagdSrfStruct *Srf1,
			       CagdSrfStruct *Srf2,
			       int DistType)
{
    CagdSrfStruct *SNrml;
    MvarMVStruct *TMV, *DiffMV, *RetMV, *MVNrml,
	*MV1 = MvarSrfToMV(Srf1),
	*MV2 = MvarSrfToMV(Srf2);

    TMV = MvarPromoteMVToMV2(MV1, 4, 0);
    MvarMVFree(MV1);
    MV1 = TMV;

    TMV = MvarPromoteMVToMV2(MV2, 4, 2);
    MvarMVFree(MV2);
    MV2 = TMV;

    DiffMV = MvarMVSub(MV1, MV2);

    switch (DistType) {
	case 0:
	    RetMV = DiffMV;
    	    break;
	case 1:
	default:
	    RetMV = MvarMVDotProd(DiffMV, DiffMV);
	    MvarMVFree(DiffMV);
	    break;
	case 2:
	case 3:
	    if (DistType == 2)
		SNrml = SymbSrfNormalSrf(Srf1);
	    else
		SNrml = SymbSrfNormalSrf(Srf2);

	    TMV = MvarSrfToMV(SNrml);
	    MVNrml = MvarPromoteMVToMV2(TMV, 4,
					DistType == 2 ? 0 : 2);
	    MvarMVFree(TMV);

	    RetMV = MvarMVDotProd(DiffMV, MVNrml);
	    MvarMVFree(DiffMV);
	    MvarMVFree(MVNrml);
	    break;
    }
	
    MvarMVFree(MV1);
    MvarMVFree(MV2);

    return RetMV;
}
