/******************************************************************************
* MvBzrSym.c - Bezier symbolic computation for multivariate library.	      *
*******************************************************************************
* (C) Gershon Elber, Technion, Israel Institute of Technology                 *
*******************************************************************************
* Written by Gershon Elber, July 97..					      *
******************************************************************************/

#include <ctype.h>
#include <stdio.h>
#include <string.h>
#include "mvar_loc.h"

#define NODE_EQUAL_SHIFT 0.8

/*****************************************************************************
* DESCRIPTION:                                                               M
* Given two Bezier multivariates - multiply them coordinatewise.	     M
*   The two multivariates are promoted to same point type before             M
* multiplication can take place.					     M
*   See also BzrMultInterpFlag.                                              M
*                                                                            *
* PARAMETERS:                                                                M
*   MV1, MV2:   The two multivariates to multiply.	                     M
*                                                                            *
* RETURN VALUE:                                                              M
*   MvarMVStruct *:  The product MV1 * MV2 coordinatewise.                   M
*                                                                            *
* KEYWORDS:                                                                  M
*   MvarBzrMVMult, product                                                   M
*****************************************************************************/
MvarMVStruct *MvarBzrMVMult(MvarMVStruct *MV1, MvarMVStruct *MV2)
{
    int i, *ProdIndices, *Indices1, *Indices2;
    CagdBType UseCagdIChooseKTable, IsNotRational;
    MvarMVStruct *ProdMV;
    int MaxCoord, *Lengths, Size;
    CagdRType **Points, **Points1, **Points2;

    if (!MVAR_IS_BEZIER_MV(MV1) || !MVAR_IS_BEZIER_MV(MV2)) {
	MVAR_FATAL_ERROR(MVAR_ERR_BEZIER_EXPECTED);
	return NULL;
    }

    MV1 = MvarMVCopy(MV1);
    MV2 = MvarMVCopy(MV2);
    if (!MvarMakeMVsCompatible(&MV1, &MV2, FALSE, FALSE)) {
	MVAR_FATAL_ERROR(MVAR_ERR_FAIL_CMPT);
	return NULL;
    }

    Lengths = (int *) IritMalloc(sizeof(int) * MV1 -> Dim);
    for (i = 0; i < MV1 -> Dim; i++)
	Lengths[i] = MV1 -> Lengths[i] +  MV2 -> Lengths[i] - 1;
    ProdMV = MvarBzrMVNew(MV1 -> Dim, Lengths, MV1 -> PType);
    IritFree((VoidPtr) Lengths);

    IsNotRational = !MVAR_IS_RATIONAL_MV(ProdMV);
    MaxCoord = CAGD_NUM_OF_PT_COORD(ProdMV -> PType);
    Size = MVAR_CTL_MESH_LENGTH(ProdMV);

    Points = ProdMV -> Points;
    Points1 = MV1 -> Points;
    Points2 = MV2 -> Points;

    for (i = IsNotRational; i <= MaxCoord; i++)
	ZAP_MEM(Points[i], sizeof(CagdRType) * Size);

    ProdIndices = (int *) IritMalloc(sizeof(int) * ProdMV -> Dim);
    Indices1 = (int *) IritMalloc(sizeof(int) * MV1 -> Dim);
    Indices2 = (int *) IritMalloc(sizeof(int) * MV2 -> Dim);

    UseCagdIChooseKTable = TRUE;
    for (i = 0; i < ProdMV -> Dim; i++) {
        if (ProdMV -> Orders[i] >= CAGD_MAX_BEZIER_CACHE_ORDER) {
	    UseCagdIChooseKTable = FALSE;
	    break;
	}
    }

    ZAP_MEM(Indices1, sizeof(int) * MV1 -> Dim);
    do {
	ZAP_MEM(Indices2, sizeof(int) * MV2 -> Dim);
	do {
	    int ProdIndex,
	        Index1 = MvarGetPointsMeshIndices(MV1, Indices1),
		Index2 = MvarGetPointsMeshIndices(MV2, Indices2),
		*I1 = Indices1,
		*I2 = Indices2,
		*IP = ProdIndices,
		*O1 = MV1 -> Orders,
		*O2 = MV2 -> Orders,
		*OP = ProdMV -> Orders;
	    CagdRType Coef, CoefNumer, CoefDenom;

	    if (UseCagdIChooseKTable) {
	        /* Use the precomputed table */
	      *IP = *I1 + *I2;

	      CoefNumer = CagdIChooseKTable[*O1++ - 1][*I1++] *
		          CagdIChooseKTable[*O2++ - 1][*I2++];
	      CoefDenom = CagdIChooseKTable[*OP++ - 1][*IP++];

	        for (i = 1; i < ProdMV -> Dim; i++) {
		    *IP = *I1 + *I2;

		    CoefNumer *= CagdIChooseKTable[*O1++ - 1][*I1++] *
		                 CagdIChooseKTable[*O2++ - 1][*I2++];
		    CoefDenom *= CagdIChooseKTable[*OP++ - 1][*IP++];
		}
	    }
	    else {
	        /* This is going to be slow! */
	      *IP = *I1 + *I2;

	      CoefNumer = CagdIChooseK(*I1++, *O1++ - 1) *
		          CagdIChooseK(*I2++, *O2++ - 1);
	      CoefDenom = CagdIChooseK(*IP++, *OP++ - 1);

	        for (i = 1; i < ProdMV -> Dim; i++) {
		    *IP = *I1 + *I2;

		    CoefNumer *= CagdIChooseK(*I1++, *O1++ - 1) *
		                 CagdIChooseK(*I2++, *O2++ - 1);
		    CoefDenom *= CagdIChooseK(*IP++, *OP++ - 1);
		}
	    }
	    Coef = CoefNumer / CoefDenom;

	    ProdIndex = MvarGetPointsMeshIndices(ProdMV, ProdIndices);

	    for (i = IsNotRational; i <= MaxCoord; i++)
		Points[i][ProdIndex] +=
		    Coef * Points1[i][Index1] * Points2[i][Index2];

	}
	while (MvarIncrementMeshIndices(MV2, Indices2));
    }
    while (MvarIncrementMeshIndices(MV1, Indices1));

    IritFree((VoidPtr) ProdIndices);
    IritFree((VoidPtr) Indices1);
    IritFree((VoidPtr) Indices2);

    MvarMVFree(MV1);
    MvarMVFree(MV2);

    return ProdMV;
}

/*****************************************************************************
* DESCRIPTION:                                                               M
* Given a rational Bezier multivariates - computes its derivative surface    M
* in direction Dir, using the quotient rule for differentiation.	     M
*                                                                            *
* PARAMETERS:                                                                M
*   MV:         Rational Bezier multivariate to differentiate.               M
*   Dir:        Direction of Differentiation.				     M
*                                                                            *
* RETURN VALUE:                                                              M
*   MvarMVStruct *:    Differentiated rational Bezier  multivariate.         M
*                                                                            *
* SEE ALSO:                                                                  M
*   MvarMVDerive, MvarBzrMVDerive, MvarBzrMVDerive, MvarBzrMVDeriveRational  M
*                                                                            *
* KEYWORDS:                                                                  M
*   MvarBzrMVDeriveRational, derivatives                                     M
*****************************************************************************/
MvarMVStruct *MvarBzrMVDeriveRational(MvarMVStruct *MV, MvarMVDirType Dir)
{
    int i;
    MvarMVStruct *DMVScalars[CAGD_MAX_PT_SIZE], *MVScalars[CAGD_MAX_PT_SIZE],
	*TMV1, *TMV2, *DeriveMV;

    MVAR_SPLIT_SCALARS(MV, MVScalars);

    if (MVScalars[0])
	DMVScalars[0] = MvarBzrMVDerive(MVScalars[0], Dir);
    else
	MVAR_FATAL_ERROR(MVAR_ERR_RATIONAL_EXPECTED);

    for (i = 1; i <= CAGD_MAX_PT_COORD && MVScalars[i] != NULL; i++) {
	DMVScalars[i] = MvarBzrMVDerive(MVScalars[i], Dir);

	TMV1 = MvarBzrMVMult(DMVScalars[i], MVScalars[0]);
	TMV2 = MvarBzrMVMult(MVScalars[i],  DMVScalars[0]);

	MvarMVFree(DMVScalars[i]);
	DMVScalars[i] = MvarMVSub(TMV1, TMV2);
	MvarMVFree(TMV1);
	MvarMVFree(TMV2);
    }
    
    TMV1 = MvarBzrMVMult(MVScalars[0], MVScalars[0]);
    MvarMVFree(MVScalars[0]);
    MVScalars[0] = TMV1;

    MVAR_FREE_SCALARS(MVScalars);

    DeriveMV = MvarMVMergeScalar(DMVScalars);

    MVAR_FREE_SCALARS(DMVScalars);

    return DeriveMV;
}
