/*****************************************************************************
* A facility to convert line segments into triangles.                        *
******************************************************************************
* (C) Gershon Elber, Technion, Israel Institute of Technology                *
******************************************************************************
* Written by:  David Shafrir & Alex Reicher       Ver 0.3, Sep. 2003         *
*****************************************************************************/

#include "polyline.h"
#include <geom_lib.h>

static int LineCross(PointType d1,
		     PointType d2,
		     PointType p1,
		     PointType p2,
		     PointType Result);

/*****************************************************************************
* DESCRIPTION:                                                               *
*   Computes intersection between two lines, given in (point,vector) form.   *
*                                                                            *
* PARAMETERS:                                                                *
*   d1:       IN, the direction vector of the first line.                    *
*   d2:       IN, the direction vector of the second line.                   *
*   p1:       IN, a point on the first line.                                 *
*   p1:       IN, a point on the second line.                                *
*   Result:        OUT, the intersection point.                              *
*                                                                            *
* RETURN VALUE:                                                              *
*    int : was there an itersection.                                         *
*****************************************************************************/
static int LineCross(PointType d1,
		     PointType d2,
		     PointType p1,
		     PointType p2,
		     PointType Result)
{
    RealType Det, A1, A2;

    Det = d1[1] * (-d2[0]) - d2[1] * (-d1[0]);
    if (APX_EQ(Det, 0))
        return FALSE;                       /* Line projections are parallel.*/

    A1 = p1[X_AXIS] * d1[1] - p1[Y_AXIS] * d1[0];
    A2 = p2[X_AXIS] * d2[1] - p2[Y_AXIS] * d2[0];
    Result[X_AXIS] = (A1 * (-d2[0]) - A2 * (-d1[0])) / Det;
    Result[Y_AXIS] = (d1[1] * A2 - d2[1] * A1) / Det;
    Result[Z_AXIS] = (p1[Z_AXIS] + p2[Z_AXIS]) * 0.5;
    Result[W_AXIS] = (p1[W_AXIS] + p2[W_AXIS]) * 0.5;

    return TRUE;
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Initialize the segment structure, using the PolylineOptions.             M
*   Should be called when object is created.                                 M
*                                                                            *
* PARAMETERS:                                                                M
*   Seg:           IN, OUT, pointer to the line segment.                     M
*   PolyOptions:   IN, the polyline options structure.                       M
*                                                                            *
* RETURN VALUE:                                                              M
*   void                                                                     M
*                                                                            *
* KEYWORDS:                                                                  M
*   LineSegmentInit, Initialize Options                                      M
*****************************************************************************/
void LineSegmentInit(LineSegmentStruct *Seg,
		     PolylineOptionsStruct *PolyOptions)
{
    Seg -> Tri = IPAllocPolygon(0, NULL, NULL);
    Seg -> TriVertex[0] = Seg -> Tri -> PVertex =
        IPAllocVertex2(Seg -> Tri -> PVertex);
    Seg -> TriVertex[1] = Seg -> TriVertex[0] -> Pnext =
        IPAllocVertex2(Seg -> TriVertex[1]);
    Seg -> TriVertex[2] = Seg -> TriVertex[1] -> Pnext =
        IPAllocVertex2(Seg -> TriVertex[2]);
    Seg -> TriVertex[2] -> Pnext = NULL;
    Seg -> Tri -> Plane[X_AXIS] = 0;
    Seg -> Tri -> Plane[Y_AXIS] = 0;
    Seg -> Tri -> Plane[Z_AXIS] = -1;
    Seg -> Tri -> Plane[W_AXIS] = 0;
    IPUpdateVrtxNrml(Seg -> Tri, Seg -> Tri -> Plane);
    Seg -> TrianglesNum = 0;
    if ( PolyOptions)
        LineSegmentSetOptions(Seg, PolyOptions);
    else {
        Seg -> k = 0;
        Seg -> PolyOptions.MaxWidth = Seg -> PolyOptions.MinWidth = 0.01;
        Seg -> PolyOptions.ZNear = Seg -> PolyOptions.ZFar = 0;
    }
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Changes the PolyOptions.                                                 M
*                                                                            *
* PARAMETERS:                                                                M
*   Seg:           IN, OUT, pointer to the line segment.                     M
*   PolyOptions:   IN, the polyline options structure.                       M
*                                                                            *
* RETURN VALUE:                                                              M
*   void                                                                     M
*                                                                            *
* KEYWORDS:                                                                  M
*   LineSegmentSetOptions,Initialize Options                                 M
*****************************************************************************/
void LineSegmentSetOptions(LineSegmentStruct *Seg,
			   PolylineOptionsStruct *PolyOptions)
{
    if (!PolyOptions)
	return;

    memcpy(&Seg -> PolyOptions, PolyOptions, sizeof(PolylineOptionsStruct));

    Seg -> k = APX_EQ(PolyOptions -> ZNear, PolyOptions -> ZFar) ?
        0 : (PolyOptions -> MaxWidth - PolyOptions -> MinWidth) /
        (PolyOptions -> ZNear - PolyOptions -> ZFar);
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Frees the memory allocated by the object.                                M
*                                                                            *
* PARAMETERS:                                                                M
*   Seg:   IN, OUT, pointer to the line segment.                             M
*                                                                            *
* RETURN VALUE:                                                              M
*   void                                                                     M
*                                                                            *
* KEYWORDS:                                                                  M
*   LineSegmentRelease, Release free Options                                 M
*****************************************************************************/
void LineSegmentRelease(LineSegmentStruct *Seg)
{
    IPFreePolygon(Seg -> Tri);
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Begins a new line.                                                       M
*                                                                            *
* PARAMETERS:                                                                M
*   Seg:   IN, OUT, pointer to the line segment.                             M
*                                                                            *
* RETURN VALUE:                                                              M
*   void                                                                     M
*                                                                            *
* KEYWORDS:                                                                  M
*   LineSegmentStart, begin start                                            M
*****************************************************************************/
void LineSegmentStart(LineSegmentStruct *Seg)
{
    Seg -> NumVertex = -1;
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Sets the next point for the line.                                        M
*                                                                            *
* PARAMETERS:                                                                M
*   Seg:      IN, OUT, pointer to the line segment.                          M
*   Vertex:   IN, the new point.                                             M
*                                                                            *
* RETURN VALUE:                                                              M
*   void                                                                     M
*                                                                            *
* KEYWORDS:                                                                  M
*   LineSegmentSet, add point line-to                                        M
*****************************************************************************/
void LineSegmentSet(LineSegmentStruct *Seg, PointType4 Vertex)
{
    RealType Width, Length;
    PointType d;
    PointType4 LastPoint, p1, p2;
    int WasIntersection;

    Seg -> NumVertex++;
    if (Seg -> NumVertex == 0) {
        PT_COPY4(Seg -> LastPoint, Vertex);
        Seg -> TrianglesNum  = 0 ;
        return;
    }
    PT_SUB(d, Vertex, Seg -> LastPoint);

    Length = sqrt(SQR(d[X_AXIS]) + SQR(d[Y_AXIS]));
    PT_SCALE(d, 1 / Length);

    if (APX_EQ(Length, 0)) {
        Seg -> TrianglesNum = 0;
        return;                                 /* Segment has zero length.  */
    }

    if (Seg -> NumVertex == 1) {
        PT_COPY4(LastPoint, Seg -> LastPoint);
        Width = (LastPoint[Z_AXIS] - Seg -> PolyOptions.ZFar) * Seg ->
            k + Seg -> PolyOptions.MinWidth;
        PT_COPY4(Seg -> Vertex[2], LastPoint);

        Seg -> Vertex[2][X_AXIS] -= Width * d[Y_AXIS];
        Seg -> Vertex[2][Y_AXIS] += Width * d[X_AXIS];

        PT_COPY4(Seg -> Vertex[3], LastPoint);
        Seg -> Vertex[3][X_AXIS] += Width * d[Y_AXIS];
        Seg -> Vertex[3][Y_AXIS] -= Width * d[X_AXIS];
        Seg -> TrianglesNum = 0;
        /* Don't return triangle. */
    }
    else {
        Seg -> BendType = BEND_NORM;
        if (DOT_PROD_2D(Seg -> LastDelta, d) < 0.0) {
            Seg -> BendType = CROSS_PROD_2D(Seg -> LastDelta, d) > 0 ?
                BEND_SHARP_CCW : BEND_SHARP_CW;
        }
        PT_COPY4(Seg -> Vertex[0], Seg -> Vertex[2]);
        PT_COPY4(Seg -> Vertex[1], Seg -> Vertex[3]);

        Width = (Seg -> LastPoint[Z_AXIS] - Seg -> PolyOptions.ZFar) *
            Seg -> k + Seg -> PolyOptions.MinWidth;
        PT_COPY4(p1, Seg -> LastPoint);
        PT_COPY4(p2, Seg -> LastPoint);
        p1[X_AXIS] -= Width * Seg -> LastDelta[Y_AXIS];
        p2[X_AXIS] -= Width * d[Y_AXIS];

        p1[Y_AXIS] += Width * Seg -> LastDelta[X_AXIS];
        p2[Y_AXIS] += Width * d[X_AXIS];

        WasIntersection = LineCross(Seg -> LastDelta, d,
				    p1, p2, Seg -> Vertex[2]);
        if (!WasIntersection ||
	    Seg -> BendType == BEND_SHARP_CW ||
	    Seg -> BendType == BEND_SHARP_CCW) {
            PT_COPY4(Seg -> Vertex[2], p2);
            PT_COPY4(Seg -> Vertex[4], p1);                /* For sharp bend.*/
        }
        PT_COPY4(p1, Seg -> LastPoint);
        PT_COPY4(p2, Seg -> LastPoint);
        p1[X_AXIS] += Width * Seg -> LastDelta[Y_AXIS] ;
        p2[X_AXIS] += Width * d[Y_AXIS] ;

        p1[Y_AXIS] -= Width * Seg -> LastDelta[X_AXIS] ;
        p2[Y_AXIS] -= Width * d[X_AXIS] ;

        LineCross(Seg -> LastDelta, d, p1, p2, Seg -> Vertex[3]);
        if  (!WasIntersection ||
	     Seg -> BendType == BEND_SHARP_CW ||
	     Seg -> BendType == BEND_SHARP_CCW) {
            PT_COPY4(Seg -> Vertex[3], p2);
            PT_COPY4(Seg -> Vertex[4], p1);               /* For sharp bend. */
        }

        /* Two new traingles can be returned. */
        Seg -> TrianglesNum = Seg -> BendType  == BEND_NORM ? 2 : 3;
    }
    PT_COPY4(Seg -> LastPoint, Vertex);
    PT_COPY(Seg -> LastDelta, d);
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Ends a  line.Should be called when after the last point was added.       M
*                                                                            *
* PARAMETERS:                                                                M
*   Seg:   IN, OUT, pointer to the line segment.                             M
*                                                                            *
* RETURN VALUE:                                                              M
*   void                                                                     M
*                                                                            *
* KEYWORDS:                                                                  M
*   LineSegmentEnd, line termination                                         M
*****************************************************************************/
void LineSegmentEnd(LineSegmentStruct *Seg)
{
    /* We simulate as if there is a new vertex in the tangent direction. */
    PointType4 Pt;

    PT_COPY4(Pt, Seg -> LastPoint);
    Pt[0] += Seg -> LastDelta[0];
    Pt[1] += Seg -> LastDelta[1];
    LineSegmentSet(Seg, Pt);
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Retrives the triangles compromising the current segment.                 M
*                                                                            *
* PARAMETERS:                                                                M
*   Seg:      IN, OUT, pointer to the line segment.                          M
*   NumTri:   IN, the number of triangle, shold be < TrianglesNum.           M
*                                                                            *
* RETURN VALUE:                                                              M
*   IPPolygonStruct *: the triangle.                                         M
*                                                                            *
* KEYWORDS:                                                                  M
*   LineSegmentGetTri, triangulation                                         M
*****************************************************************************/
IPPolygonStruct *LineSegmentGetTri(LineSegmentStruct *Seg, int NumTri)
{
    int Map[3], i;

    if (NumTri >= Seg -> TrianglesNum) {
        return NULL;
    }

    if (NumTri < 2) {
        Map[0] = 0;
        Map[1] = Seg -> BendType != BEND_SHARP_CCW ? 3 : 4;
        Map[2] =  NumTri == 0 ? 1 : (Seg -> BendType != BEND_SHARP_CW ? 2 : 4);
    }
    else {
        /* Third traingle in a sharp bend. */
        if (Seg -> BendType == BEND_SHARP_CCW) {
            Map[0] = 2; Map[1] = 4; Map[2] = 3;
        }
        else {
            Map[0] = 2; Map[1] = 3; Map[2] = 4;
        }
    }

    for ( i = 0; i < 3 ; i++) {
        PT_COPY(Seg -> TriVertex[i] -> Coord, Seg -> Vertex[ Map[i] ]);
        AttrSetRealAttrib(&Seg -> TriVertex[i] ->  Attr,
            "_1/W", Seg -> Vertex[ Map[i] ][3]);
    }

    return Seg -> Tri;
}
