/******************************************************************************
* Bump_Map.c - real bump mapping support.         			      *
*******************************************************************************
* (C) Gershon Elber, Technion, Israel Institute of Technology                 *
*******************************************************************************
* Written by Gershon Elber, Nov. 2001.					      *
******************************************************************************/

#include "irit_sm.h"
#include "cagd_lib.h"
#include "symb_lib.h"
#include "iritprsr.h"
#include "ip_cnvrt.h"
#include "bool_lib.h"
#include "allocate.h"
#include "geom_lib.h"
#include "user_loc.h"

STATIC_DATA int GlblLclUV, GlblSrfUClosed, GlblSrfVClosed, GlblShiftDirPower;
STATIC_DATA CagdVType GlblShiftDir;
STATIC_DATA CagdRType GlblU, GlblV, GlblDu, GlblDv;
STATIC_DATA CagdSrfStruct *GlblSrf, *GlblDuSrf, *GlblDvSrf;
STATIC_DATA IPObjectStruct *GlblRetObjs;

static void UserBumpMapProcessGeom(IPObjectStruct *PObj, MatrixType Mat);

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Maps the given bump texture defined in Pls over the surface Srf,         M
* duplicating it (UDup x VDup) times over the parametric domain of Srf.      M
*                                                                            *
* PARAMETERS:                                                                M
*   Srf:     Surface to derive a bump mapping for.                           M
*   Texture: Geometry defining the single tile of the bump texture.          M
*	     The geometry is assumed to span [0..1] in both x and y.	     M
*   UDup, VDup: The U and V duplication factors.			     M
*   LclUV:   TRUE to keep local UV coordinates for each tile, FALSE to       M
*	     employ the UV coordinates of the surface Srf.		     M
*                                                                            *
* RETURN VALUE:                                                              M
*   IPObjectStruct *:  A set of objects that derive the bumpy surface        M
*	above Srf.							     M
*                                                                            *
* SEE ALSO:                                                                  M
*                                                                            M
*                                                                            *
* KEYWORDS:                                                                  M
*   UserBumpMapPolysOverSrf                                                  M
*****************************************************************************/
IPObjectStruct *UserBumpMapPolysOverSrf(CagdSrfStruct *Srf,
					IPObjectStruct *Texture,
					RealType UDup,
					RealType VDup,
					int LclUV)
{
    CagdRType UMin, UMax, VMin, VMax;

    GlblShiftDirPower = AttrGetObjectIntAttrib(Texture, "DispTextDirPower");

    if (!IP_ATTR_IS_BAD_INT(GlblShiftDirPower)) {
	char
	    *p = AttrGetObjectStrAttrib(Texture, "DispTextDir");

	if (sscanf(p, "%lf,%lf,%lf",
		   &GlblShiftDir[0],
		   &GlblShiftDir[1],
		   &GlblShiftDir[2]) != 3 &&
	    sscanf(p, "%lf %lf %lf",
		   &GlblShiftDir[0],
		   &GlblShiftDir[1],
		   &GlblShiftDir[2]) != 3) {
	    GlblShiftDir[0] = GlblShiftDir[1] = 0.0;
	    GlblShiftDir[2] = -1.0;
	}
    }

    GlblRetObjs = IPGenLISTObject(NULL);

    GlblLclUV = LclUV;

    GlblSrf = Srf;
    GlblDuSrf = CagdSrfDerive(Srf, CAGD_CONST_U_DIR),
    GlblDvSrf = CagdSrfDerive(Srf, CAGD_CONST_V_DIR);

    GlblSrfUClosed = CagdIsClosedSrf(Srf, CAGD_CONST_U_DIR),
    GlblSrfVClosed = CagdIsClosedSrf(Srf, CAGD_CONST_V_DIR);

    CagdSrfDomain(Srf, &UMin, &UMax, &VMin, &VMax);
    GlblDu = (UMax - UMin) / UDup;
    GlblDv = (VMax - VMin) / VDup;

    for (GlblU = UMin; GlblU < UMax; GlblU += GlblDu) {
	for (GlblV = VMin; GlblV < VMax; GlblV += GlblDv) {
	    MatrixType Mat;

	    MatGenUnitMat(Mat);

	    IPTraverseObjListHierarchy(Texture, Mat, UserBumpMapProcessGeom);
	}
    }

    CagdSrfFree(GlblDuSrf);
    CagdSrfFree(GlblDvSrf);

    return GlblRetObjs;
}

/*****************************************************************************
* DESCRIPTION:                                                               *
*   Process a single leaf in the bump texture geometry.                      *
*                                                                            *
* PARAMETERS:                                                                *
*   PObj:   Object to process.  A leaf in the geometry tree.                 *
*   Mat:    Transformation matrix in the tree hierarchy.  Ignored here.      *
*                                                                            *
* RETURN VALUE:                                                              *
*   void                                                                     *
*****************************************************************************/
static void UserBumpMapProcessGeom(IPObjectStruct *PObj, MatrixType Mat)
{
    CagdRType UMin, UMax, VMin, VMax;

    CagdSrfDomain(GlblSrf, &UMin, &UMax, &VMin, &VMax);

    if (IP_IS_POLY_OBJ(PObj)) {
        int l;
	IPPolygonStruct *Pl;
        IPObjectStruct
	    *CpObj = IPCopyObject(NULL, PObj, TRUE);

	for (Pl = CpObj -> U.Pl; Pl != NULL; Pl = Pl -> Pnext) {
	    CagdRType Vu, Vv, z, t, *R;
	    CagdPType SrfPos;
	    CagdVType DuVec, DvVec, NrmlVec, TVec;
	    IPVertexStruct *V;

	    V = Pl -> PVertex;
	    do {
	        Vu = V -> Coord[0] * GlblDu + GlblU;
		Vv = V -> Coord[1] * GlblDv + GlblV;

		/* Make sure inside surface domain. */
		while (TRUE) {
		    if (GlblSrfUClosed) {
		        if (Vu > UMax)
			    Vu = UMin + (Vu - UMax);
			else if (Vu < UMin)
			    Vu = UMax + (UMin - Vu);
			else
			    break;
		    }
		    else {
		        if (Vu > UMax)
			    Vu = UMax;
			else if (Vu < UMin)
			    Vu = UMin;
			else
			    break;
		    }
		}
		while (TRUE) {
		    if (GlblSrfVClosed) {
		        if (Vv > VMax)
			    Vv = VMin + (Vv - VMax);
			else if (Vv < VMin)
			    Vv = VMax + (VMin - Vv);
			else
			    break;
		    }
		    else {
		        if (Vv > VMax)
			    Vv = VMax;
			else if (Vv < VMin)
			    Vv = VMin;
			else
			    break;
		    }
		}

		/* Evaluate local frame in Euclidean space. */
		z = V -> Coord[2];
		R = CagdSrfEval(GlblSrf, Vu, Vv);
		CagdCoerceToE3(SrfPos, &R, -1, GlblSrf -> PType);
		PT_COPY(V -> Coord, SrfPos);

		/* Compute a frame spanning R^3: */ 
		R = CagdSrfEval(GlblDuSrf, Vu, Vv);
		CagdCoerceToE3(DuVec, &R, -1, GlblDuSrf -> PType);
		VEC_NORMALIZE(DuVec);

		R = CagdSrfEval(GlblDvSrf, Vu, Vv);
		CagdCoerceToE3(DvVec, &R, -1, GlblDvSrf -> PType);
		VEC_NORMALIZE(DvVec);

		CROSS_PROD(NrmlVec, DuVec, DvVec);
		VEC_NORMALIZE(NrmlVec);

		/* Evaluate the z contribution. */
		VEC_COPY(TVec, NrmlVec);
		VEC_SCALE(TVec, z);
		VEC_ADD(V -> Coord, V -> Coord, TVec);

		if (GlblShiftDirPower > 0) {
		    t = pow(z, GlblShiftDirPower);
		    VEC_COPY(TVec, GlblShiftDir);
		    VEC_SCALE(TVec, t); 
		    VEC_ADD(V -> Coord, V -> Coord, TVec);
		}
		    
		if (IP_HAS_NORMAL_VRTX(V)) {
		    VEC_SCALE(DuVec,   V -> Normal[0]);
		    VEC_SCALE(DvVec,   V -> Normal[1]);
		    VEC_SCALE(NrmlVec, V -> Normal[2]);
		    VEC_COPY(V -> Normal, DuVec);
		    VEC_ADD(V -> Normal, V -> Normal, DvVec);
		    VEC_ADD(V -> Normal, V -> Normal, NrmlVec);
		    if (GlblShiftDirPower > 0) {
			t = pow(z, GlblShiftDirPower - 1) * GlblShiftDirPower;
			VEC_COPY(TVec, GlblShiftDir);
			VEC_SCALE(TVec, t); 
			VEC_ADD(V -> Normal, V -> Normal, TVec);
		    }
		    VEC_NORMALIZE(V -> Normal);
		}

		if (!GlblLclUV)
		    AttrSetUVAttrib(&V -> Attr, "uvvals", Vu, Vv);

		V = V -> Pnext;
	    }
	    while (V != Pl -> PVertex && V != NULL);
	}

	l = IPListObjectLength(GlblRetObjs);
	IPListObjectInsert(GlblRetObjs, l++, CpObj);
	IPListObjectInsert(GlblRetObjs, l, NULL);
    }
}
