/******************************************************************************
* visible.c - surface visibility analysis routines.			      *
*******************************************************************************
* (C) Gershon Elber, Technion, Israel Institute of Technology                 *
*******************************************************************************
* Written by Gershon Elber, Aug. 99.					      *
******************************************************************************/

#include "irit_sm.h"
#include "allocate.h"
#include "attribut.h"
#include "iritprsr.h"
#include "cagd_lib.h"
#include "symb_lib.h"
#include "geom_lib.h"
#include "user_loc.h"

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Computes a decomposition of surface Srf into regions that are visible    M
* with normals deviating in a cone of size as set via ConeAngle.	     M
*                                                                            *
* PARAMETERS:                                                                M
*   Srf:         To decompose into visibility cones.                         M
*   Resolution:  Of polygonal approximation of surface Srf.                  M
*   ConeAngle:   Of coverage over the unit sphere.  ConeAngle prescribes the M
*		 opening angle of the cone. In Radians.			     M
*                                                                            *
* RETURN VALUE:                                                              M
*   IPObjectStruct *:  A list of trimmed surface regions of surface Srf that M
*		       are visible from a selected direction (available as   M
*		       "VIewDir" attribute on them) ad that the union of the M
*		       list covers the entire surface Srf.		     M
*                                                                            *
* SEE ALSO:                                                                  M
*   UserViewingConeSrfDomains, UserVisibilityClassify                        M
*                                                                            *
* KEYWORDS:                                                                  M
*   UserSrfVisibConeDecomp                                                   M
*****************************************************************************/
IPObjectStruct *UserSrfVisibConeDecomp(CagdSrfStruct *Srf,
				       CagdRType Resolution,
				       CagdRType ConeAngle)
{
    CagdRType
	ConeSize = sin(ConeAngle);
    CagdSrfStruct
        *NSrf = SymbSrfNormalSrf(Srf);
    IPObjectStruct *PTmp,
	*RetObjs = NULL,
        *SpherePtsDist = GMPointCoverOfUnitHemiSphere(ConeSize),
        *SrfDomains = UserViewingConeSrfDomains(Srf,
						NSrf,
						SpherePtsDist -> U.Pl,
						Resolution,
						ConeAngle,
						FALSE);

    for (PTmp = SrfDomains; PTmp != NULL; PTmp = PTmp -> Pnext) {
	TrimSrfStruct
	    *TSrfList = TrimSrfsFromContours(Srf, PTmp -> U.Pl);
	IPObjectStruct *PObj,
	    *ViewDirObj = AttrGetObjectObjAttrib(PTmp, "ViewDir");

	/* Eliminate the trimmed surfaces in PObj that are outside the     */
	/* viewing cone, in place.					   */
	TSrfList = UserVisibilityClassify(
				    AttrGetObjectObjAttrib(PTmp, "SclrSrf"),
				    TSrfList);
	if (TSrfList != NULL) {
	    PObj = IPGenTRIMSRFObject(TSrfList);
	    if (ViewDirObj != NULL)
	        AttrSetObjectObjAttrib(PObj, "ViewDir", ViewDirObj, TRUE);
	    LIST_PUSH(PObj, RetObjs);
	}
    }

    CagdSrfFree(NSrf);
    IPFreeObjectList(SrfDomains);
    IPFreeObject(SpherePtsDist);

    return RetObjs;
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Given a decomposition of surface Srf into a set of TrimmedSrfs into      M
* regions that are inside the normal viewing cone as prescribed by SclrSrf,  M
* eliminate all regions (trimmed surfaces) that are outside the viewing      M
*  cone in the given list, TrimmedSrfs.					     M
*                                                                            *
* PARAMETERS:                                                                M
*   SclrSrf:	   The scalar surface computed symbollically in		     M
*		   UserViewingConeSrfDomains, forming this decomosition.     M
*   TrimmedSrfs:   To verify they are within the viewing code of ViewingDir. M
*		   Trimmed surfaces inside this list that are outside the    M
*		   viewing cone are eliminated, in place.		     M
*                                                                            *
* RETURN VALUE:                                                              M
*   TrimSrfStruct *: The trimmed regions inside the viewing cone, as a       M
*		   subset of TrimmedSrfs.			             M
*                                                                            *
* SEE ALSO:                                                                  M
*   UserViewingConeSrfDomains, UserSrfVisibConeDecomp                        M
*                                                                            *
* KEYWORDS:                                                                  M
*   UserVisibilityClassify                                                   M
*****************************************************************************/
TrimSrfStruct *UserVisibilityClassify(IPObjectStruct *SclrSrf,
				      TrimSrfStruct *TrimmedSrfs)
{
    TrimSrfStruct *TSrf,
	*PrevTSrf = NULL;

    if (SclrSrf == NULL || !IP_IS_SRF_OBJ(SclrSrf)) {
	UserFatalError(USER_ERR_MISSING_ATTRIB);
	return NULL;
    }

    for (TSrf = TrimmedSrfs; TSrf != NULL; ) {
	CagdRType
	    *UVInside = TrimPointInsideTrimmedCrvs(TSrf -> TrimCrvList, TSrf),
	    *R = CagdSrfEval(SclrSrf -> U.Srfs, UVInside[0], UVInside[1]);

	if (R[1] > 0) {
	    if (PrevTSrf != NULL) {
		PrevTSrf -> Pnext = TSrf -> Pnext;
		TrimSrfFree(TSrf);
		TSrf = PrevTSrf -> Pnext;
	    }
	    else {
		TrimmedSrfs = TrimmedSrfs -> Pnext;
		TrimSrfFree(TSrf);
		TSrf = TrimmedSrfs;
	    }
	}
	else {
	    PrevTSrf = TSrf;
	    TSrf = TSrf ->Pnext;
	}
    }

    return TrimmedSrfs;
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Computes the domain in given surface Srf that is inside the given cones  M
* of directions ConeDirs and opening angle ConeAngle.			     M
*                                                                            *
* PARAMETERS:                                                                M
*   Srf:           To compute its visibility.                                M
*   NSrf:	   The normal surface computed symbollically for Srf.        M
*   ConeDirs:      Direction of cone's axes.                                 M
*   Resolution:    Of Cone - normal surface intersection, for polygonal      M
*		   approximation accuracy.				     M
*   ConeAngle:     The opening angle of the cone, in Radians.		     M
*   Euclidean:     Contours are in Euclidean (TRUE) or parametric (FALSE)    M
*		   space of Srf.					     M
*                                                                            *
* RETURN VALUE:                                                              M
*   IPObjectStruct *:   Set of piecewise linear contours, approximating the  M
*		   intersections of the normal viewing cones and the         M
*		   original surface.					     M
*                                                                            *
* SEE ALSO:                                                                  M
*   UserVisibilityClassify, UserSrfVisibConeDecomp                           M
*                                                                            *
* KEYWORDS:                                                                  M
*   UserViewingConeSrfDomains                                                M
*****************************************************************************/
IPObjectStruct *UserViewingConeSrfDomains(CagdSrfStruct *Srf,
					  CagdSrfStruct *NSrf,
					  IPPolygonStruct *ConeDirs,
					  CagdRType Resolution,
					  CagdRType ConeAngle,
					  CagdRType Euclidean)
{
    int OldCirc;
    IPVertexStruct
        *VPts = ConeDirs -> PVertex;
    IPObjectStruct
	*CntrObjList = NULL;

    if (ConeAngle < 0.0 || ConeAngle >= M_PI / (2 + IRIT_EPS)) {
	UserFatalError(USER_ERR_WRONG_ANGLE);
	return NULL;
    }

    for ( ; VPts != NULL; VPts = VPts -> Pnext) {
	STATIC_DATA PlaneType
	    Plane = { 1.0, 0.0, 0.0, 0.0 };
	int i;
	CagdBType HasPosVals, HasNegVals;
	CagdRType *R;
	MatrixType Mat, InvMat;
	CagdSrfStruct *SrfW, *SrfX, *SrfY, *SrfZ, *TmpSrf, *SclrSrf;

	GMGenMatrixZ2Dir(Mat, VPts -> Coord);
	MatInverseMatrix(Mat, InvMat);
	TmpSrf = CagdSrfCopy(NSrf);
	CagdSrfMatTransform(TmpSrf, InvMat);

	SymbSrfSplitScalar(TmpSrf, &SrfW, &SrfX, &SrfY, &SrfZ);
	CagdSrfFree(TmpSrf);
	if (SrfW != NULL) {
	    fprintf(stderr,
		    IRIT_EXP_STR("No support for rationals, at this time.\n"));
	    exit(1);
	}

	TmpSrf = SymbSrfMult(SrfX, SrfX);
	CagdSrfFree(SrfX);
	SrfX = TmpSrf;

	TmpSrf = SymbSrfMult(SrfY, SrfY);
	CagdSrfFree(SrfY);
	SrfY = TmpSrf;

	TmpSrf = SymbSrfMult(SrfZ, SrfZ);
	CagdSrfFree(SrfZ);
	SrfZ = SymbSrfScalarScale(TmpSrf, SQR(tan(ConeAngle)));
	CagdSrfFree(TmpSrf);

	TmpSrf = SymbSrfAdd(SrfX, SrfY);
	CagdSrfFree(SrfX);
	CagdSrfFree(SrfY);

	SclrSrf = SymbSrfSub(TmpSrf, SrfZ);
	CagdSrfFree(TmpSrf);
	CagdSrfFree(SrfZ);

	R = SclrSrf -> Points[1];
	HasPosVals = HasNegVals = FALSE;
	for (i = SclrSrf -> ULength * SclrSrf -> VLength; i >= 0; i--) {
	    if (*R > 0.0)
	        HasPosVals = TRUE;
	    if (*R++ < 0.0)
	        HasNegVals = TRUE;
	}
	if (HasPosVals && HasNegVals) {
	    IPPolygonStruct *Cntrs, *Cntr;

	    OldCirc = IPSetPolyListCirc(TRUE);
	    Cntrs = UserCntrSrfWithPlane(SclrSrf, Plane, Resolution);
	    IPSetPolyListCirc(OldCirc);

	    if (Cntrs != NULL) {
		IPObjectStruct *CntrObj, *VecObj, *SclrSrfObj;
		IPVertexStruct *V;

		if (Euclidean) {
		    for (Cntr = Cntrs; Cntr != NULL; Cntr = Cntr -> Pnext) {
			for (V = Cntr -> PVertex; V != NULL; V = V -> Pnext) {
			    CagdRType
			        *P = CagdSrfEval(Srf, V -> Coord[1],
						      V -> Coord[2]);

			    CagdCoerceToE3(V -> Coord, &P, -1, Srf -> PType);
			}
		    }
		}
		else {
		    for (Cntr = Cntrs; Cntr != NULL; Cntr = Cntr -> Pnext) {
			for (V = Cntr -> PVertex; V != NULL; V = V -> Pnext) {
			    V -> Coord[0] = V -> Coord[1];
			    V -> Coord[1] = V -> Coord[2];
			    V -> Coord[2] = 0.0;
			}
		    }
		}

		CntrObj = IPGenPOLYObject(Cntrs);
		IP_SET_POLYLINE_OBJ(CntrObj);

		/* Save the viewing direction as an attribute. */
		VecObj = IPGenVECObject(&VPts -> Coord[0],
					&VPts -> Coord[1],
					&VPts -> Coord[2]);
		AttrSetObjectObjAttrib(CntrObj, "ViewDir", VecObj, FALSE);

		SclrSrfObj = IPGenSRFObject(SclrSrf);
		AttrSetObjectObjAttrib(CntrObj, "SclrSrf", SclrSrfObj, FALSE);

		LIST_PUSH(CntrObj, CntrObjList);
	    }
	    else {
	        CagdSrfFree(SclrSrf);
	    }
	}
	else {
	    CagdSrfFree(SclrSrf);
	}
    }

    return CntrObjList;
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Computes the views on the unit sphere at which the number of silhouette  M
* curves is changing.  This is a subset of the events typically considered   M
* in aspect graphs and, for example, events such as silhouette cusps are not M
* considered.								     M
*   The viewing directions where the number of silhouette curves/loops can   M
* be modified are views in the tangent plane of the surface at parabolic     M
* points of the surfaces, viewed from the zero principal curvatures'         M
* direction. 								     M
*   Hence, we simply derive the parabolic lines of the surface and then      M
* remap them to the direction of the zero principal curvature there.         M
*                                                                            *
* PARAMETERS:                                                                M
*   PSrf:       To compute the aspect graph's topology partitioning for.     M
*   Tolerance:  Accuracy of parabolic piecewise linear approximation.        M
*                                                                            *
* RETURN VALUE:                                                              M
*   IPPolygonStruct *:   Polylines on the unit sphere, depicting the aspect  M
*		graph's partitioning lines.				     M
*                                                                            *
* SEE ALSO:                                                                  M
*   SymbSrfGaussCurvature, UserCntrSrfWithPlane, MvarSrfSilhInflections      M
*                                                                            *
* KEYWORDS:                                                                  M
*   UserSrfTopoAspectGraph                                                   M
*****************************************************************************/
IPPolygonStruct *UserSrfTopoAspectGraph(CagdSrfStruct *PSrf,
					CagdRType Tolerance)
{
    STATIC_DATA PlaneType
	GaussPlane = { 1.0, 0.0, 0.0, 1.060964e-10 };
    CagdBType
	OldInterpFlag = BspMultInterpFlag(FALSE);
    CagdSrfStruct
	*GaussSrf = SymbSrfGaussCurvature(PSrf, TRUE);
    IPPolygonStruct *Pl, *Plls2,
	*Plls = UserCntrSrfWithPlane(GaussSrf, GaussPlane, Tolerance);

    CagdSrfFree(GaussSrf);

    BspMultInterpFlag(OldInterpFlag);

    if (Plls == NULL)
	return NULL;

    /* Evaluate the principal curvatures/directions along the parabolic     */
    /* curve.  This is somewhat redundant as we have the parabolic lines    */
    /* already but is more robust to precisely evaluate the surface there.  */
    SymbEvalSrfCurvPrep(PSrf, TRUE);
    for (Pl = Plls; Pl != NULL; Pl = Pl -> Pnext) {
	CagdRType K1, K2;
	CagdVType D1, D2, V1;
	IPVertexStruct *V,
	    *VPrev = NULL;

	for (V = Pl -> PVertex; V != NULL; V = V -> Pnext) {
	    SymbEvalSrfCurvature(PSrf, V -> Coord[1], V -> Coord[2], FALSE,
				 &K1, &K2, D1, D2);

	    /* Copy direction with zero curvature as the viewing direction. */
	    if (fabs(K1) < fabs(K2)) {
		VEC_COPY(V -> Coord, D1);
	    }
	    else {
		VEC_COPY(V -> Coord, D2);
	    }

	    /* Check the orientation for consistency.		            */
	    if (VPrev != NULL) {
		VEC_COPY(V1, VPrev -> Coord);
		VEC_SCALE(V1, -1);

		if (DOT_PROD(V -> Coord, V1) >
		    DOT_PROD(V -> Coord, VPrev -> Coord))
		    VEC_SCALE(V -> Coord, -1);
	    }

	    VPrev = V;
	} 
    }
    SymbEvalSrfCurvPrep(PSrf, FALSE);

    /* For each polyline with +V vertices, duplicate and make -V vertices. */
    Plls2 = IPCopyPolygonList(Plls);
    for (Pl = Plls2; Pl != NULL; Pl = Pl -> Pnext) {
	IPVertexStruct *V;

	for (V = Pl -> PVertex; V != NULL; V = V -> Pnext) {
	    VEC_SCALE(V -> Coord, -1);
	}
    }
    IPGetLastPoly(Plls) -> Pnext = Plls2;

    return Plls;
}

#ifdef DEBUG_VISIBILITY

void main(int argc, char **argv)
{
    IPObjectStruct *PObj;
    RealType
	Size = 0.6,
	Resolution = 20;

    if (argc > 1 && strcmp(argv[1], "-s") == 0) {
	Size = atof(argv[2]);
	argv += 2;
	argc -= 2;
    }
    if (argc > 1 && strcmp(argv[1], "-r") == 0) {
	Resolution = atof(argv[2]);
	argv += 2;
	argc -= 2;
    }

    fprintf(stderr, "Doing decomposition of size %f, resolution %f\n",
	    Size, Resolution);

    if ((PObj = IPGetDataFiles(&argv[1], 1, TRUE, TRUE)) != NULL &&
	IP_IS_SRF_OBJ(PObj)) {
	int FileIndex = 1;
	IPObjectStruct *PTmp,
	    *PTrimSrfObjs = UserSrfVisibConeDecomp(PObj -> U.Srfs,
						   Resolution, Size);

	for (PTmp = PTrimSrfObjs; PTmp != NULL; PTmp = PTmp -> Pnext) {
	    int Handler;
	    char FileName[LINE_LEN];

	    sprintf(FileName, "ViewDir%d.itd", FileIndex++);
	    if ((Handler = IPOpenDataFile(FileName,
						FALSE, FALSE)) >= 0) {
		IPPutObjectToHandler(Handler, PTmp);
		IPCloseStream(Handler, TRUE);
	    }
	}
	IPFreeObjectList(PTrimSrfObjs);
    }
    else {
	fprintf(stderr, "Expecting a surface");
    }

    exit(0);
}

#endif /* DEBUG_VISAIBILITY */
